/*
 * $Id: 1541.c,v 1.3 93/06/21 13:36:32 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *
 * $Log:	1541.c,v $
 * Revision 1.3  93/06/21  13:36:32  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:07:03  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:12:53  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * 1541.c -- 1541 Floppy Drive emulation. Emulator emulates 1541 reasonably
 *	     well as long as 1541 is used through ROM-interface (no
 *	     turboloaders supported). Also REL-file type is still missing.
 *
 */

#ifdef __hpux
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#endif

/*#ifdef */
/*#define d_namlen d_reclen  /* Solaris */
/*#endif*/


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include <dirent.h>
#include <sys/param.h>
#include <sys/stat.h>

#ifndef STANDALONE_1541
#include "macro.h"
#include "extern.h"
#else
typedef unsigned char BYTE;
#define UPPER(ad)		(((ad)>>8)&0xff)
#define LOWER(ad)		((ad)&0xff)
#ifdef __cplusplus
#ifdef __STDC__
#undef __STDC__
#endif
#define __STDC__ 0   /* kludge */
#endif
#endif

#include "serial.h"

typedef struct bufferinfo_s {
    int     mode;	/* mode on this buffer */
    int     readmode;	/* is this for reading or writing */
    int     secondary;	/* this is not needed */
    BYTE   *buffer;	/* use this to save data */
    BYTE   *slot;	/* save data for directory-slot */
    int     bufptr;	/* use this to save/read data to disk */
    int     length;	/* directory-read length */
}       bufferinfo_t;


#define BUFFER_NOT_IN_USE		0
#define BUFFER_DIRECTORY_READ		1
#define BUFFER_SEQUENTIAL		2
#define BUFFER_MEMORY_BUFFER		3
#define BUFFER_OTHER			4
#define BUFFER_COMMAND_CHANNEL		5


#define WRITE_BLOCK             512


/*
 * Actually serial-code errors ...
 */
#define FLOPPY_COMMAND_OK		0
#define FLOPPY_ERROR			(2)
#define FLOPPY_EOF			(64)

#define SLOT_TYPE_OFFSET		2
#define SLOT_FIRST_TRACK		3
#define SLOT_FIRST_SECTOR		4
#define SLOT_NAME_OFFSET		5
#define SLOT_NR_BLOCKS			30

#define BAM_FIRST_TRACK			0
#define BAM_FIRST_SECTOR		1
#define BAM_FORMAT_TYPE			2
#define BAM_BIT_MAP			4
#define BAM_DISK_NAME			144
#define BAM_DISK_ID			162
#define BAM_VERSION			165


#define BAM_SET(n)    (bamp[1+(n)/8] |= (1 << ((n) % 8)))
#define BAM_CLR(n)    (bamp[1+(n)/8] &= ~(1 << ((n) % 8)))
#define BAM_ISSET(n)  (bamp[1+(n)/8] & (1 << ((n) % 8)))

/*
 * At the beginning of each image is header that makes sure
 * only c1541 images are used. There is room for additional
 * information that is not used yet (errors etc.)
 */

#define HEADER_MAGIC_OFFSET	0

#define HEADER_MAGIC_1		'C'
#define HEADER_MAGIC_2		(1*16 + 5)
#define HEADER_MAGIC_3		(4*16 + 1)
#define HEADER_MAGIC_4		(6*16 + 4)

#define HEADER_VERSION_OFFSET	4
#define C1541_VERSION_MAJOR	1
#define C1541_VERSION_MINOR	1

#define HEADER_LENGTH		64

static BYTE bam[256];

static int sector_map[36] =
{
    0,
    21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21,
    19, 19, 19, 19, 19, 19, 19,
    18, 18, 18, 18, 18, 18,
    17, 17, 17, 17, 17
};


static char *slot_type[] =
{
    "   ", "SEQ", "PRG", "USR", "REL", "   ", "   ", "   "
};

static bufferinfo_t floppy_buffers[16];

static int floppy_find_slot;
static BYTE floppy_find_buffer[256];
static int floppy_find_current_track;
static int floppy_find_current_sector;

int     FloppyActiveFd = -1;

static char *floppy_find_name;
static int floppy_find_length;
static int floppy_find_type;

#ifndef __CEXTRACT__
#ifdef __STDC__

static void trap_1541_ready ( void );
extern int initialize_1541 ( void );
extern int open_1541 ( char *name, int length, int secondary );
extern int close_1541 ( int secondary );
extern int read_1541 ( unsigned char *data, int secondary );
extern int write_1541 ( unsigned char data, int secondary );
extern void flush_1541 ( int secondary );
static int floppy_parse_name ( char *name, int length, char *realname, int *reallength, int *readmode, int *filetype );
static int floppy_create_directory ( char *name, int length, int filetype, int secondary, unsigned char *outputptr );
static void set_find_first_slot ( char *name, int length, int type );
static unsigned char *find_next_slot ( void );
static void no_a0_pads ( unsigned char *ptr, int l );
static int floppy_read_block ( unsigned char *buf, int track, int sector );
static int floppy_write_block ( unsigned char *buf, int track, int sector );
static void floppy_error ( int code, int track, int sector );
static void remove_slot ( unsigned char *slot );
static int write_sequential_buffer ( unsigned char *slot, unsigned char *buf, int length );
static int alloc_free_sector ( int *track, int *sector );
static int allocate_sector ( int track, int sector );
static off_t offset_from_track_and_sector ( int track, int sector );
static int free_sector ( int track, int sector );
static int floppy_name_match ( unsigned char *slot, char *name, int length, int type );
static void floppy_close_all_channels ( void );
static int do_initialize ( void );
static int do_validate ( void );
static int do_format ( char *name, unsigned char *id, unsigned char *minus );
static int do_command_channel_write ( unsigned char *buf, int length );
static int do_block_command ( char command, char *buffer );
extern int check_header ( int fd, int *v1, int *v2 );
extern void attach_floppy_image ( char *name );
static int mystrncpy ( unsigned char *d, unsigned char *s, int n );
extern int write_fs ( unsigned char data, int secondary );
extern int petconv ( int c );
extern void petconvstring ( char *c );
extern int read_fs ( unsigned char *data, int secondary );
extern int open_fs ( char *name, int length, int secondary );
extern int close_fs ( int secondary );

#else /* __STDC__ */

static void trap_1541_ready (/* void */);
extern int initialize_1541 (/* void */);
extern int open_1541 (/* char *name, int length, int secondary */);
extern int close_1541 (/* int secondary */);
extern int read_1541 (/* unsigned char *data, int secondary */);
extern int write_1541 (/* unsigned char data, int secondary */);
extern void flush_1541 (/* int secondary */);
static int floppy_parse_name (/* char *name, int length, char *realname, int *reallength, int *readmode, int *filetype */);
static int floppy_create_directory (/* char *name, int length, int filetype, int secondary, unsigned char *outputptr */);
static void set_find_first_slot (/* char *name, int length, int type */);
static unsigned char *find_next_slot (/* void */);
static void no_a0_pads (/* unsigned char *ptr, int l */);
static int floppy_read_block (/* unsigned char *buf, int track, int sector */);
static int floppy_write_block (/* unsigned char *buf, int track, int sector */);
static void floppy_error (/* int code, int track, int sector */);
static void remove_slot (/* unsigned char *slot */);
static int write_sequential_buffer (/* unsigned char *slot, unsigned char *buf, int length */);
static int alloc_free_sector (/* int *track, int *sector */);
static int allocate_sector (/* int track, int sector */);
static off_t offset_from_track_and_sector (/* int track, int sector */);
static int free_sector (/* int track, int sector */);
static int floppy_name_match (/* unsigned char *slot, unsigned char *name, int length, int type */);
static void floppy_close_all_channels (/* void */);
static int do_initialize (/* void */);
static int do_validate (/* void */);
static int do_format (/* char *name, unsigned char *id, unsigned char *minus */);
static int do_command_channel_write (/* unsigned char *buf, int length */);
static int do_block_command (/* unsigned char command, unsigned char *buffer */);
extern int check_header (/* int fd, int *v1, int *v2 */);
extern void attach_floppy_image (/* char *name */);
static int mystrncpy (/* unsigned char *d, unsigned char *s, int n */);
extern int write_fs (/* unsigned char data, int secondary */);
extern int petconv (/* int c */);
extern void petconvstring (/* unsigned char *c */);
extern int read_fs (/* unsigned char *data, int secondary */);
extern int open_fs (/* char *name, int length, int secondary */);
extern int close_fs (/* int secondary */);

#endif /* __STDC__ */
#endif /* __CEXTRACT__ */


#if 0
static BYTE *find_next_slot();
static void set_find_first_slot();
static void no_a0_pads();
static int floppy_read_block();
static int floppy_write_block();
static void floppy_error();
static void remove_slot();
static int write_sequential_buffer();
static int alloc_free_sector();
static int free_sector();
static off_t offset_from_track_and_sector();
static int floppy_name_match();
static void floppy_close_all_channels();
static int do_command_channel_write();
static int mystrncpy();
static int do_block_command();
static int floppy_parse_name();
static int floppy_create_directory();
static int allocate_sector();

void    flush_1541();
int     read_1541();
int     write_1541();
int     open_1541();
int     close_1541();

int     read_fs();
int     write_fs();
int     open_fs();
int     close_fs();
#endif

#ifndef STANDALONE_1541
/*
 * Kernal loops serial-port (0xdd00) to see when serial is ready. Fake it
 */
#ifndef __STDC__
static void trap_1541_ready()
#else /* __STDC__ */
static void trap_1541_ready(void)
#endif /* __STDC__ */
{
    AC = 0x01;
    SET_SIGN(AC);
    SET_ZERO(AC);
    PC = 0xEEB2;	/* RTS */
}


static trap_t trap_1541 =
{
    "1541 ready",
    0xEEA9,
    {0xAD, 0x00, 0xDD},
    trap_1541_ready,
};
#endif

#ifndef __STDC__
int     initialize_1541()
#else /* __STDC__ */
int     initialize_1541(void)
#endif /* __STDC__ */
{
    int     i;

    for (i = 0; i < 15; i++)
	floppy_buffers[i].mode = BUFFER_NOT_IN_USE;

    floppy_buffers[15].mode = BUFFER_COMMAND_CHANNEL;
    floppy_buffers[15].buffer = (BYTE *) malloc(256);

#ifndef STANDALONE_1541
    if (attach_serial_device(8, "1541 Disk Drive", read_1541, write_1541,
	    open_1541, close_1541, flush_1541)) {
	printf("could not initialize 1541 ????\n");
	exit(0);
    }
    if (attach_serial_device(9, "1541 FS Drive", read_fs, write_fs,
	    open_fs, close_fs, NULL)) {
	printf("could not initialize 1541 ????\n");
	exit(0);
    }
    set_trap(&trap_1541);
#endif

    floppy_error(73, 0, 0);

    return 0;
}

#ifndef __STDC__
int     open_1541(name, length, secondary)
    char   *name;
    int     length;
    int     secondary;
#else /* __STDC__ */
int     open_1541(char *name, int length, int secondary)
#endif /* __STDC__ */
{
    bufferinfo_t *p = &floppy_buffers[secondary];
    char    realname[256];
    int     reallength;
    int     readmode;
    int     filetype;

    BYTE *slot;
    int     i;

    /*
     * No floppy in drive ?
     */
    if (FloppyActiveFd < 0)
	return FLOPPY_ERROR;

    /*
     * Clear error flag
     */
    floppy_error(0, 0, 0);

    /*
     * If channel is command channel, name will be used as write. Return only
     * status of last write ...
     */
    if (p -> mode == BUFFER_COMMAND_CHANNEL) {
	int     i;
	int     status = FLOPPY_COMMAND_OK;
	for (i = 0; i < length; i++)
	    status = write_1541(name[i], secondary);
	return status;
    }
    /*
     * In use ?
     */
    if (p -> mode != BUFFER_NOT_IN_USE) {
	floppy_error(70, 0, 0);
	return FLOPPY_ERROR;
    }
    /*
     * Filemode / type
     */
    if (secondary == 1)
	readmode = 0;
    else
	readmode = 1;

    filetype = 0;

    i = floppy_parse_name(name, length, realname, &reallength,
	&readmode, &filetype);

    if (i != FLOPPY_COMMAND_OK)
	return i;

    /*
     * Internal buffer ?
     */
    if (*name == '#') {
	p -> mode = BUFFER_MEMORY_BUFFER;
	p -> buffer = (BYTE *) malloc(256);
	p -> bufptr = 0;
	return FLOPPY_COMMAND_OK;
    }
    /*
     * Directory read
     */
    if (*name == '$') {
	if (!readmode)
	    return FLOPPY_ERROR;

	p -> mode = BUFFER_DIRECTORY_READ;
	p -> buffer = (BYTE *) malloc(8192);

	/*
	 * fix this someday
	 */
	if (*realname != '$')
	    p -> length = floppy_create_directory(realname, reallength, filetype,
		secondary, p -> buffer);
	else
	    p -> length = floppy_create_directory("*", 1, 0, secondary,
		p -> buffer);
	p -> bufptr = 0;

	return FLOPPY_COMMAND_OK;
    }
    /*
     * now, set filetype according secondary address, if it was not specified
     * on filename
     */
    if (!filetype)
	filetype = secondary < 2 ? 2 : 1;

    /*
     * Check that there is room on directory.
     */
    set_find_first_slot(name, length, 0);
    slot = find_next_slot();

    p -> readmode = readmode;

    /*
     * Open file for reading
     */
    if (readmode) {
	int     type;
	int     status;

	if (!slot) {
	    close_1541(secondary);
	    floppy_error(62, 0, 0);
	    return FLOPPY_ERROR;
	}
	type = slot[SLOT_TYPE_OFFSET] & 0x07;

	if (type != filetype) {
	    floppy_error(64, 0, 0);
	    return FLOPPY_ERROR;
	}
	/*
	 * Seq, Prg
	 */
	if (type == 1 || type == 2) {
	    p -> mode = BUFFER_SEQUENTIAL;
	    p -> bufptr = 2;
	    p -> buffer = (BYTE *) malloc(256);

	    status = floppy_read_block(p -> buffer, (int) slot[SLOT_FIRST_TRACK],
		(int) slot[SLOT_FIRST_SECTOR]);

	    if (status < 0) {
		close_1541(secondary);
		return FLOPPY_ERROR;
	    }
	    return FLOPPY_COMMAND_OK;
	}
	/*
	 * Unsupported filetype
	 */
	return FLOPPY_ERROR;
    }
    /*
     * Write
     */
    if (slot) {
	if (*name == '@')
	    remove_slot(slot);
	else {
	    close_1541(secondary);
	    floppy_error(63, 0, 0);
	    return FLOPPY_ERROR;
	}
    }
    /*
     * Create slot information. (XXXX: should write directoryentry to disk ?)
     */
    p -> slot = (BYTE *) malloc(32);
    memset(p -> slot, 0, 32);
    memset(p -> slot + SLOT_NAME_OFFSET, 0xa0, 16);
    memcpy(p -> slot + SLOT_NAME_OFFSET, realname, reallength);

    p -> slot[SLOT_TYPE_OFFSET] = filetype;

    p -> buffer = (BYTE *) malloc(256);
    p -> mode = BUFFER_SEQUENTIAL;
    p -> bufptr = 2;

    return FLOPPY_COMMAND_OK;
}


#ifndef __STDC__
int     close_1541(secondary)
    int     secondary;
#else /* __STDC__ */
int     close_1541(int secondary)
#endif /* __STDC__ */
{
    BYTE   *e;
    bufferinfo_t *p = &floppy_buffers[secondary];

    switch (p -> mode) {
      case BUFFER_NOT_IN_USE:
	return FLOPPY_ERROR;

      case BUFFER_MEMORY_BUFFER:
      case BUFFER_DIRECTORY_READ:
	free((char *)p -> buffer);
	p -> mode = BUFFER_NOT_IN_USE;
	p -> buffer = NULL;
	break;

      case BUFFER_SEQUENTIAL:
	if (!p -> readmode) {
	    /*
	     * Flush bytes and write slot to directory
	     */
	    write_sequential_buffer(p -> slot, p -> buffer, p -> bufptr);
	    set_find_first_slot(NULL, -1, 0);

	    e = find_next_slot();

	    if (!e) {
		p -> mode = BUFFER_NOT_IN_USE;
		free((char *)p -> buffer);
		p -> buffer = NULL;

		floppy_error(72, 0, 0);
		return FLOPPY_ERROR;
	    }
	    p -> slot[SLOT_TYPE_OFFSET] |= 0x80;
	    memcpy(&floppy_find_buffer[floppy_find_slot * 32 + 2], p -> slot + 2,
		30);

	    floppy_write_block(floppy_find_buffer,
		floppy_find_current_track,
		floppy_find_current_sector);
	    floppy_write_block(bam, 18, 0);
	}
	p -> mode = BUFFER_NOT_IN_USE;
	free((char *)p -> buffer);
	p -> buffer = NULL;
	break;

      case BUFFER_COMMAND_CHANNEL:
	floppy_close_all_channels();
	break;

      default:
	printf("unknown floppy-close-mode\n");
	exit(0);
    }

    return FLOPPY_COMMAND_OK;
}


#ifndef __STDC__
int     read_1541(data, secondary)
    BYTE   *data;
    int     secondary;
#else /* __STDC__ */
int     read_1541(unsigned char *data, int secondary)
#endif /* __STDC__ */
{
    bufferinfo_t *p = &floppy_buffers[secondary];

    switch (p -> mode) {
      case BUFFER_NOT_IN_USE:
	floppy_error(61, 0, 0);
	return FLOPPY_ERROR;

      case BUFFER_DIRECTORY_READ:
	if (p -> bufptr >= p -> length)
	    return SERIAL_EOF;
	*data = p -> buffer[p -> bufptr];
	p -> bufptr++;
	break;

      case BUFFER_MEMORY_BUFFER:
	if (p -> bufptr >= 256)
	    return SERIAL_EOF;
	*data = p -> buffer[p -> bufptr];
	p -> bufptr++;
	break;

      case BUFFER_SEQUENTIAL:
	if (!p -> readmode)
	    return FLOPPY_ERROR;

	/*
	 * Read next block if needed
	 */
	if (p -> buffer[0]) {
	    if (p -> bufptr >= 256) {
		floppy_read_block(p -> buffer,
		    (int) p -> buffer[0],
		    (int) p -> buffer[1]);
		p -> bufptr = 2;
	    }
	} else {
	    if (p -> bufptr > p -> buffer[1])
		return SERIAL_EOF;
	}

	*data = p -> buffer[p -> bufptr];
	p -> bufptr++;
	break;

      case BUFFER_COMMAND_CHANNEL:
	if (p -> bufptr > p -> length) {
	    floppy_error(0, 0, 0);
	    return SERIAL_EOF;
	}
	*data = p -> buffer[p -> bufptr];
	p -> bufptr++;
	break;

      default:
	printf("unknown buffermode on floppy-read\n");
	exit(0);
    }

    return SERIAL_OK;
}


#ifndef __STDC__
int     write_1541(data, secondary)
    BYTE    data;
    int     secondary;
#else /* __STDC__ */
int     write_1541(unsigned char data, int secondary)
#endif /* __STDC__ */
{
    int     total = 0;
    int     e;
    bufferinfo_t *p = &floppy_buffers[secondary];

    switch (p -> mode) {
      case BUFFER_NOT_IN_USE:
	floppy_error(61, 0, 0);
	return FLOPPY_ERROR;

      case BUFFER_DIRECTORY_READ:
	floppy_error(60, 0, 0);
	return FLOPPY_ERROR;

      case BUFFER_MEMORY_BUFFER:
	if (p -> bufptr >= 256)
	    return FLOPPY_ERROR;
	p -> buffer[p -> bufptr] = data;
	p -> bufptr++;
	return FLOPPY_COMMAND_OK;

      case BUFFER_SEQUENTIAL:
	if (p -> readmode)
	    return FLOPPY_ERROR;

	if (p -> bufptr >= 256) {
	    p -> bufptr = 2;
	    e = write_sequential_buffer(p -> slot, p -> buffer, WRITE_BLOCK);
	    if (e < 0)
		return FLOPPY_ERROR;
	}
	p -> buffer[p -> bufptr] = data;
	p -> bufptr++;
	break;

      case BUFFER_COMMAND_CHANNEL:
	if (p -> bufptr >= 256)
	    return FLOPPY_ERROR;
	p -> buffer[p -> bufptr] = data;
	p -> bufptr++;
	break;

      default:
	printf("Unknown write mode\n");
	exit(-1);
    }

    return total;
}


#ifndef __STDC__
void    flush_1541(secondary)
    int     secondary;
#else /* __STDC__ */
void    flush_1541(int secondary)
#endif /* __STDC__ */
{
    bufferinfo_t *p = &floppy_buffers[secondary];

    if (p -> mode != BUFFER_COMMAND_CHANNEL)
	return;

    do_command_channel_write(p -> buffer, p -> bufptr);
    p -> bufptr = 0;
    floppy_error(0, 0, 0);
}



/*
 * Parse name realname, type and read/write mode. '@' on write must
 * be checked elsewhere
 */
#ifndef __STDC__
static int floppy_parse_name(name, length, realname, reallength, readmode,
            filetype)
    char   *name;
    int     length;
    char   *realname;
    int    *reallength;
    int    *readmode;
    int    *filetype;
#else /* __STDC__ */
static int floppy_parse_name(char *name, int length, char *realname, int *reallength, int *readmode, int *filetype)
#endif /* __STDC__ */
{
    char   *p;
    char   *c;
    int     t;

    p = (char *)memchr(name, ':', length);
    if (p)
	p++;
    else
	p = name;

    if (*name == '@' && p == name)
	p++;

    t = length - (p - name);
    *reallength = 0;
    c = realname;

    while (t > 0 && *p != ',') {
	t--;
	(*reallength)++;
	*c++ = *p++;
    }

    /*
     * Change modes ?
     */
    while (t > 0) {
	t--;
	p++;

	if (t == 0) {
	    return FLOPPY_ERROR;
	}
	if (*p == 'R')
	    *readmode = 1;
	else if (*p == 'W')
	    *readmode = 0;
	else if (*p == 'S')
	    *filetype = 1;
	else if (*p == 'P')
	    *filetype = 2;
	else if (*p == 'U')
	    *filetype = 3;
	else if (*p == 'R')
	    *filetype = 4;
	else
	    return FLOPPY_ERROR;

	c = (char *)memchr(p, ',', t);

	if (c) {
	    t -= (c - p);
	    p = c;
	} else
	    t = 0;
    }

    return FLOPPY_COMMAND_OK;
}


#define SET_LO_HI(p, val) {*(p)++ = (val) & 0xff; *(p)++ = ((val)/256) & 0xff;}

/*
 * Create directory listing.
 * If filetype is 0, match for all
 */
#ifndef __STDC__
static int floppy_create_directory(name, length, filetype, secondary,
            outputptr)
    char   *name;
    int     length;
    int     filetype;
    int     secondary;
     BYTE   *outputptr;
#else /* __STDC__ */
static int floppy_create_directory(char *name, int length, int filetype, int secondary, unsigned char *outputptr)
#endif /* __STDC__ */
{
    BYTE   *l;
    BYTE *p;
    BYTE   *tmpsl;
    BYTE   *origptr = outputptr;
    int     t;
    int     bf;

    int     addr;


    set_find_first_slot(name, length, filetype);

    l = outputptr;

    /*
     * Start-address
     */
    addr = 0x401;
    SET_LO_HI(l, addr);

    /*
     * Set next line offset later
     */
    l += 2;

    /*
     * Line number 0
     */
    SET_LO_HI(l, 0);

    /*
     * Reverse on
     */
    *l++ = (char) 0x12;

    *l++ = '\"';
    memcpy(l, &bam[BAM_DISK_NAME], 16);
    no_a0_pads(l, 16);
    l += 16;
    *l++ = '\"';
    *l++ = ' ';
    memcpy(l, &bam[BAM_DISK_ID], 5);
    no_a0_pads(l, 5);
    l += 5;

    *l++ = 0;

    /*
     * Address of next line
     */
    addr += ((l - outputptr) - 2);

    outputptr[2] = addr & 0xff;
    outputptr[3] = (addr / 256) & 0xff;

    outputptr = l;

    /*
     * Now, list files
     */
    while ((p = find_next_slot())) {
	if (p[SLOT_TYPE_OFFSET]) {

	    l += 2;

	    /*
	     * Length and spaces
	     */
	    t = p[SLOT_NR_BLOCKS] + p[SLOT_NR_BLOCKS + 1] * 256;

	    SET_LO_HI(l, t);

	    if (t < 10)
		*l++ = ' ';
	    if (t < 100)
		*l++ = ' ';

	    /*
	     * name
	     */
	    *l++ = ' ';
	    *l++ = '\"';
	    memcpy(l, &p[SLOT_NAME_OFFSET], 16);

	    for (t = 0; (t < 16) && (p[SLOT_NAME_OFFSET + t] != 0xa0);)
		t++;

	    no_a0_pads(l, 16);

	    l[16] = ' ';
	    l[t] = '\"';
	    l += 17;

	    /*
	     * type
	     */
	    *l++ = p[SLOT_TYPE_OFFSET] & 0x80 ? ' ' : '*';

	    tmpsl = (BYTE *)slot_type[p[SLOT_TYPE_OFFSET] & 0x07];

	    memcpy(l, tmpsl, 3);
	    l += 3;

	    /*
	     * End
	     */
	    *l++ = ' ';
	    *l++ = ' ';
	    *l++ = (char) 0;

	    /*
	     * New address
	     */
	    addr += l - outputptr;

	    outputptr[0] = addr & 0xff;
	    outputptr[1] = (addr / 256) & 0xff;

	    outputptr = l;
	}
    }

    /*
     * calculate blocks free
     */
    bf = 0;

    for (t = 1; t <= 35; t++) {
	if (t != 18)
	    bf += bam[BAM_BIT_MAP + 4 * (t - 1)];
    }

    *l++ = 0;
    *l++ = 0;
    SET_LO_HI(l, bf);
    memcpy(l, "BLOCKS FREE.", 12);
    l += 12;
    *l++ = (char) 0;

    addr += l - outputptr;

    outputptr[0] = addr & 0xff;
    outputptr[1] = (addr / 256) & 0xff;

    /*
     * end
     */
    *l++ = (char) 0;
    *l++ = (char) 0;
    *l++ = (char) 0;

    return l - origptr;
}



/*
 * Initialize slot find
 */
#ifndef __STDC__
static void set_find_first_slot(name, length, type)
    char   *name;
    int     length;
    int     type;
#else /* __STDC__ */
static void set_find_first_slot(char *name, int length, int type)
#endif /* __STDC__ */
{
    floppy_find_name = name;
    floppy_find_length = length;
    floppy_find_type = type;

    /*
     * Make sure find_next_slot() loads first directoryblock ...
     */
    floppy_find_slot = 8;
    floppy_read_block(floppy_find_buffer, 18, 0);
}

#ifndef __STDC__
static BYTE *find_next_slot()
#else /* __STDC__ */
static BYTE *find_next_slot(void)
#endif /* __STDC__ */
{
    static BYTE return_slot[32];
    int     status;

    floppy_find_slot++;

    /*
     * Loop all directoryblocks staring from track 18, sector 1
     */
    do {
	/*
	 * Load next(first) directory block ?
	 */
	if (floppy_find_slot >= 8) {
	    if (!(*floppy_find_buffer))
		return NULL;

	    floppy_find_slot = 0;

	    floppy_find_current_track = (int) floppy_find_buffer[0];
	    floppy_find_current_sector = (int) floppy_find_buffer[1];

	    status = floppy_read_block(floppy_find_buffer,
		floppy_find_current_track,
		floppy_find_current_sector);
	}
	while (floppy_find_slot < 8) {
	    if (floppy_name_match(&floppy_find_buffer[floppy_find_slot * 32],
		    floppy_find_name, floppy_find_length,
		    floppy_find_type)) {
		memcpy(return_slot,
		    &floppy_find_buffer[floppy_find_slot * 32], 32);
		return return_slot;
	    }
	    floppy_find_slot++;
	}
    } while (*floppy_find_buffer);

    /*
     * If length < 0, create new directory-entry if possible
     */
    if (floppy_find_length < 0) {
	int     s;

	for (s = 1; s < sector_map[18]; s++) {
	    if (allocate_sector(18, s)) {
		floppy_find_buffer[0] = 18;
		floppy_find_buffer[1] = s;
		floppy_write_block(floppy_find_buffer,
		    floppy_find_current_track,
		    floppy_find_current_sector);
		floppy_find_slot = 0;
		memset(floppy_find_buffer, 0, 256);
		floppy_find_buffer[1] = 0xff;
		floppy_find_current_sector = s;
		return floppy_find_buffer;
	    }
	}

    }
    return NULL;
}


#ifndef __STDC__
static void no_a0_pads(ptr, l)
    BYTE *ptr;
    int     l;
#else /* __STDC__ */
static void no_a0_pads(unsigned char *ptr, int l)
#endif /* __STDC__ */
{
    while (l--) {
	if (*ptr == 0xa0)
	    *ptr = 0x20;

	ptr++;
    }
}



/*
 * Read one block
 */
#ifndef __STDC__
static int floppy_read_block(buf, track, sector)
    BYTE   *buf;
    int     track;
    int     sector;
#else /* __STDC__ */
static int floppy_read_block(unsigned char *buf, int track, int sector)
#endif /* __STDC__ */
{
    off_t   offset;

    offset = offset_from_track_and_sector(track, sector);

    if (offset < 0)
	return -1;

    offset += HEADER_LENGTH;

    lseek(FloppyActiveFd, offset, SEEK_SET);

    if (read(FloppyActiveFd, (char *)buf, 256) < 256) {
	printf("floppy read error\n");
	exit(-1);
    }
    return 0;
}



/*
 * Write one block
 */
#ifndef __STDC__
static int floppy_write_block(buf, track, sector)
    BYTE   *buf;
    int     track;
    int     sector;
#else /* __STDC__ */
static int floppy_write_block(unsigned char *buf, int track, int sector)
#endif /* __STDC__ */
{
    off_t   offset;

    offset = offset_from_track_and_sector(track, sector);

    if (offset < 0)
	return -1;

    offset += HEADER_LENGTH;

    lseek(FloppyActiveFd, offset, SEEK_SET);

    if (write(FloppyActiveFd, (char *)buf, 256) < 256) {
	printf("floppy write error\n");
	exit(-1);
    }
    return 0;
}


/*
 * Error messages
 */
typedef struct errortext_s {
    int     nr;
    char   *text;
}       errortext_t;

static errortext_t floppy_error_messages[] =
{
    {0, "OK"},
    {1, "FILES SCRATCHED"},
    {30, "SYNTAX ERROR"},
    {60, "WRITE FILE OPEN"},
    {61, "FILE NOT OPEN"},
    {62, "FILE NOT FOUND"},
    {63, "FILE EXISTS"},
    {64, "FILE TYPE MISMATCH"},
    {70, "NO CHANNEL"},
    {72, "DISK FULL"},
    {73, "TVR 1541 EMULATOR V1.0"},
    {-1, 0},
};


/*
 * Should set values to somewhere so that they could be read from
 * command channel
 */
#ifndef __STDC__
static void floppy_error(code, track, sector)
    int     code;
    int     track;
    int     sector;
#else /* __STDC__ */
static void floppy_error(int code, int track, int sector)
#endif /* __STDC__ */
{
    char   *message;
    bufferinfo_t *p = &floppy_buffers[15];

    errortext_t *e = &floppy_error_messages[0];

    while (e -> nr >= 0 && e -> nr != code)
	e++;

    if (e -> nr >= 0)
	message = e -> text;
    else
	message = "UNKNOWN ERROR NUMBER";

    sprintf((char *)p -> buffer, "%02d, %s,%d,%d", code, message, track, sector);
    p -> length = strlen((char *)p -> buffer);
    p -> bufptr = 0;
}


#ifndef __STDC__
static void remove_slot(slot)
    BYTE *slot;
#else /* __STDC__ */
static void remove_slot(unsigned char *slot)
#endif /* __STDC__ */
{
    BYTE    buf[256];
    int     tmp;
    int     t, s;

    /*
     * Find slot
     */
    for (tmp = 0; (tmp < 16) && slot[SLOT_NAME_OFFSET + tmp] != 0xa0; tmp++);

    set_find_first_slot((char *)&slot[SLOT_NAME_OFFSET], tmp,
	slot[SLOT_TYPE_OFFSET] & 0x07);

    /*
     * If slot slot found, remove
     */
    if (find_next_slot()) {
	/*
	 * Free all buffers from bam
	 */
	t = (int) floppy_find_buffer[floppy_find_slot * 32 + SLOT_FIRST_TRACK];
	s = (int) floppy_find_buffer[floppy_find_slot * 32 + SLOT_FIRST_SECTOR];

	while (t) {
	    free_sector(t, s);
	    floppy_read_block(buf, t, s);
	    t = (int) buf[0];
	    s = (int) buf[1];
	}

	/*
	 * Update bam
	 */
	floppy_write_block(bam, 18, 0);

	/*
	 * Update directory entry
	 */
	floppy_find_buffer[floppy_find_slot * 32 + SLOT_TYPE_OFFSET] = 0;
	floppy_write_block(floppy_find_buffer,
	    floppy_find_current_track,
	    floppy_find_current_sector);
    }
}



#ifndef __STDC__
static int write_sequential_buffer(slot, buf, length)
    BYTE *slot;
    BYTE   *buf;
    int     length;
#else /* __STDC__ */
static int write_sequential_buffer(unsigned char *slot, unsigned char *buf, int length)
#endif /* __STDC__ */
{
    static int t = 0, s;
    int     t_new, s_new;
    int     e;

    /*
     * First block of a file ?
     */
    if (t == 0) {
	e = alloc_free_sector(&t, &s);
	if (e < 0) {
	    floppy_error(72, 0, 0);
	    return -1;
	}
	slot[SLOT_FIRST_TRACK] = t;
	slot[SLOT_FIRST_SECTOR] = s;
    }
    if (length == WRITE_BLOCK) {
	/*
	 * Write current sector and allocate next
	 */
	e = alloc_free_sector(&t_new, &s_new);
	if (e < 0) {
	    floppy_error(72, 0, 0);
	    return -1;
	}
	buf[0] = t_new;
	buf[1] = s_new;

	floppy_write_block(buf, t, s);

	t = t_new;
	s = s_new;

	if (!(++slot[SLOT_NR_BLOCKS]))
	    ++slot[SLOT_NR_BLOCKS + 1];

    } else {
	/*
	 * Write last block
	 */
	buf[0] = 0;
	buf[1] = length - 1;

	floppy_write_block(buf, t, s);

	t = 0;

	if (!(++slot[SLOT_NR_BLOCKS]))
	    ++slot[SLOT_NR_BLOCKS + 1];
    }
    return 0;
}



#ifndef __STDC__
static int alloc_free_sector(track, sector)
    int    *track;
    int    *sector;
#else /* __STDC__ */
static int alloc_free_sector(int *track, int *sector)
#endif /* __STDC__ */
{
    int     t, s;

    for (t = 17; t >= 1; t--) {
	for (s = 0; s < sector_map[t]; s++) {
	    if (allocate_sector(t, s)) {
		*track = t;
		*sector = s;
		return 0;
	    }
	}
    }

    for (t = 19; t <= 35; t++) {
	for (s = 0; s < sector_map[t]; s++) {
	    if (allocate_sector(t, s)) {
		*track = t;
		*sector = s;
		return 0;
	    }
	}
    }
    return -1;
}


#ifndef __STDC__
static int allocate_sector(track, sector)
    int     track, sector;
#else /* __STDC__ */
static int allocate_sector(int track, int sector)
#endif /* __STDC__ */
{
    /*
     * Macros use this
     */
    BYTE   *bamp;

    bamp = &bam[BAM_BIT_MAP + 4 * (track - 1)];

    if (BAM_ISSET(sector)) {
	(*bamp)--;
	BAM_CLR(sector);

	return 1;
    }
    return 0;
}


/*
 * Return block offset on 'disk file' according to track and sector
 */
#ifndef __STDC__
static off_t offset_from_track_and_sector(track, sector)
    int     track;
    int     sector;
#else /* __STDC__ */
static off_t offset_from_track_and_sector(int track, int sector)
#endif /* __STDC__ */
{
    int     sectors, i;

    if ((track < 1) || (track > 35))
	return (-1);

    for (sectors = 0, i = 1; i < track; i++) {
	sectors += sector_map[i];
    }

    if ((sector < 0) || (sector >= sector_map[track]))
	return (-1);

    sectors += sector;

    return 256 * sectors;
}



#ifndef __STDC__
static int free_sector(track, sector)
    int     track, sector;
#else /* __STDC__ */
static int free_sector(int track, int sector)
#endif /* __STDC__ */
{
    BYTE   *bamp;

    bamp = &bam[BAM_BIT_MAP + 4 * (track - 1)];

    if (!(BAM_ISSET(sector))) {
	BAM_SET(sector);
	(*bamp)++;
	return 1;
    }
    return 0;
}


/*
 * If type is 0, match for next slot with same name
 * If type is nonzero, match next that has the same id and name
 */
#ifndef __STDC__
static int floppy_name_match(slot, name, length, type)
    BYTE *slot;
     char *name;
    int     length;
    int     type;
#else /* __STDC__ */
static int floppy_name_match(unsigned char *slot, char *name, int length, int type)
#endif /* __STDC__ */
{
    int     i;
    /* int comma = 0; */

    if (length < 0) {
	if (slot[SLOT_TYPE_OFFSET])
	    return 0;
	else
	    return 1;
    }
    if (!slot[SLOT_TYPE_OFFSET])
	return 0;

    if (length >= 16)
	length = 16;

    for (i = 0; i < length; i++) {
	switch (name[i]) {
	  case '?':
	    /*
	     * Match any character
	     */
	    break;

	  case '*':
	    /*
	     * Make a match
	     */
	    i = 16;
	    break;

	  default:
	    if (name[i] != slot[i + SLOT_NAME_OFFSET])
		return 0;
	}
    }

    /*
     * Got name match ?
     */
    if (i < 16 && slot[SLOT_NAME_OFFSET + i] != 0xa0)
	return 0;

    if (type && type != (slot[SLOT_TYPE_OFFSET] & 0x07))
	return 0;

    return 1;
}




/*
 * Close all channels. This happens on 'I' -command and on command-
 * channel close
 */
#ifndef __STDC__
static void floppy_close_all_channels()
#else /* __STDC__ */
static void floppy_close_all_channels(void)
#endif /* __STDC__ */
{
    int     i;
    bufferinfo_t *p;

    for (i = 0; i <= 15; i++) {
	p = &floppy_buffers[i];

	if (p -> mode != BUFFER_NOT_IN_USE && p -> mode != BUFFER_COMMAND_CHANNEL)
	    close_1541(i);
    }
}


#ifndef __STDC__
static int do_initialize()
#else /* __STDC__ */
static int do_initialize(void)
#endif /* __STDC__ */
{

    floppy_close_all_channels();
    return FLOPPY_COMMAND_OK;
}

#ifndef __STDC__
static int do_validate()
#else /* __STDC__ */
static int do_validate(void)
#endif /* __STDC__ */
{
    int     t, s;
    BYTE *b;
    BYTE    tmp[256];

    do_initialize();

    b = (BYTE *) bam + BAM_BIT_MAP;

    for (t = 1; t <= 35; t++) {
	*b++ = 0;
	*b++ = 0;
	*b++ = 0;
	*b++ = 0;

	for (s = 0; s < sector_map[t]; s++)
	    free_sector(t, s);
    }

    set_find_first_slot("*", 1, 0);

    while ((b = (BYTE *) find_next_slot())) {
	if (b[SLOT_TYPE_OFFSET]) {
	    floppy_find_buffer[floppy_find_slot * 32 + SLOT_TYPE_OFFSET] |= 0x80;
	    floppy_write_block(floppy_find_buffer, floppy_find_current_track,
		floppy_find_current_sector);
	    t = (int) b[SLOT_FIRST_TRACK];
	    s = (int) b[SLOT_FIRST_SECTOR];

	    while (t) {
		allocate_sector(t, s);
		floppy_read_block(tmp, t, s);
		t = (int) tmp[0];
		s = (int) tmp[1];
	    }
	}
    }

    floppy_write_block(bam, 18, 0);
    return FLOPPY_COMMAND_OK;
}


#ifndef __STDC__
static int do_format(name, id, minus)
    char   *name;
    BYTE   *id;
    BYTE   *minus;
#else /* __STDC__ */
static int do_format(char *name, unsigned char *id, unsigned char *minus)
#endif /* __STDC__ */
{
    BYTE    tmp[256];
    int     status;

    BYTE    null = 0;

    if (!name) {
	floppy_error(30, 0, 0);
	return FLOPPY_ERROR;
    }
    /*
     * If id, skip comma
     */
    if (id)
	*id++ = 0;
    else
	id = &null;

    /*
     * Make dir-entry
     */
    memset(tmp, 0, 256);
    tmp[1] = 255;
    floppy_write_block(tmp, 18, 1);

    /*
     * Make bam
     */
    memset(bam, 0, 256);
    bam[0] = 18;
    bam[1] = 1;
    bam[2] = 65;
    memset(bam + BAM_DISK_NAME, 0xa0, 16);
    mystrncpy(bam + BAM_DISK_NAME, (BYTE *)name + 1, 16);

    memset(bam + BAM_DISK_ID, 0xa0, 5);
    mystrncpy(bam + BAM_DISK_ID, id, 2);
    bam[BAM_VERSION] = 50;
    bam[BAM_VERSION + 1] = 65;
    status = do_validate();
    floppy_write_block(bam, 18, 0);
    return status;
}


#ifndef __STDC__
static int do_command_channel_write(buf, length)
    BYTE   *buf;
    int     length;
#else /* __STDC__ */
static int do_command_channel_write(unsigned char *buf, int length)
#endif /* __STDC__ */
{
    int     status = FLOPPY_ERROR;
    BYTE   *p;
    char   *name;
    BYTE   *minus;
    BYTE   *id;

    p = (BYTE *) malloc(length + 1);
    memcpy(p, buf, length);
    p[length] = 0;

    name = (char *)memchr(p, ':', length);
    id = (BYTE *)memchr(p, ',', length);
    minus = (BYTE *)memchr(p, '-', length);

    floppy_error(0, 0, 0);

    switch (p[0]) {
      case 'N':
	status = do_format(name, id, minus);
	break;
      case 'C':
	break;

      case 'R':
	break;

      case 'S':
	break;

      case 'I':
	status = do_initialize();
	break;

      case 'V':
	status = do_validate();
	break;

      case 'B':
	if (!minus || !name)
	    floppy_error(30, 0, 0);
	else
	    status = do_block_command(minus[1], name + 1);
	break;

      case 'U':
	switch (p[1]) {
	  case '1':
	  case 'A':
	    if (name)
		status = do_block_command('R', name + 1);
	    break;

	  case '2':
	  case 'B':
	    if (name)
		status = do_block_command('W', name + 1);
	    break;
	}
	break;

    }

    free((char *)p);
    return status;
}



#ifndef __STDC__
static int do_block_command(command, buffer)
    char    command;
    char   *buffer;
#else /* __STDC__ */
static int do_block_command(char command, char *buffer)
#endif /* __STDC__ */
{
    int     l;

    int     channel = 0;
    int     drive = 0;
    int     track = 0;
    int     sector = 0;
    int     position = 0;

    switch (command) {
      case 'R':
      case 'W':
	l = sscanf(buffer, "%d %d %d %d", &channel, &drive,
	    &track, &sector);
	if (l < 4) {
	    l = sscanf(buffer, "%d,%d,%d,%d", &channel, &drive,
		&track, &sector);
	    if (l < 4) {
		floppy_error(30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}
	if (floppy_buffers[channel].mode != BUFFER_MEMORY_BUFFER) {
	    floppy_error(30, 0, 0);
	    return FLOPPY_ERROR;
	}
	if (command == 'W')
	    floppy_write_block(floppy_buffers[channel].buffer,
		track, sector);
	else
	    floppy_read_block(floppy_buffers[channel].buffer,
		track, sector);

	floppy_buffers[channel].bufptr = 0;
	break;

      case 'A':
      case 'F':
	l = sscanf(buffer, "%d %d %d", &drive, &track, &sector);

	if (l < 3) {
	    l = sscanf(buffer, "%d,%d,%d", &drive, &track, &sector);

	    if (l < 3) {
		floppy_error(30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}
	if (command == 'A')
	    allocate_sector(track, sector);
	else
	    free_sector(track, sector);
	break;

      case 'P':
	l = sscanf(buffer, "%d %d", &channel, &position);

	if (l < 2) {
	    l = sscanf(buffer, "%d,%d", &channel, &position);

	    if (l < 2) {
		floppy_error(30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}
	floppy_buffers[channel].bufptr = position;
	break;

      default:
	floppy_error(30, 0, 0);
	return FLOPPY_ERROR;
    }
    return FLOPPY_COMMAND_OK;
}

#ifndef __STDC__
int     check_header(fd, v1, v2)
    int     fd;
    int    *v1;
    int    *v2;
#else /* __STDC__ */
int     check_header(int fd, int *v1, int *v2)
#endif /* __STDC__ */
{
    BYTE    header[HEADER_LENGTH];

    if (read(fd, (char *)header, sizeof (header)) != sizeof (header)) {
	printf("cannot read image header\n");
	return 1;
    }
    if (header[HEADER_MAGIC_OFFSET + 0] != HEADER_MAGIC_1 ||
	header[HEADER_MAGIC_OFFSET + 1] != HEADER_MAGIC_2 ||
	header[HEADER_MAGIC_OFFSET + 2] != HEADER_MAGIC_3 ||
	header[HEADER_MAGIC_OFFSET + 3] != HEADER_MAGIC_4) {
	return 1;
    }
    *v1 = header[HEADER_VERSION_OFFSET + 0];
    *v2 = header[HEADER_VERSION_OFFSET + 1];

    return 0;
}


#ifndef __STDC__
void    attach_floppy_image(name)
    char   *name;
#else /* __STDC__ */
void    attach_floppy_image(char *name)
#endif /* __STDC__ */
{
    int     v1, v2;

    floppy_close_all_channels();

    if (FloppyActiveFd >= 0)
	close(FloppyActiveFd);

    FloppyActiveFd = -1;

    if (!name) {
	printf("detaching floppyimage\n");
	return;
    }

#ifdef C64_MSDOS
    FloppyActiveFd = open(name, O_RDWR | O_BINARY);
#else
    FloppyActiveFd = open(name, O_RDWR);
#endif

    if (FloppyActiveFd >= 0) {
	if (check_header(FloppyActiveFd, &v1, &v2)) {
	    printf("file %s does not have valid C1541 header\n", name);
	    close(FloppyActiveFd);
	    FloppyActiveFd = -1;
	    return;
	}
	if (v1 > C1541_VERSION_MAJOR ||
	    (v1 == C1541_VERSION_MAJOR && v2 > C1541_VERSION_MINOR)) {
	    printf("c1541 image file %s (version %d.%02d) newer that emulator version %d.%02d\n",
		name, v1, v2, C1541_VERSION_MAJOR, C1541_VERSION_MINOR);
	    close(FloppyActiveFd);
	    FloppyActiveFd = -1;
	    return;
	}

#ifndef STANDALONE_1541
	printf("c1541 image file %s (version %d.%02d) attached\n", name,
	    v1, v2);
#endif

	floppy_read_block(bam, 18, 0);
    }
}



#ifndef __STDC__
static int mystrncpy(d, s, n)
    BYTE   *d;
    BYTE   *s;
    int     n;
#else /* __STDC__ */
static int mystrncpy(unsigned char *d, unsigned char *s, int n)
#endif /* __STDC__ */
{
    while (n-- && *s)
	*d++ = *s++;
    return (n);
}



/*
 * File system interface
 */

enum fsmode {Write, Read, Directory };

struct fs_buffer_info {
    FILE   *fd;
    DIR    *dp;
    enum fsmode mode;
    char    dir[MAXPATHLEN];
    BYTE    buf[MAXPATHLEN + 5];
    int     buflen;
    BYTE   *bufp;
    int     eof;
}       fs_info[16];

#ifndef __STDC__
int     write_fs(data, secondary)
     BYTE    data;
    int     secondary;
#else /* __STDC__ */
int     write_fs(unsigned char data, int secondary)
#endif /* __STDC__ */
{
    if (fs_info[secondary].mode != Write)
	return FLOPPY_ERROR;

    if (fs_info[secondary].fd) {
	fputc(data, fs_info[secondary].fd);
	return FLOPPY_COMMAND_OK;
    };

    return FLOPPY_ERROR;
}


#ifndef __STDC__
int     petconv(c)
     int c;
#else /* __STDC__ */
int     petconv(int c)
#endif /* __STDC__ */
{
    switch (c & 0xe0) {
	case 0x40:
	case 0x60:
	return (c ^ 0x20);
    }
    return (c);
}

#ifndef __STDC__
void    petconvstring(c)
     char *c;
#else /* __STDC__ */
void    petconvstring(char *c)
#endif /* __STDC__ */
{
    int i;
    for (i = strlen(c) - 1; i >= 0; i--)
	c[i] = petconv(c[i]);
}

#ifndef __STDC__
int     read_fs(data, secondary)
    BYTE   *data;
    int     secondary;
#else /* __STDC__ */
int     read_fs(unsigned char *data, int secondary)
#endif /* __STDC__ */
{
    int     i;
    unsigned short blocks;
    struct dirent *dirp;
    struct stat statbuf;
    struct fs_buffer_info *info = &fs_info[secondary];

    switch (info -> mode) {
      case Write:
	return FLOPPY_ERROR;

      case Read:
	if (info -> fd) {
	    i = fgetc(info -> fd);
	    if (ferror(info -> fd))
		return FLOPPY_ERROR;
	    if (feof(info -> fd))
		return SERIAL_EOF;
	    *data = i;
	    return SERIAL_OK;
	}
	break;

      case Directory:
	if (info -> dp) {
	    if (info -> buflen <= 0) {
		char    buf[MAXPATHLEN];
		if (info -> eof)
		    return SERIAL_EOF;

		if ((dirp = readdir(info -> dp))) {
		    BYTE  *p = info -> buf;
		    int     j;

		    strcpy(buf, info -> dir);
		    strcat(buf, "/");
		    strcat(buf, dirp -> d_name);
		    if (stat(buf, &statbuf) < 0)
			return FLOPPY_ERROR;

		    *p++ = 1;
		    *p++ = 4;
		    blocks = (unsigned short) (statbuf.st_size / 256);
		    *p++ = LOWER(blocks);
		    *p++ = UPPER(blocks);
		    for (j = 0; blocks > 9; j++)
			blocks /= 10;
		    i = 4 - j;
		    while (i-- > 0)
			*p++ = ' ';
		    *p++ = '"';

		    for (i = 0; i < dirp -> d_namlen; i++)
			*p++ = petconv(dirp -> d_name[i]);

		    *p++ = '"';
		    *p++ = '\0';

		    info -> buflen = (int) p - (int) info -> buf;

		} else {
		    info -> buf[0] = 0;
		    info -> buf[1] = 0;
		    info -> buf[2] = 0;
		    info -> buflen = 3;
		    info -> eof++;
		}
		info -> bufp = info -> buf;
	    }
	    *data = *info -> bufp++;
	    info -> buflen--;
	    return SERIAL_OK;
	}
	break;
    }

    return FLOPPY_ERROR;
}



#ifndef __STDC__
int     open_fs(name, length, secondary)
    char   *name;
    int     length;
    int     secondary;
#else /* __STDC__ */
int     open_fs(char *name, int length, int secondary)
#endif /* __STDC__ */
{
    FILE   *fd;
    DIR    *dp;
    char    tmp[256], *p;

    if (fs_info[secondary].fd)
	return FLOPPY_ERROR;

    /*
     * Support only load / save
     */
    if (secondary < 0 || secondary >= 2)
	return FLOPPY_ERROR;

    memcpy(tmp, name, length);
    tmp[length] = 0;
    petconvstring(tmp);
    
    if (*tmp == '$') {
	if (secondary != 0)
	    return FLOPPY_ERROR;

	if (tmp[1]) {
	    if (!(dp = opendir((char *)tmp + 1))) {
		for (p = tmp; *p; p++)
		    if (isupper(*p))
			*p = tolower(*p);
		if (!(dp = opendir((char *)tmp + 1)))
		    return FLOPPY_ERROR;
	    }
	    strcpy(fs_info[secondary].dir, tmp + 1);
	} else {
	    if (!(dp = opendir(".")))
		return FLOPPY_ERROR;
	    strcpy(fs_info[secondary].dir, ".");
	}

	fs_info[secondary].mode = Directory;
	fs_info[secondary].dp = dp;
	fs_info[secondary].eof = 0;
	fs_info[secondary].buflen = 2;
	fs_info[secondary].bufp = fs_info[secondary].buf;
	fs_info[secondary].buf[0] = 0;
	fs_info[secondary].buf[1] = 4;

    } else {
	fs_info[secondary].mode = (secondary == 1 ? Write : Read);
	fd = fopen(tmp, fs_info[secondary].mode ? "r" : "w+");

	if (!fd) {
	    for (p = tmp; *p; p++)
		if (isupper(*p))
		    *p = tolower(*p);

	    fd = fopen(tmp, fs_info[secondary].mode ? "r" : "w+");

	    if (!fd)
		return FLOPPY_ERROR;
	}
	fs_info[secondary].fd = fd;
    }
    return FLOPPY_COMMAND_OK;
}



#ifndef __STDC__
int     close_fs(secondary)
    int     secondary;
#else /* __STDC__ */
int     close_fs(int secondary)
#endif /* __STDC__ */
{
    switch (fs_info[secondary].mode) {
      case Write:
      case Read:
	if (!fs_info[secondary].fd)
	    return FLOPPY_ERROR;

	fclose(fs_info[secondary].fd);
	fs_info[secondary].fd = NULL;
	break;

      case Directory:
	if (!fs_info[secondary].dp)
	    return FLOPPY_ERROR;

	closedir(fs_info[secondary].dp);
	fs_info[secondary].dp = NULL;
	break;
    }

    return FLOPPY_COMMAND_OK;
}


#ifdef STANDALONE_1541
void    create_image(fd)
    int     fd;
{
    BYTE    header[HEADER_LENGTH];
    BYTE    block[256];
    int     i;

    memset(header, 0, sizeof (header));
    memset(block, 0, sizeof (block));

    header[HEADER_MAGIC_OFFSET + 0] = HEADER_MAGIC_1;
    header[HEADER_MAGIC_OFFSET + 1] = HEADER_MAGIC_2;
    header[HEADER_MAGIC_OFFSET + 2] = HEADER_MAGIC_3;
    header[HEADER_MAGIC_OFFSET + 3] = HEADER_MAGIC_4;

    header[HEADER_VERSION_OFFSET + 0] = C1541_VERSION_MAJOR;
    header[HEADER_VERSION_OFFSET + 1] = C1541_VERSION_MINOR;

    printf("creating header\n");

    if (write(fd, (char *)header, sizeof (header)) != sizeof (header)) {
	printf("cannot write header\n");
	exit(1);
    }
    printf("writing blocks\n");

    for (i = 0; i < 683; i++) {
	if (write(fd, (char *)block, sizeof (block)) != sizeof (block)) {
	    printf("cannot write block %d\n", i);
	    exit(1);
	}
    }
}

void    open_image(name, create)
    char   *name;
    int     create;
{
    int     fd;

    if (create) {
	if ((fd = open(name, O_RDWR | O_CREAT, 0666)) < 0) {
	    printf("could not create image %s\n", name);
	    exit(1);
	}
	create_image(fd);
	close(fd);
    }
    attach_floppy_image(name);

    if (FloppyActiveFd < 0)
	exit(1);
}

void    print_listing()
{
    BYTE   *p;
    int     length;
    /*
     * int   i; int   t;
     */
    int     address;

    p = floppy_buffers[2].buffer;
    length = floppy_buffers[2].length;

    p += 2;

    while (1) {
	address = (*p) + (*(p + 1)) * 256;
	p += 2;
	if (!address)
	    break;
	printf("%d ", (*p) + (*(p + 1)) * 256);
	p += 2;
	while (*p) {
	    if (isprint(*p))
		printf("%c", *p);
	    p++;
	}
	p++;
	printf("\n");
    }
}

void    usage()
{
    printf("%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n",
	"usage: following commands are supported on c1541",
	"    c1541 -f image name,id           Format image. Create if needed",
	"    c1541 -d image files             Delete files from image",
	"    c1541 -l image                   List files on image",
	"    c1541 -c image dump              Make image out of raw blockdump",
	"                                          (683 256byte blocks)",
	"    c1541 -v image                   Validate disk",
	"    c1541 -i image                   Show image and 1541 version",
	"    c1541 -r image 1541name fsname   Read 1541name on image to file fsname",
	"    c1541 -w image fsname 1541name   Write file fsname to image");
    exit(1);
}


/*
 * This is ugly ... (and a quick hack)
 */
int     main(argc, argv)
    int     argc;
    char   *argv[];
{
    char    tmp[256];
    char    newname[256];
    int     fd, v1, v2;
    FILE   *f;

    initialize_1541();

    if (argc < 3 || argv[1][0] != '-')
	usage();

    switch (argv[1][1]) {
      case 'f':
	if (argc != 4)
	    usage();
	if (!memchr(argv[3], ',', strlen(argv[3]))) {
	    printf("There must be ID on a name\n");
	    usage();
	}
	open_image(argv[2], 1);
	printf("formatting image\n");
	strcpy(newname, argv[3]);
	petconvstring(newname);
	strcpy(tmp, "N:");
	strcat(tmp, newname);
	do_command_channel_write((BYTE *)tmp, strlen(tmp));
	break;

      case 'd':
	printf("not supported yet\n");
	exit(1);

      case 'r':
	if (argc != 5)
	    usage();
	open_image(argv[2], 0);
	strcpy((char *)newname, argv[3]);
	petconvstring(newname);
	if (open_1541(newname, strlen(newname), 0)) {
	    printf("cannot open %s on image %s\n", argv[3], argv[2]);
	    exit(1);
	}
	if (!(f = fopen(argv[4], "w"))) {
	    printf("cannot open %s for writing\n", argv[4]);
	    exit(1);
	}
	printf("reading file from image\n");

	{
	    char    c;
	    while (!read_1541((BYTE *)&c, 0))
		fputc(c, f);
	}
	fclose(f);
	close_1541(0);
	break;

      case 'w':
	if (argc != 5)
	    usage();
	open_image(argv[2], 0);
	if (!(f = fopen(argv[3], "r"))) {
	    printf("cannot open %s for reading\n", argv[3]);
	    exit(1);
	}
	strcpy((char *)newname, argv[4]);
	petconvstring(newname);
	if (open_1541(newname, strlen(newname), 1)) {
	    printf("cannot open %s for writing on image %s\n",
		argv[4], argv[2]);
	    exit(1);
	}
	printf("writing file to image\n");

	{
	    int     c;
	    while (EOF != (c = fgetc(f))) {
		if (write_1541(c, 1)) {
		    printf("No space on image ?\n");
		    break;
		}
	    }
	}
	fclose(f);
	close_1541(1);
	break;

      case 'l':
	if (argc != 3)
	    usage();
	open_image(argv[2], 0);
	open_1541("$", 1, 2);
	print_listing();
	close_1541(2);
	break;

      case 'c':
	if (argc != 4)
	    usage();
	open_image(argv[2], 1);
	{
	    int     i;

	    fd = open(argv[3], O_RDONLY);
	    if (fd < 0) {
		printf("cannot open %s\n", argv[3]);
		exit(1);
	    }
	    printf("copying blocks to image\n");

	    lseek(FloppyActiveFd, HEADER_LENGTH, SEEK_SET);
	    for (i = 0; i < 683; i++) {
		char    tmp[256];
		if (read(fd, tmp, 256) != 256) {
		    printf("cannot read block %d of %s\n", i, argv[3]);
		    exit(1);
		}
		if (write(FloppyActiveFd, tmp, 256) != 256) {
		    printf("cannot read block %d of %s\n", i, argv[3]);
		    exit(1);
		}
	    }
	    close(fd);
	}
	break;

      case 'v':
	if (argc != 3)
	    usage();
	open_image(argv[2], 0);
	printf("validating\n");
	do_validate();
	break;

      case 'i':
	if (argc != 3)
	    usage();
	fd = open(argv[2], O_RDONLY);
	check_header(fd, &v1, &v2);
	close(fd);
	printf("c1541 image version %d.%02d, emulator version %d.%02d\n",
	    v1, v2, C1541_VERSION_MAJOR, C1541_VERSION_MINOR);
	break;

      default:
	usage();
    }
    if (FloppyActiveFd >= 0)
	close(FloppyActiveFd);
    return (0);
}
#endif
