/*
 * $Id: asm.c,v 1.3 93/06/21 13:37:56 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains functions to implement simple MOS6502
 * Machine Language Monitor
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	asm.c,v $
 * Revision 1.3  93/06/21  13:37:56  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:07:37  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "macro.h"
#include "asm.h"
#include "extern.h"


static ADDRESS address;	/* Default place to put code in Commodore 64. */

static char *modename[] = {
    "IMPLIED", "IMMEDIATE", "ZERO_PAGE", "ZERO_PAGE_X", "ZERO_PAGE_Y",
    "ABSOLUTE", "ABSOLUTE_X", "ABSOLUTE_Y", "INDIRECT", "INDIRECT_X",
    "INDIRECT_Y", "RELATIVE", "ACCUMULATOR", "ASS_CODE",
};	       /* ADDRESSING MODES */

/*static char *convptr;*/	/* pointer: end of scan */

#ifdef __STDC__
static int parse_arguments(ADDRESS, int, char *, BYTE *, int *, int);
static int interpret_instr(char *, int);
int     parse_argstr(char *, char *, int, int, int);
int     sconv(char *, int, int);
char   *e_gets(char *);
#else

static int parse_arguments();
static int interpret_instr();
int     parse_argstr();
int     sconv();
char   *e_gets();
#endif

/*
 * This function implements a simple MOS6502 assembler.
 * The interpreter selects default number base according to the
 * value of hexflg.
 */

#ifndef __STDC__
    void    ass(adr, mode)
    ADDRESS adr;
    int     mode;
#else	       /* __STDC__ */
void    ass(ADDRESS adr, int mode)
#endif	       /* __STDC__ */
{
    char   *line;
    int     err;

    address = adr;

    for (;;) {
	line = read_line(address, 0);

	if (line == NULL || *line == '.' || *line == 'x')
	    break;

	else {
	    if ((err = interpret_instr(line, mode)) > 512)
		printf("**** %.4X SYNTAX ERROR (%d): %s ****\n",
		    address, err, line);
	}
    }
}


/*
 * This routine interprets lines which are not direct mode commands for
 * the program itself.
 * On success data is stored to memory, otherwise an error is returned.
 */

#ifndef __STDC__
static int interpret_instr(line, mode)
    char   *line;
    int     mode;
#else	       /* __STDC__ */
static int interpret_instr(char *line, int mode)
#endif	       /* __STDC__ */
{
    /*static char buf[10];*/
    int     i = 0;
    int     instr = 0;
    int     len = 0;
    int attempt = 0;
    unsigned char arg[4];

    do {
    while (*(line) && isspace(*line))
	line++;

    for (i = 0; line[i] && i < 3; i++) {
	if (isalpha (line[i]))
	    line[i] = toupper (line[i]);
    }	    
	
    /*
     * Machine language commands
     * 
     * First all legal mnemonics are searched and if right instruction is found
     * the remaining line is parsed. The numeric code of  the instruction is
     * returned and arguments are returned in arg array.
     */


	for (i = 0; i < TOTAL_CODES; i++)
	    if (0 == strncmp(lookup[i].mnemonic, line, 3)) {

		instr = parse_arguments
		    (address, i, strlen(line) > 4 ? &line[4] : NULL,
		    arg, &len, mode);
		
		if (instr >= 512)
		    return (instr);
		/* Error code from parse_arguments */

		/*if (tstflg)
		    printf("OK. %d bytes: %d %s\n", len + 1, instr,
			lookup[instr].mnemonic);*/

		ram[address] = instr;
		address += 1;
		if (len) {
		    memcpy(ram + address, arg, len);
		    address += len;
		}
		printf(".%04X %.2X ", address - len - 1, instr);
		switch (len) {
		  case 0:
		    printf("\t%s\n", sprint_opcode(address - len - 1, 1));
		    break;
		  case 1:
		    printf("%.2X\t%s\n", arg[0], sprint_opcode(address - len - 1, 1));
		    break;
		  case 2:
		    printf("%.2X %.2X\t%s\n", arg[0], arg[1],
			sprint_opcode(address - len - 1, 1));
		}
		return (0);

	    }  /* end of matching mnemonic */
    
    if (!*line || !sconv(line, 0, mode | MODE_QUERY))
	return(E_BAD_IDENTIFIER);

    address = sconv(line, 0, mode);
    while(*++line > ' ');

/*printf("rescan %04x ===>%s<===\n", address, line); */

    } while(*line && !attempt++);
    return (E_ERROR);
}	       /* end of interpret_instr */


#ifndef __STDC__
static int parse_arguments(adr, code, line, arg, len, mode)
    ADDRESS adr;
    int     code;
    char   *line;
    BYTE   *arg;
    int    *len;
    int     mode;
#else	       /* __STDC__ */
static int parse_arguments(ADDRESS adr, int code, char *line, BYTE * arg, int *len, int mode)
#endif	       /* __STDC__ */
{
    int     comma_found = 0;	/* True if argument contains comma. */
    int     right_parenthesis = 0;
    int     left_parenthesis = 0;
    int     x_found = 0;
    int     y_found = 0;
    int     zero_page = 0;
    int     addr_mode = 0;
    int     ind_x_found = 0;
    int     ind_y_found = 0;
    int     ival;
    char   *linep;
    int     i;


    if (tstflg)
	printf("LINE####%s\n", line);

    if (NULL == line) {	/* IMPLIED addressing mode. */
	for (i = code; i <= OP_IMPL_MAX; i += OP_IMPL_SPC)
	    if (((IMPLIED == lookup[i].addr_mode) ||
		    (ACCUMULATOR == lookup[i].addr_mode)) &&
		0 == strcmp(lookup[code].mnemonic, lookup[i].mnemonic)) {
		*len = 0;
		return i;
	    }
	return E_PARAMETER_SYNTAX;
    }
    if (tstflg)
	printf("CODE %d LINE %s\n", code, line);

    /*
     * ACCUMULATOR addressing. This piece of code checks if the length of the
     * argument is 1 and if this argument is A. If those conditions are true
     * appropriate machine code where addressing mode is right is chosen,
     * otherwise E_PARAMETER_SYNTAX is returned. This errorcode means that
     * argument is A (=accumulator), but this instruction does not support
     * accumulator addressing. If the parameter A is not present then it
     * matches IMPLIED addressing mode which is implemented above.
     */

    if (tstflg)
	printf("strlen %d --%s--\n", strlen(line), line);

    if (1 == strlen(line) && 0 == strcmp(line, "A")) {
	for (i = code; i < 256; ++i)
	    if (ACCUMULATOR == lookup[i].addr_mode &&
		0 == strcmp(lookup[code].mnemonic, lookup[i].mnemonic)) {
		*len = 0;
		return i;
	    }
	return E_PARAMETER_SYNTAX;
    }
    if (tstflg)
	printf("ACCU testattu\n");


    /* IMMEDIATE addressing mode */

    if ('#' == *line) {
	for (i = code; i < OP_IMM_MAX; ++i)
	    if (IMMEDIATE == lookup[i].addr_mode &&
		0 == strcmp(lookup[code].mnemonic, lookup[i].mnemonic)) {
		*len = 1;
		*arg = (BYTE) sconv(&line[1], 0, mode);
		return i;
	    }
	return E_PARAMETER_SYNTAX;
    }
    /* Next check if relative addressing mode was used. */

    if (RELATIVE == lookup[i = code].addr_mode) {
	if (tstflg)
	    printf("addressing mode RELATIVE\n");

	*len = 1;
	ival = sconv(line + 1, 0, mode);

	if (tstflg)
	    printf("rel disp %d  %.4x\n", ival - adr - 2, ival - adr - 2);

	if ((ival -= (adr + 2)) <= 127 && ival >= -128) {
	    *arg = (BYTE) ((ival) & 0xFF);
	    return i;
	}
	return E_LONG_BRANCH;
    }
    /*
     * Now it's time to parse absolute and zeropage addressing modes with
     * their indexed and indirect versions. First check if there are any
     * commas. And count also parenthesis.
     */
    linep = line;

    while (*linep)
	switch (*linep++) {

	  case ')':
	    ++right_parenthesis;
	    break;
	  case '(':
	    ++left_parenthesis;
	    break;

	  case ',':
	    ++comma_found;
	    while (*linep == ' ')
		linep++;

	    if (*linep) {
		if ((*linep & 0xdf) == 'X') {
		    if (!left_parenthesis)
			++x_found;
		    else
			++ind_x_found;
		    linep++;
		} else if ((*linep & 0xdf) == 'Y') {
		    if (!left_parenthesis)
			++y_found;
		    else
			++ind_y_found;
		    linep++;
		}
		if (*linep && !strchr(" )", linep[1]))
		    return E_BAD_INDEX;
	    }
	    break;
	}

    if (comma_found > 1)	/* There can not be more than one comma! */
	return E_TOO_MANY_COMMAS;

    if (right_parenthesis > 1)	/* Only one ) can be found. */
	return E_RIGHT_PARENTHESIS;

    if (left_parenthesis > 1)	/* Only one ( can be found. */
	return E_LEFT_PARENTHESIS;

    if (left_parenthesis != right_parenthesis)	/* Must be equal. */
	return E_PARENTHESIS;

    if (comma_found && !(x_found || y_found || ind_x_found || ind_y_found))
	return E_MISSING_XY;

    if (x_found && y_found)
	return E_MIXED_XY;	/* There is not x AND y indexed addressing
				 * mode */


    linep = line;

    if (right_parenthesis)
	while (*linep && (*linep++ != '('));

    /* ival = atoi( linep ); */
    ival = sconv(linep, 0, mode);

    /* E_PARAMETER_SYNTAX indicates an error */

    if (tstflg)
	printf("integer: %d\n", ival);
    if ((ival < 256) && !y_found)
	zero_page = 1;

    if (right_parenthesis && !comma_found)
	addr_mode = INDIRECT;

    if (zero_page) {
	if (!right_parenthesis && !comma_found)
	    addr_mode = ZERO_PAGE;
	if (x_found)
	    addr_mode = ZERO_PAGE_X;
	if (y_found)
	    addr_mode = ZERO_PAGE_Y;
	if (ind_x_found)
	    addr_mode = INDIRECT_X;
	if (ind_y_found)
	    addr_mode = INDIRECT_Y;
    }
    if (!zero_page) {
	if (!comma_found && !right_parenthesis)
	    addr_mode = ABSOLUTE;
	if (x_found)
	    addr_mode = ABSOLUTE_X;
	if (y_found)
	    addr_mode = ABSOLUTE_Y;
    }
    if (tstflg)
	printf("addressing mode 1 : %s  -- code $%.2x\n",
	    modename[addr_mode], code);

    if (0 == addr_mode)
	return E_ERROR;

    /*
     * Addressing mode was found and the appropriate opcode is being selected
     */

    for (i = code; i < TOTAL_CODES; ++i)
	if (addr_mode == lookup[i].addr_mode &&
	    0 == strcmp(lookup[code].mnemonic, lookup[i].mnemonic)) {
	    switch (addr_mode) {

	      case ZERO_PAGE:
	      case ZERO_PAGE_X:
	      case ZERO_PAGE_Y:
	      case INDIRECT_X:
	      case INDIRECT_Y:
		*len = 1;
		*arg = (BYTE) ival;
		break;

	      case INDIRECT:
		if ((ival & 0xff) == 0xff)
		    printf("%.4X: Warning: Indirect JMP($%.4X)\n", adr, ival);
		/* The famous bug in jmp($indirect) */
		/* fall trough */

	      case ABSOLUTE:
	      case ABSOLUTE_X:
	      case ABSOLUTE_Y:
		*len = 2;
		*arg++ = (BYTE) ival & 0xFF;
		*arg = (BYTE) (ival >> 8) & 0xFF;
	    }
	    return i;
	}
    /*
     * If machine code has not zeropage addressing mode, absolute addressing
     * mode is tried instead.
     */

    if (tstflg)
	printf("addressing mode 2 : ABSOLUTE\n");
    if (ZERO_PAGE & addr_mode) {
	for (i = code; i <= OP_ABS_MAX; i += OP_ABS_SPC)
	    if ((ABSOLUTE | (~ZERO_PAGE & addr_mode)) == lookup[i].addr_mode &&
		0 == strcmp(lookup[code].mnemonic, lookup[i].mnemonic)) {
		*len = 2;
		*arg++ = (unsigned char) ival & 0xFF;
		*arg = (unsigned char) (ival >> 8) & 0xFF;
		return i;
	    }
    }
    fprintf(stderr, "addressing mode failed\n");
    return E_PARAMETER_SYNTAX;
}
