/*
 * $Id: blist.c,v 1.3 93/06/21 13:38:03 jopi Exp $
 *
 * This file is part of Commodore 64 emulator
 *      and Program Development System.
 * See README for copyright notice
 *
 * This program expands your tokenised C64/128 BASIC programs
 * into 7-bit ASCII text. Unprintable characters are shown
 * as hexadecimal codes in parenthesis.
 * It is also possible to convert programs from ascii listings
 * into tokenised basic v2.0, v7.0 or simon's basic programs.
 * This program does not convert control code names to codes.
 *
 * Usage: blist [-c] [-w{2|s|7}] files
 *    or  cat file | blist [-c] [-w{2|s|7}]
 *
 *
 * In shell:
 *  tr '\015A-Za-z\301-\332\\\|\[\{\]\}' '\012a-zA-ZA-Z\|\\\{\[\}\]'
 * or
 *  tr '\015A-Za-z\\\|\[\{\]\}' '\012a-zA-Z\|\\\{\[\}\]'
 *
 * Written by
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	blist.c,v $
 * Revision 1.3  93/06/21  13:38:03  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:07:42  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

void b_expand();
void b_tokenise();

#define BASIC_2   1
#define BASIC_S   2
#define BASIC_7   3
#define MAX_KWCE  0xb
#define MAX_KWFE  0x27
#define KW_NONE   0xff

/* local */

#ifdef __STDC__
void b_expand(char *, int);
void b_tokenise(int version);
unsigned char sstrcmp(char *, char**, int, int);
#else
void b_expand();
void b_tokenise();
unsigned char sstrcmp();
#endif


 /* basic keywords */

char   *keyword[] = {
    /* 80 - ca */
    "end", "for", "next", "data", "input#", "input", "dim", "read",
    "let", "goto", "run", "if", "restore", "gosub", "return", "rem",
    "stop", "on", "wait", "load", "save", "verify", "def", "poke",
    "print#", "print", "cont", "list", "clr", "cmd", "sys", "open",
    "close", "get", "new", "tab(", "to", "fn", "spc(", "then",
    "not", "step", "+", "-", "*", "/", "^", "and",
    "or", ">", "=", "<", "sgn", "int", "abs", "usr", "fre", "pos",
    "sqr", "rnd", "log", "exp", "cos", "sin",
    "tan", "atn", "peek", "len", "str$", "val", "asc", "chr$",
    "left$", "right$", "mid$",
/* the following codes (0xcb- 0xfe) are for v7 only */
    "go", "rgr", "rlcr", /* 0xce -- prefix */ "", "joy",
    "rdot", "dec", "hex$", "err$", "instr", "else", "resume", "trap",
    "tron", "troff", "sound", "vol", "auto", "pudef", "graphic", "paint",
    "char", "box", "circle", "gshape", "sshape", "draw", "locate", "color",
    "scnclr", "scale", "help", "do", "loop", "exit", "directory", "dsave",
    "dload", "header", "scratch", "collect", "copy", "rename", "backup",
    "delete", "renumber", "key", "monitor", "using", "while", "until",
    /* 0xfe -- prefix */ "", "pi"
};

char   *kwce[] = {
    " ", " ", "pot", "bump", "pen", "rspos", "rsprite", "rspcolor",
    "xor", "rwindow", "pointer"
};

char   *kwfe[] = {
    "", "", "bank", "filter", "play", "tempo", "movspr", "sprite", "sprcolor",
    "rreg", "envelope", "sleep", "catalog", "dopen", "append", "dclose",
    "bsave", "bload", "record", "concat", "dverify", "dclear", "sprsav",
    "collision", "begin", "bend", "window", "boot", "width", "sprdef",
    "quit", "stash", " ", "fetch", " ", "swap", "off", "fast", "slow"
};

char   *simonskw[] = {
    "", "hires", "plot", "line", "block", "fchr", "fcol", "fill",
    "rec", "rot", "draw", "char", "hi col", "inv", "frac", "move",
    "place", "upb", "upw", "leftw", "leftb", "downb", "downw", "rightb",
    "rightw", "multi", "colour", "mmob", "bflash", "mob set", "music", "flash",
    "repeat", "play", ">>", "centre", "envelope", "cgoto", "wave", "fetch",
    "at(", "until", ">>", ">>", "use", ">>", "global", ">>",
    "reset", "proc", "call", "exec", "end proc", "exit", "end loop", "on key",
    "disable", "resume", "loop", "delay", ">>", ">>", ">>", ">>",
    "secure", "disapa", "circle", "on error", "no error", "local", "rcomp", "else",
    "retrace", "trace", "dir", "page", "dump", "find", "option", "auto",
    "old", "joy", "mod", "div", ">>", "dup", "inkey", "inst",
    "test", "lin", "exor", "insert", "pot", "penx", ">>", "peny",
    "sound", "graphics", "design", "rlocmob", "cmob", "bckgnds", "pause", "nrm",
    "mob off", "off", "angl", "arc", "cold", "scrsv", "scrld", "text",
    "cset", "vol", "disk", "hrdcpy", "key", "paint", "low col", "copy",
    "merge", "renumber", "mem", "detect", "check", "display", "err", "out"
};

 /* control characters.
  * NOTE: On C128 code 0x09 means TAB instead of ENSH
  */

char   *control1[] = {
    "", "", "", "", "", "(wht)", "", "",
    "(dish)", "(ensh)", "", "", "", "\n", "(swlc)", "",
    "", "(down)", "(rvon)", "(home)", "(del)", "", "", "",
    "", "", "", "(esc)", "(red)", "(rght)", "(grn)", "(blu)"
};

char   *control2[] = {
    "", "(orng)", "", "", "", "(F1)", "(F3)", "(F5)",
    "(F7)", "(F2)", "(F4)", "(F6)", "(F8)", "(sret)", "swuc)", "",
    "(blk)", "(up)", "(rvof)", "(clr)", "(inst)", "(brn)", "(lred)", "(gry1)",
    "(gry2)", "(lgrn)", "(lblu)", "(gry3)", "(pur)", "(left)", "(yel)", "(cyn)"
};

FILE    *source;
int     fil = 0, kwlen;


#ifndef __STDC__
void    main(argc, argv)
    int     argc;
    char   *argv[];
#else
void    main(int argc, char **argv)
#endif
{
    char   *progname;
    int     wr_mode = 0;
    int     no_ctrls= 0;

    progname = argv[0];
    while (--argc && ((*++argv)[0] == '-')) {	/* parse arguments */

	if (!strcmp(argv[0], "-w2") && !wr_mode) {
	    wr_mode = BASIC_2;
	    continue;
	}
	else if (!strcmp(argv[0], "-ws") && !wr_mode) {
	    wr_mode = BASIC_S;
	    continue;
	}
        else if (!strcmp(argv[0], "-w7") && !wr_mode) {
	    wr_mode = BASIC_7;
	    continue;
	}
        else if (!strcmp(argv[0], "-c")) {
	    no_ctrls++;
	    continue;
	}
	else fprintf(stderr,
		"\nUsage: %7s [-c] [-w2 | -ws | -w7] [file list]\n", argv[0]);
	exit(1);
    }

    if (argc)
	fil++;

    do {
	if (!fil)
	    source = stdin;
	else {
	    if ((source = fopen(argv[0], "r")) == NULL) {
		fprintf(stderr,
			"\n%s: Can't open file %s\n", progname, argv[0]);
		exit(-1);
	    }
	}

	if (wr_mode)
	    b_tokenise(wr_mode);
	else
	    b_expand(argv[0], no_ctrls);

	if (fil)
	    fclose(source);
    } while (fil && --argc && ++argv);	/* next file */
}

 
#ifndef __STDC__
void b_expand(fname, clean)
     char *fname;
     int clean;
#else
void b_expand(char *fname, int clean)
#endif
{
    static char line[4];
    unsigned int c;
    int quote;

    /* show load address */
    fread(line, 1, 2, source);
    if (fil)
	printf("\n\n%s ", fname);
    printf("==%04X==\n", line[0] + 256 * line[1]);

    while ((fread(line, 1, 4, source) == 4) && (line[0] || line[1])) {
	quote = 0;
	printf(" %4d ", (line[2] & 0xff) + ((line[3] & 0xff) << 8));

	/* prevent list protection from terminating listing */

	while ((c = getc(source)) != EOF && !c);
	do {
	    if (c == 0x22)
		quote ^= c;

	    /*
	     * simon's basic. Any flag for this is not needed since all
	     * implemented modes are mutually exclusive
	     */

	    if (!quote && (c == 0x64)) {
		if ((c = getc(source)) < 0x80) {
		    printf("%s", simonskw[c]);
		    continue;
		} else
		    printf("($64)");	/* it wasn't prefix */
	    }
	    /* basic 2.0 and 7.0 */

	    if (!quote && c > 0x7f) {
		if (c == 0xce) {
		    if ((c = getc(source)) < MAX_KWCE)
			printf("%s", kwce[c]);
		    else
			printf("($ce%.2x)", c);
		} else if (c == 0xfe) {
		    if ((c = getc(source)) < MAX_KWFE)
			printf("%s", kwfe[c]);
		    else
			printf("($fe%.2x)", c);
		} else
		    printf("%s", keyword[c & 0x7f]);
		continue;
	    }
	    if ((c >= 'A') && (c <= 0x7e))
		putchar(c ^ 0x20);
	    else if (isprint(c))
		putchar(c);
	    else {
		if ((c < 0x20) && !clean && control1[c])
		    printf("%s", control1[c]);
		else if (((c & 0x7f) < 0x20) && !clean && control2[c & 0x1f])
		    printf("%s", control2[c & 0x1f]);
		else if (c >= 0xc0 && c < 0xe0)
		    putchar(c - 0x80);
		else if (c == 0xff)
		    putchar(0x7e);
		else
		    printf("($%.2x)", c & 0xff);
	    }
	} while ((c = getc(source)) != EOF && c);
	printf("\n");

    }      /* line */
}

 
#ifndef __STDC__
void b_tokenise(version)
     int version;
#else
void b_tokenise(int version)
#endif
{
    static char line[256];
    unsigned char c;
    char *p1, *p2;
    int   quote, len = 0, linum = 10, match, addr = 0x0801;

    printf("%c%c", 0x01, 0x08);

    while((p1 = p2 = fgets(line, 255, source)) != NULL) {

	if (sscanf(line, "%d%n", &linum, &len) == 1)
	    p2 += len;

	quote = 0;
	while (isspace(*p2)) p2++;

	while (*p2) {
	    if (*p2 == 0x22) {
		quote ^= *p2;
	    }
	    if (*p2 == 0x0a || *p2 == 0x0d)
		break;

	    match = 0;
	    if (!quote && isalpha(*p2)) {
		if ((c = sstrcmp(p2, keyword, 0,
				 ((version==BASIC_7) ? 0x80 :
				  0x4b)) ) != KW_NONE) {
			*p1 = c | 0x80;
			p2 += kwlen;
			match++;
		    }
		if (!match && (version == BASIC_7)) {
		    if ((c = sstrcmp(p2, kwfe, 2, MAX_KWFE)) != KW_NONE) {
			*p1++ = 0xfe;
			*p1 = c;
			p2 += kwlen;
			match++;
		    }
		    else if ((c = sstrcmp(p2, kwce, 2, MAX_KWCE)) != KW_NONE) {
			*p1++ = 0xce;
			*p1 = c;
			p2 += kwlen;
			match++;
		    }
		}
		if (!match && (version == BASIC_S)) {
		    if ((c = sstrcmp(p2, simonskw, 1, 0x80)) != KW_NONE) {
			*p1++ = 0x64;
			*p1 = c;
			p2 += kwlen;
			match++;
		    }
		}
	    } /* quote */

	    if (!match) {
		if ((*p2 >= 0x5b) && (*p2 <= 0x7e))
		    *p1 = *p2 ^ 0x20;

		else if ((*p2 >= 'A') && (*p2 <= 'Z'))
		    *p1 = *p2 + 0x80;

		else if (*p2 == 0x7e)
		    *p1 = 0xff;
		/*
		 *  control code evaluation should be placed here
		 */
		else
		    *p1 = *p2;

		++p2;
	    } /* match */

	    ++p1;
	} /* while */

	*p1 = 0;
	if ((len = strlen(line) ) /*>0*/) {
	    addr += len + 5;
	    printf("%c%c%c%c%s%c", addr & 255, (addr>>8) & 255,
		   linum & 255, (linum>>8) & 255, line, 0);

	    linum += 2; /* auto line numbering as default */
	}
    }
}


#ifndef __STDC__
unsigned char sstrcmp(line, wordlist, token, maxitems)
     char *line, **wordlist;
     int token, maxitems;
#else
unsigned char sstrcmp(char *line, char **wordlist, int token, int maxitems)
#endif
{
    int bi = 0, j;
    char *p, *q;

    kwlen = 0;
    /* search for keyword */
    for (; token < maxitems; token++) {
	for (p = wordlist[token], q = line, j = 0;
	     *p && *q && *p == *q; p++, q++, j++);

	/* printf ("compare %s %s - %d %d %d\n",
		wordlist[token], line, j, kwlen, bi);*/

	if (j > kwlen && !*p) {
	    kwlen = j;
	    bi = token;

	    /* found an exact keyword */
	    if (!*p)
		return token;
	}
    } /* for */

    return KW_NONE;
}
