/*
 * $Id: cia1.c,v 1.3 93/06/21 13:38:13 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 * 
 * $Log:	cia1.c,v $
 * Revision 1.3  93/06/21  13:38:13  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:07:58  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include "macro.h"
#include "extern.h"
#include <stdio.h>
#include <time.h>

/* global */

BYTE    cia1[16];
BYTE    keyarr[8];

/* local functions */

#ifndef __STDC__
static void update_cia1ta();
static void update_cia1tb();
#else
static void update_cia1ta(int);
static void update_cia1tb(int);
#endif

/* local variables */

static CLOCK cia1tau;	/* time when cia 1 timer a is updated */
static CLOCK cia1tbu;	/* time when cia 1 timer b is updated */

static int cia1ta;	/* value of cia 1 timer a at last update */
static int cia1tb;	/* value of cia 1 timer b at last update */

static int tod1freq;	/* frequency of cia 1 TOD */
static int tod1rel;	/* TOD 1 difference to system clock */
static int tod1alarm;	/* TOD 1 alarm time */

static int cia1int;	/* interrupt register for cia1 */

/* ------------------------------------------------------------------------- */
/* CIA 1 */

#ifndef __STDC__
void    store_cia1(addr, byte)
    ADDRESS addr;
    BYTE    byte;
#else	       /* __STDC__ */
void    store_cia1(ADDRESS addr, BYTE byte)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("store cia1[%x] %x\n", (int) addr, (int) byte);

    switch (addr) {
      case 0: /* port a */
	cia1[0] = (cia1[0] & ~cia1[2]) | (byte & cia1[2]);
	break;

      case 1: /* port b */
	cia1[1] = (cia1[1] & ~cia1[3]) | (byte & cia1[3]);
	break;

      case 2: /* data direction register for port a */
      case 3: /* data direction register for port b */
	cia1[addr] = byte;
	break;

      case 4: /* timer a low */
      case 5: /* timer a high */
	cia1[addr] = byte;
	update_cia1ta(1);
	break;

      case 6: /* timer b low */
      case 7: /* timer b high */
	cia1[addr] = byte;
	update_cia1tb(1);
	break;

      case 8: /* Time Of Day clock 1/10 s */
      case 9: /* Time Of Day clock sec */
      case 10:/* Time Of Day clock min */
      case 11:/* Time Of Day clock hour */
	cia1[addr] = byte;
	tod1rel = tod1alarm = (int)time(NULL);
	break;

      case 12:/* serial port shift register */
	cia1[addr] = byte;
	break;

      case 13:/* interrupt control register */
	if (byte & 0x80) {
	    /* set interrupts */
	    if (byte & 1) {
		update_cia1ta(0);
		int_clk[I_CIA1TA] = cia1tau + cia1ta;
	    }
	    if (byte & 2) {
		update_cia1tb(0);
		int_clk[I_CIA1TB] = cia1tbu + cia1tb;
	    }
	    if (byte & 4)
		int_clk[I_CIA1TOD] = 0;	/* tod not implemented */
	    if (byte & 8)
		int_clk[I_CIA1SP] = 0;	/* serial data empty/full not impl. */
	    if (byte & 0x10)
		int_clk[I_CIA1FL] = 0;	/* FLAG1 not impl. */
	    cia1[addr] |= byte & 0x7f;
	} else {
	    /* clear interrupts */
	    if (byte & 1)
		int_clk[I_CIA1TA] = 0;
	    if (byte & 2)
		int_clk[I_CIA1TB] = 0;
	    if (byte & 4)
		int_clk[I_CIA1TOD] = 0;
	    if (byte & 8)
		int_clk[I_CIA1SP] = 0;
	    if (byte & 0x10)
		int_clk[I_CIA1FL] = 0;
	    cia1[addr] &= ~byte;
	}
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;

      case 14:/* control register A */
	cia1[addr] = byte;

	/* bit 7 tod frequency */
	tod1freq = (byte & 0x80) ? 50 : 60;

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1ta(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer */
	    update_cia1ta(0);
	}
	break;

      case 15:/* control register B */
	cia1[addr] = byte;
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1tb(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer if mode = 01 or 11 */
	    update_cia1tb(0);
	}
	break;
    }
}

#ifndef __STDC__
BYTE    read_cia1(addr)
    ADDRESS addr;
#else	       /* __STDC__ */
BYTE    read_cia1(ADDRESS addr)
#endif	       /* __STDC__ */
{
    switch (addr) {
	case 0:/* port a */
	{
	    int     t, i;
	    BYTE    j;
	    x_get_keyarr();
	    for (j = i = 0, t = cia1[1] & cia1[3]; i < 8; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    if (debugflg)
		printf("read port a %d\n", j);
	    return (((j & ~cia1[2]) | (cia1[0] & cia1[2])) & ~joy1);
	}

      case 1: /* port b */
	{
	    int     t, i;
	    BYTE    j;
	    x_get_keyarr();
	    for (j = i = 0, t = cia1[0] & cia1[2]; i < 8; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    j = ~j;
	    if (debugflg) {
		printf("read port b %d\n", j);
		printf("a: %x b:%x ca: %x cb: %x joy: %x\n",
		    (int) cia1[0], (int) cia1[1], (int) cia1[2], (int) cia1[3],
		    joy1);
	    }
	    return (((j & ~cia1[3]) | (cia1[1] & cia1[3])) & ~joy1);
	}

      case 2: /* data direction register for port a */
      case 3: /* data direction register for port b */
	return (cia1[addr]);

      case 4: /* timer a low */
	update_cia1ta(0);
	return (cia1ta & 0xff);

      case 5: /* timer a high */
	update_cia1ta(0);
	return ((cia1ta >> 8) & 0xff);

      case 6: /* timer b low */
	update_cia1tb(0);
	return (cia1tb & 0xff);

      case 7: /* timer b high */
	update_cia1tb(0);
	return ((cia1tb >> 8) & 0xff);

      case 8: /* Time Of Day clock 1/10 s */
      case 9: /* Time Of Day clock sec */
      case 10:/* Time Of Day clock min */
      case 11:/* Time Of Day clock hour */
	return (cia1[addr]);

      case 12:/* serial port shift register */
	return (cia1[addr]);

      case 13:/* interrupt control register */
	{
	    BYTE    t = cia1int;
	    cia1int = 0;
	    return (t);
	}

      case 14:/* control register A */
      case 15:/* control register B */
	break;
    }
    return (cia1[addr]);
}

#ifndef __STDC__
void    int_cia1ta()
#else	       /* __STDC__ */
void    int_cia1ta(void)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("cia1 timer a interrupt\n");
    if (cia1[0xe] & 8) {
	/* one-shot mode */
	cia1ta = 0;
	cia1tau = clk;
	int_clk[I_CIA1TA] = 0;
    } else {
	/* continuous mode */
	update_cia1ta(1);
	int_clk[I_CIA1TA] = cia1tau + cia1ta;
    }
    cia1int = 0x80 | 1;

    /* count timer a underflow pulses */
    if ((cia1[15] & 0x61) == 0x41)
	if (--cia1tb <= 0)
	    int_clk[I_CIA1TB] = clk;
}

#ifndef __STDC__
static void update_cia1ta(force)
    int     force;
#else	       /* __STDC__ */
static void update_cia1ta(int force)
#endif	       /* __STDC__ */
{
    if (force)
	cia1ta = cia1[4] + (cia1[5] << 8);
    else if ((cia1[14] & 0x21) == 1)
	cia1ta = (int)(clk - cia1tau + cia1ta) & 0xffff;
    cia1tau = clk;
}

/*
 * Timer b can run in 2 (4) modes
 * cia1[f] & 0x60 == 0x00   count system 02 pulses
 * cia1[f] & 0x60 == 0x40   count timer a underflows
 * cia1[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


#ifndef __STDC__
void    int_cia1tb()
#else	       /* __STDC__ */
void    int_cia1tb(void)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("cia1 timer b interrupt\n");

    if (cia1[0xf] & 8) {
	/* one-shot mode */
	cia1tb = 0;
	cia1tbu = clk;
	int_clk[I_CIA1TB] = 0;
    } else {
	/* continuous mode */
	update_cia1tb(1);
	if ((cia1[15] & 0x61) == 0x01)
	    int_clk[I_CIA1TB] = cia1tbu + cia1tb;
    }
    cia1int = 0x80 | 2;
}

#ifndef __STDC__
static void update_cia1tb(force)
    int     force;
#else	       /* __STDC__ */
static void update_cia1tb(int force)
#endif	       /* __STDC__ */
{
    if (force) {
	cia1tb = cia1[4] + (cia1[5] << 8);
    } else if ((cia1[15] & 0x61) == 1)
	/* count system 02 pulses */
	cia1tb = (int)(clk - cia1tbu + cia1tb) & 0xffff;
    cia1tbu = clk;
}

#ifndef __STDC__
int     show_keyarr()
#else	       /* __STDC__ */
int     show_keyarr(void)
#endif	       /* __STDC__ */
{
    int     i, j;
    for (j = 0; j < 8; j++) {
	printf("%d:", j);
	for (i = 0x80; i; i >>= 1)
	    printf(" %d", keyarr[j] & i);
	printf("\n");
    }
    return (0);
}
