/*
 * $Id: cia2.c,v 1.3 93/06/21 13:38:16 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	cia2.c,v $
 * Revision 1.3  93/06/21  13:38:16  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:02  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include "macro.h"
#include "extern.h"
#include <stdio.h>
#include <time.h>

/* global */

BYTE    cia2[16];

/* local functions */

#ifndef __STDC__
static void update_cia2ta();
static void update_cia2tb();
#else
static void update_cia2ta(int);
static void update_cia2tb(int);
#endif

/* local variables */

static CLOCK cia2tau;	/* time when cia 2 timer a is updated */
static CLOCK cia2tbu;	/* time when cia 2 timer b is updated */

static int cia2ta;	/* value of cia 2 timer a at last update */
static int cia2tb;	/* value of cia 2 timer b at last update */

static int tod2freq;	/* frequency of cia 2 TOD */
static int tod2rel;	/* TOD 2 difference to system clock */
static int tod2alarm;	/* TOD 2 alarm time */

static int cia2int;	/* interrupt register for cia2 */


/* ------------------------------------------------------------------------- */
/* CIA 2 */

#ifndef __STDC__
void    store_cia2(addr, byte)
    ADDRESS addr;
    BYTE    byte;
#else	       /* __STDC__ */
void    store_cia2(ADDRESS addr, BYTE byte)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("store cia2[%x] %x\n", (int) addr, (int) byte);

    switch (addr) {
      case 0: /* port a */
	{
	    BYTE    tmp = cia2[0];
	    cia2[0] = (cia2[0] & ~cia2[2]) | (byte & cia2[2]);
	    vbank = ((cia2[0] & 3) ^ 3) << 2;
	    if ((tmp ^ cia2[0]) & 3)
		screen_init();
	    break;
	}

      case 1: /* port b */
	cia2[1] = (cia2[1] & ~cia2[3]) | (byte & cia2[3]);
	break;

      case 2: /* data direction register for port a */
      case 3: /* data direction register for port b */
	cia2[addr] = byte;
	break;

      case 4: /* timer a low */
      case 5: /* timer a high */
	cia2[addr] = byte;
	update_cia2ta(1);
	break;

      case 6: /* timer b low */
      case 7: /* timer b high */
	cia2[addr] = byte;
	update_cia2tb(1);
	break;

      case 8: /* Time Of Day clock 1/10 s */
      case 9: /* Time Of Day clock sec */
      case 10:/* Time Of Day clock min */
      case 11:/* Time Of Day clock hour */
	cia2[addr] = byte;
	tod2rel = tod2alarm = (int)time(NULL);
	break;

      case 12:/* serial port shift register */
	cia2[addr] = byte;
	break;

      case 13:/* interrupt control register */

	if (byte & 0x80) {
	    /* set interrupts */
	    if (byte & 1) {
		/* update_cia2ta(0); */
		int_clk[I_CIA2TA] = cia2tau + cia2ta;
	    }
	    if (byte & 2) {
		/* update_cia2tb(); */
		int_clk[I_CIA2TB] = cia2tbu + cia2tb;
	    }
	    if (byte & 4)
		int_clk[I_CIA2TOD] = 0;	/* tod not implemented */
	    if (byte & 8)
		int_clk[I_CIA2SP] = 0;	/* serial data empty/full not impl. */
	    if (byte & 0x10)
		int_clk[I_CIA2FL] = 0;	/* FLAG1/ RESTORE key */
	    cia2[addr] |= byte & 0x7f;
	} else {

	    /* clear interrupts */
	    if (byte & 1)
		int_clk[I_CIA2TA] = 0;
	    if (byte & 2)
		int_clk[I_CIA2TB] = 0;
	    if (byte & 4)
		int_clk[I_CIA2TOD] = 0;
	    if (byte & 8)
		int_clk[I_CIA2SP] = 0;
	    if (byte & 0x10)
		int_clk[I_CIA2FL] = 0;
	    cia2[addr] &= ~byte;
	}
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;

      case 14:/* control register A */
	cia2[addr] = byte;

	/* bit 7 tod frequency */
	tod2freq = (byte & 0x80) ? 50 : 60;

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    /* update_cia2ta(1); */
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer */
	    update_cia2ta(0);
	}
	break;

      case 15:/* control register B */
	cia2[addr] = byte;
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia2tb(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer if mode = 01 or 11 */
	    update_cia2tb(0);
	}
	break;
    }
}

#ifndef __STDC__
BYTE    read_cia2(addr)
    ADDRESS addr;
#else	       /* __STDC__ */
BYTE    read_cia2(ADDRESS addr)
#endif	       /* __STDC__ */
{
    switch (addr) {
	case 0:/* port a */
	return ((cia2[0] & cia2[2]) | (0xff & ~cia2[2]));

      case 1: /* port b */
	return ((cia2[1] & cia2[3]) | (0xff & ~cia2[3]));

      case 2: /* data direction register for port a */
      case 3: /* data direction register for port b */
	return (cia2[addr]);

      case 4: /* timer a low */
	update_cia2ta(0);
	return (cia2ta & 0xff);

      case 5: /* timer a high */
	update_cia2ta(0);
	return ((cia2ta >> 8) & 0xff);

      case 6: /* timer b low */
      case 7: /* timer b high */
	return (cia2[addr]);

      case 8: /* Time Of Day clock 1/10 s */
      case 9: /* Time Of Day clock sec */
      case 10:/* Time Of Day clock min */
      case 11:/* Time Of Day clock hour */
	return (cia2[addr]);

      case 12:/* serial port shift register */
	return (cia2[addr]);

      case 13:/* interrupt control register */
	{
	    BYTE    t = cia2int;
	    cia2int = 0;
	    return (t);
	}

      case 14:/* control register A */
      case 15:/* control register B */
	break;
    }
    return (cia2[addr]);
}

#ifndef __STDC__
void    int_cia2ta()
#else	       /* __STDC__ */
void    int_cia2ta(void)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("cia2 timer a interrupt\n");
    if (cia2[0xe] & 8) {
	/* one-shot mode */
	cia2ta = 0;
	cia2tau = clk;
	int_clk[I_CIA2TA] = 0;
    } else {
	/* continuous mode */
	update_cia2ta(1);
	int_clk[I_CIA2TA] = cia2tau + cia2ta;
    }
    cia2int = 0x80 | 1;

    /* count timer a underflow pulses */
    if ((cia2[15] & 0x61) == 0x41)
	if (--cia2tb <= 0)
	    int_clk[I_CIA2TB] = clk;
}

#ifndef __STDC__
static void update_cia2ta(force)
    int     force;
#else	       /* __STDC__ */
static void update_cia2ta(int force)
#endif	       /* __STDC__ */
{
    if (force)
	cia2ta = cia2[4] + (cia2[5] << 8);
    else if ((cia2[14] & 0x21) == 1)
	cia2ta = (int)(clk - cia2tau + cia2ta) & 0xffff;
    cia2tau = clk;
}

#ifndef __STDC__
void    int_cia2tb()
#else	       /* __STDC__ */
void    int_cia2tb(void)
#endif	       /* __STDC__ */
{
    if (debugflg)
	printf("cia2 timer b interrupt\n");

    if (cia2[0xf] & 8) {
	/* one-shot mode */
	cia2tb = 0;
	cia2tbu = clk;
	int_clk[I_CIA2TB] = 0;
    } else {
	/* continuous mode */
	update_cia2tb(1);
	if ((cia2[15] & 0x61) == 0x01)
	    int_clk[I_CIA2TB] = cia2tbu + cia2tb;
    }
    cia2int = 0x80 | 2;
}

#ifndef __STDC__
static void update_cia2tb(force)
    int     force;
#else	       /* __STDC__ */
static void update_cia2tb(int force)
#endif	       /* __STDC__ */
{
    if (force) {
	cia2tb = cia2[4] + (cia2[5] << 8);
    } else if ((cia2[15] & 0x61) == 1)
	/* count system 02 pulses */
	cia2tb = (int)(clk - cia2tbu + cia2tb) & 0xffff;
    cia2tbu = clk;
}
