/*
 * $Id: ed.c,v 1.3 93/06/21 13:38:26 jopi Exp $
 *
 * See README for copyright notice
 *
 * Copyright (c) 1993
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 * $Log:	ed.c,v $
 * Revision 1.3  93/06/21  13:38:26  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:13  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * now generates ophack2, cpu emulation with switch case structure
 *
 *
 */

/*#define COMPLETE*/
/*#define MAIN*/
#define ONELINE

#include "vmachine.h"

#ifdef EDIT
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifdef COMPLETE
#include <sys/types.h>
#include <dirent.h>
#endif

#ifdef __cplusplus
#define __STDC__ 0
#define fpt   int(*)(char *, int)
#else
#define fpt   int(*)()
#endif

#include "proto.h"

#define MAXKEYS 100
#define MAXHIST 50
#define TTYFD 0
#define DRAW 1
#define EMPTY 2
#define COUNT 3

#define beep() rawout('\07')
#define bufout(ch) outbuf[outcount++]=(ch)

char   *history[MAXHIST];
static struct termios old_termio;
static char k_buf[10], rivi[1024], yank[1024], outbuf[1024];
static int ind, rind, curhist, xpos, curpos, minpos, realpos;
static int outcount;

int     key_init, tty_reset;

#ifdef COMPLETE
char  **(*complete_lister) ();
char  **make_list();
int     k_complete(), k_complete_list();
#endif

struct k_com {
    int     com;
    char   *str;
}       comlist[MAXKEYS];

struct k_func {
    char   *name;

#ifdef __STDC__
    int     (*ptr) (char *, int);
#else
    int     (*ptr) ();
#endif
}       funlist[] = {

    {
	"no-operation", (fpt) k_nop
    },
    {
	"self-insert-key", (fpt) k_self_insert
    },
    {
	"previous-history", (fpt) k_prev_history
    },
    {
	"next-history", (fpt) k_next_history
    },
    {
	"forward-char", (fpt) k_forward
    },
    {
	"backward-char", (fpt) k_backward
    },
    {
	"backspace", (fpt) k_backspace
    },
    {
	"end-of-line", (fpt) k_eol
    },
    {
	"end-of-file", (fpt) k_eof
    },
    {
	"goto-end", (fpt) k_end
    },
    {
	"goto-begin", (fpt) k_begin
    },
    {
	"redraw-line", (fpt) k_redraw
    },
    {
	"delete-char", (fpt) k_delete
    },
    {
	"kill-to-begin", (fpt) k_kill_begin
    },
    {
	"kill-to-end", (fpt) k_kill_end
    },
    {
	"yank", (fpt) k_yank
    },

#ifdef COMPLETE
    {
	"complete-file", k_complete
    },
    {
	"list-complete", k_complete_list
    },
#endif

    {
	NULL, NULL
    }
};

/* edit functions */

#ifndef __STDC__
void    e_setterm()
#else
void    e_setterm(void)
#endif

{
    struct termios newt;

#ifdef TCGETA
    ioctl(TTYFD, TCGETA, &newt);
#else
    tcgetattr(TTYFD, &newt);
#endif

    /* if we have got out with a signal ... */
    if (!tty_reset)
	old_termio = newt;
    tty_reset = 1;
    newt.c_iflag &= ~(INLCR | ICRNL);
    newt.c_lflag &= ~(

#ifdef MAIN
	ISIG |
#endif

	ICANON | ECHO);
    newt.c_cc[VTIME] = 0;
    newt.c_cc[VMIN] = 1;

#ifdef TCSETA
    ioctl(TTYFD, TCSETA, &newt);
#else
    tcsetattr(TTYFD, TCSANOW, &newt);
#endif

    if (!key_init) {
	int     i;
	key_init++;
	for (i = 0; i < MAXHIST; i++)
	    history[i] = NULL;
	for (i = 0; i < MAXKEYS; i++)
	    comlist[i].str = NULL;
	/* must be first in comlist */
	set_ckey(0xff, "self-insert-key");
	/* set_ckey (newt.c_cc[VEOL], F_EOL); */
	/* set_ckey (newt.c_cc[VEOF], F_EOF); */
	set_ckey(newt.c_cc[VERASE], "backspace");
	set_ckey(newt.c_cc[VKILL], "kill-to-begin");
	e_bind_key("\04", "end-of-file");
	e_bind_key("\r", "end-of-line");
	e_bind_key("\n", "end-of-line");
	e_bind_key("\033[A", "previous-history");
	e_bind_key("\033[B", "next-history");
	e_bind_key("\033[C", "forward-char");
	e_bind_key("\033[D", "backward-char");
	e_bind_key("\033OA", "previous-history");
	e_bind_key("\033OB", "next-history");
	e_bind_key("\033OC", "forward-char");
	e_bind_key("\033OD", "backward-char");
	e_bind_key("\05", "goto-end");
	e_bind_key("\01", "goto-begin");
	e_bind_key("\014", "redraw-line");
	e_bind_key("\013", "kill-to-end");
	e_bind_key("\031", "yank");

#ifdef COMPLETE
	e_bind_key("\011", "complete-file");
	/* set default completer */
	complete_lister = make_list;
#endif
    }
}

#ifndef __STDC__
void    e_resetterm()
#else
void    e_resetterm(void)
#endif

{

#ifdef TCSETA
    ioctl(TTYFD, TCSETA, &old_termio);
#else
    tcsetattr(TTYFD, TCSANOW, &old_termio);
#endif

    tty_reset = 0;
}

#ifndef __STDC__
void    e_error(str)
    char   *str;
#else
void    e_error(char *str)
#endif
{
    printf(str);
}

#ifndef __STDC__
void    e_bind_key(str, funn)
    char   *str;
    char   *funn;
#else
void    e_bind_key(char *str,
            char *funn)
#endif
{
    int     i;
    for (i = 0; funlist[i].name; i++)
	if (!strcmp(funlist[i].name, funn))
	    break;
    if (!funlist[i].name) {
	printf("ilegal-function '%s'\n", funn);
	return;
    }
    set_key(str, i);
}

#ifndef __STDC__
void    e_unbind_key(str)
    char   *str;
#else
void    e_unbind_key(char *str)
#endif
{
    int     i;
    switch (i = match_key(str, strlen(str))) {
      case -1:
      case 0:
	e_error("key not bound!\n");
	return;
    }
    free(comlist[i].str);
    comlist[i].com = 0;
    comlist[i].str = NULL;

/*    printf ("'%s' unbound\n", str);*/
}

#ifndef __STDC__
void    e_push_history(str)
    char   *str;
#else
void    e_push_history(char *str)
#endif
{
    int     i;
    char   *t = history[MAXHIST - 1];
    for (i = MAXHIST - 1; i > 0; i--)
	history[i] = history[i - 1];
    if (!t)
	t = (char *) malloc(strlen(str) + 1);
    else
	t = (char *) realloc(t, strlen(str) + 1);
    strcpy(t, str);
    history[0] = t;
}

#ifndef __STDC__
char   *e_gets(prompt)
    char   *prompt;
#else
char   *e_gets(char *prompt)
#endif
{
    int     exit = 0, i, match, matchlen;

    realpos = curpos = ind = 0;
    curhist = -1;
    strcpy(rivi, prompt);
    minpos = xpos = rind = strlen(prompt);
    e_setterm();
    outstr(prompt);
    bflush();
    while (!exit) {
	if (read(TTYFD, k_buf + ind, 1) != 1)
	    exit = 2;
	ind++;
	while (!exit && ind) {
	    match = match_key(k_buf, ind);
	    if (match == -1)
		/* several matches found - continue */
		break;

	    /* if key not found print it char by char (match = 0) */
	    matchlen = match ? strlen(comlist[match].str) : 1;
	    exit = (funlist[comlist[match].com].ptr) (k_buf, matchlen);

	    /* key found - remove it from buffer */
	    ind -= matchlen;
	    for (i = 0; i < ind; i++)
		k_buf[i] = k_buf[i + matchlen];
	}
    }
    e_resetterm();
    return ((exit == 2) ? 0 : rivi + minpos);	/* return EOF or rivi */
}

/* key functions */

#ifndef __STDC__
int     k_self_insert(buf, len)
    char   *buf;
    int     len;
#else
int     k_self_insert(char *buf, int len)
#endif
{
    /* make room for string */
    memmov(rivi + xpos + len, rivi + xpos, rind - xpos);
    rind += len;
    memcpy(rivi + xpos, buf, len);
    /* draw rest of line */
    drawline(xpos, rind, DRAW);
    xpos += len;
    goto_col(xpos);
    return (0);
}

#ifndef __STDC__
int     k_nop()
#else
int     k_nop(void)
#endif

{
    return (0);
}

#ifndef __STDC__
int     k_backspace()
#else
int     k_backspace(void)
#endif

{
    if (xpos > minpos) {
	int     i;
	backchar(rivi[xpos - 1]);
	for (i = xpos; i < rind; i++)
	    outchar(rivi[i - 1] = rivi[i]);
	outchar(' ');
	outchar(' ');
	backchar(' ');
	backchar(' ');
	goto_col(--xpos);
	rind--;
    } else
	beep();
    return (0);
}

#ifndef __STDC__
int     k_delete()
#else
int     k_delete(void)
#endif

{
    if (xpos < rind) {
	int     i = xpos + 1;
	while (i < rind)
	    outchar(rivi[i - 1] = rivi[i++]);
	outchar(' ');
	outchar(' ');
	backchar(' ');
	backchar(' ');
	goto_col(xpos);
	rind--;
    } else

#ifdef COMPLETE
	k_complete_list();
#else
	beep();
#endif

    return (0);
}

#ifndef __STDC__
int     k_eof()
#else
int     k_eof(void)
#endif

{
    if (rind == minpos) {
	bufout(0x0d);
	bufout(0x0a);
	bflush();
	return (2);
    }
    k_delete();
    return (0);
}

#ifndef __STDC__
int     k_eol()
#else
int     k_eol(void)
#endif

{
    rivi[rind] = '\0';
    bufout(0x0d);
    bufout(0x0a);
    bflush();
    return (1);
}

#ifndef __STDC__
int     k_prev_history()
#else
int     k_prev_history(void)
#endif

{
    if ((curhist < MAXHIST - 1) && history[curhist + 1]) {
	goto_col(minpos);
	drawline(minpos, rind, EMPTY);
	goto_col(minpos);
	get_history(++curhist);
	drawline(minpos, xpos, DRAW);
	goto_col(xpos);
    } else
	beep();
    return (0);
}

#ifndef __STDC__
int     k_next_history()
#else
int     k_next_history(void)
#endif

{
    if ((curhist > 0) && history[curhist - 1]) {
	goto_col(minpos);
	drawline(minpos, rind, EMPTY);
	goto_col(minpos);
	get_history(--curhist);
	drawline(minpos, xpos, DRAW);
	goto_col(xpos);
    } else
	beep();
    return (0);
}

#ifndef __STDC__
int     k_forward()
#else
int     k_forward(void)
#endif

{
    if (xpos < rind)
	goto_col(++xpos);
    return (0);
}

#ifndef __STDC__
int     k_backward()
#else
int     k_backward(void)
#endif

{
    if (xpos > minpos)
	goto_col(--xpos);
    return (0);
}

#ifndef __STDC__
int     k_end()
#else
int     k_end(void)
#endif

{
    goto_col(xpos = rind);
    return (0);
}

#ifndef __STDC__
int     k_begin()
#else
int     k_begin(void)
#endif

{
    goto_col(xpos = minpos);
    return (0);
}

#ifndef __STDC__
int     k_redraw()
#else
int     k_redraw(void)
#endif

{
    bufout(0x0d);
    bufout(0x0a);
    realpos = curpos = 0;
    drawline(0, rind, DRAW);
    goto_col(xpos);
    return (0);
}

#ifndef __STDC__
int     k_kill_end()
#else
int     k_kill_end(void)
#endif

{
    memcpy(yank, rivi + xpos, rind - xpos);
    yank[rind - xpos] = '\0';
    drawline(xpos, rind, EMPTY);
    goto_col(rind = xpos);
    return (0);
}

#ifndef __STDC__
int     k_kill_begin()
#else
int     k_kill_begin(void)
#endif

{
    int     len = xpos - minpos;
    goto_col(minpos);
    drawline(xpos, rind, DRAW);
    memcpy(yank, rivi + minpos, len);
    yank[len] = '\0';
    drawline(minpos, xpos, EMPTY);
    memmov(rivi + minpos, rivi + xpos, rind - xpos);
    rind -= len;
    goto_col(xpos = minpos);
    return (0);
}

#ifndef __STDC__
int     k_yank()
#else
int     k_yank(void)
#endif

{
    int     len = strlen(yank);
    /* make room for string */
    memmov(rivi + xpos + len, rivi + xpos, rind - xpos);
    memcpy(rivi + xpos, yank, len);
    /* draw rest of line */
    drawline(xpos, rind += len, DRAW);
    goto_col(xpos);
    return (0);
}

#ifdef COMPLETE
/* simple file completion */

int     sortcmp(a, b)
    char  **a, **b;
{
    return (strcmp(*a, *b));
}

char  **make_list(line, mode)
    char   *line;
    int     mode;
{
    struct comp_list {
	char   *name;
	struct comp_list *next;
    };
#define new_comp_list() (struct comp_list *)malloc(sizeof(struct comp_list));

    static struct comp_list *p = NULL;
    struct comp_list *q;
    struct dirent *dp;
    DIR    *dire;
    char  **res = NULL, *path = ".", *file = NULL;
    int     len = strlen(line), i, j, nelems = 0;

    for (i = len - 1; (i >= 0) && (line[i] != ' '); i--)
	if ((line[i] == '/') && (!file)) {
	    file = line + i + 1;
	    line[i] = '\0';
	}
    i++;

    /* slash found => directory other than . */
    if (file)
	path = line + i;
    else
	file = line + i;

    len = strlen(file);

    /* now we should have separated dir and file */

    if (!(dire = opendir(path)))
	printf("cannot open directory %s!\n", path);

    if (*file != '.') {
	readdir(dire);	/* don't list . and .. files unless we want em */
	readdir(dire);
    }
    if (!p) {
	p = new_comp_list();
	p -> next = (struct comp_list *) (p -> name = NULL);
    }
    q = p;

    while (dp = readdir(dire)) {
	if (!strncmp(dp -> d_name, file, len)) {

	    /* free previous name */
	    /* if (q->name) free (q->name); */

	    q -> name = xstrdup(dp -> d_name + len * mode);

	    nelems++;
	    if (!q -> next) {
		q -> next = new_comp_list();
		q -> next -> next = (struct comp_list *) NULL;
		q -> next -> name = (char *) NULL;
	    }
	    q = q -> next;
	}
    }
    closedir(dire);

    /* convert linked list to a matrix */
    if (nelems) {
	res = (char **) malloc(sizeof (char *) * (nelems + 1));
	res[nelems] = NULL;
	for (q = p, i = 0; i < nelems; i++, q = q -> next) {
	    res[i] = q -> name;
	}
	qsort(res, nelems, sizeof (char *), sortcmp);
    }
    if (nelems == 1) {
	DIR    *dp;
	int     clen = strlen(res[0]);
	char   *only = malloc(strlen(path) + len + clen + 2);

	sprintf(only, "%s/%s%s", path, file, res[0]);
	res[0] = realloc(res[0], clen + 2);
	res[0][clen] = (dp = opendir(only)) ? '/' : ' ';
	res[0][clen + 1] = '\0';
	if (dp)
	    closedir(dp);
    }
    return (res);
}

int     k_complete()
{
    int     i = 0, flag = 1, j;
    char    n, **list;

    if (!(list = complete_lister(strndup(rivi + minpos, xpos - minpos), 1))) {
	beep();
	return (0);
    }
    do {
	n = list[0][i];
	for (j = 1; list[j]; j++)
	    if (list[j][i] != n)
		flag = 0;
	i++;
    } while (flag && n);
/*    printf ("ok %d chars of %s\n", i -1 , list[0]);*/
    k_self_insert(list[0], i - 1);

    if (!flag)
	beep();

    /* free allocated memory */
    for (j = 0; list[j]; j++)
	free(list[j]);
    free(list);
    return (0);
}

int     k_complete_list()
{
    char  **list = make_list(strndup(rivi + minpos, xpos - minpos), 0);
    char   *p;
    int     width = 0, i, j, k;

    for (j = 0; list[j]; j++)
	if ((i = strlen(list[j])) > width)
	    width = i;
    bufout(0x0d);
    bufout(0x0a);
    realpos = 0;
    for (k = 0, j = 0; list[j]; j++) {
	for (i = 0, p = list[j]; i < width + 2; i++, k++)
	    if (*p)
		bufout(*p++);
	    else
		bufout(' ');

	if ((k + width > 78) || !list[j + 1]) {
	    bufout(0x0d);
	    bufout(0x0a);
	    bflush();
	    k = 0;
	}
	/* free allocated memory */
	free(list[j]);
    }
    free(list);
    k_redraw();
    return (0);
}
#endif

/* other functions */

#ifndef __STDC__
void    get_history(i)
    int     i;
#else
void    get_history(int i)
#endif
{
    strcpy(rivi + minpos, history[i]);
    xpos = rind = minpos + strlen(history[i]);
}

#ifndef __STDC__
int     charlen(ch)
    unsigned char ch;
#else
int     charlen(unsigned char ch)
#endif
{
    return ((ch < 0x20) || (ch > 0x7e) ? 2 : 1);
}

#ifndef __STDC__
int     goto_col(x)
    int     x;
#else
int     goto_col(int x)
#endif
{
    /* backward */
    while (x < curpos)
	backchar(rivi[curpos - 1]);
    /* forward */
    while ((x > curpos))
	outchar(rivi[curpos]);
    bflush();
    return (0);
}

#ifndef __STDC__
int     drawline(start, end, mode)
    int     start;
    int     end;
    int     mode;
#else
int     drawline(int start,
            int end,
            int mode)
#endif
{
    int     i, j;
    switch (mode) {
      case DRAW:
	for (i = start; i < end; i++)
	    outchar(rivi[i]);
	break;
      case EMPTY:
	for (i = start; i < end; i++)
	    for (j = charlen(rivi[i]); j; j--) {
		outchar(' ');
		realpos++;
	    }
	break;
      case COUNT:
	for (i = start, j = 0; i < end; i++)
	    j += charlen(rivi[i]);
	return (j);
    }
    return (0);
}

#ifndef __STDC__
void    outstr(str)
    char   *str;
#else
void    outstr(char *str)
#endif
{
    char   *p = str;
    while (*p)
	outchar(*p++);
}

#ifndef __STDC__
void    backchar(ch)
    char    ch;
#else
void    backchar(char ch)
#endif
{
    int     i = charlen(ch);
    curpos--;
    while (i--) {
	bufout(0x08);
	realpos--;
    }
}

#ifndef __STDC__
void    outchar(ch)
    unsigned char ch;
#else
void    outchar(unsigned char ch)
#endif
{
    curpos++;
    if (ch < 0x20) {
	bufout('^');
	ch |= 0x40;
	realpos++;
    }
    if (ch > 126) {
	bufout('^');
	ch = '?';
	realpos++;
    }
    bufout(ch);
    realpos++;
}

#ifndef __STDC__
void    rawout(ch)
    unsigned char ch;
#else
void    rawout(unsigned char ch)
#endif
{
    write(1, (char *) &ch, 1);
}

#ifndef __STDC__
void    bflush()
#else
void    bflush(void)
#endif

{
    if (outcount) {
	write(1, outbuf, outcount);
	outcount = 0;
    }
}

#ifndef __STDC__
void    set_ckey(ch, funn)
    unsigned char ch;
    char   *funn;
#else
void    set_ckey(unsigned char ch,
            char *funn)
#endif
{
    char    t[2];
    t[0] = ch;
    t[1] = '\0';
    e_bind_key(t, funn);
}

#ifndef __STDC__
int     match_key(str, len)
    char   *str;
    int     len;
#else
int     match_key(char *str,
            int len)
#endif
{
    int     match = 0, i;
    for (i = 0; i < MAXKEYS; i++)
	if (comlist[i].str && !strncmp(comlist[i].str, str, len))
	    match = match ? -1 : i;
    return (match);
}

#ifndef __STDC__
void    set_key(str, fun)
    char   *str;
    unsigned int fun;
#else
void    set_key(char *str,
            unsigned int fun)
#endif
{
    int     i;
    switch (i = match_key(str, strlen(str))) {
      case 0:
	for (i = 0; i < MAXKEYS; i++)
	    if (!comlist[i].str)
		break;
	if (i == MAXKEYS) {
	    e_error("not enough room!\n");
	    return;
	}
	break;
      case -1:
	e_error("impossible!\n");
	return;
      default:
	free(comlist[i].str);
    }
    comlist[i].com = fun;
    comlist[i].str = xstrdup(str);
/*    outchar ('\'');
    outstr (str);
    printf ("' bound to '%s'\n", funlist[fun].name);*/
}
#endif	       /* EDIT */

#ifdef MAIN
main()
{
    char   *p;
    while (p = e_gets("*>"))
	if (*p) {
	    e_push_history(p);
	    printf("\n<%s>\n", p);
	}
}
#endif
