/*
 * $Id: main.c,v 1.3 93/06/21 13:38:36 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	main.c,v $
 * Revision 1.3  93/06/21  13:38:36  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:20  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include "patchlevel.h"
#include "macro.h"
#include "extern.h"

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>

/* Profiling arrays. */

#ifdef DEBUG
int     rd[RAMSIZE];
int     wr[RAMSIZE];
long    instructions[300];
#endif

/* CPU internals */

ADDRESS program_counter;
BYTE    x_register, y_register, stack_pointer;
BYTE    accumulator, status_register;

int     zero_flag;
int     sign_flag;
int     overflow_flag;
int     break_flag;
int     decimal_flag;
int     interrupt_flag;
int     carry_flag;

CLOCK   clk;
CLOCK   next_int_clk;
CLOCK   int_clk[NUMOFINT];
int     next_int;

int     runflg = 0;
int     hexflg = 1;
int     tstflg = 0;
int     verflg = 0;
int     debugflg = 0;
int     autodump = 0;
int     module = 0;
int     reuflg = 0;

char   *ram_name = RAMNAME;
char   *basic_name = BASICNAME;
char   *kernel_name = KERNELNAME;
char   *progname;
char   *project_dir = PROJECTDIR;
char   *exrom_name = EXROMNAME;
#ifdef REU
char   *reufile_name = REUNAME;
#endif

#ifndef __STDC__
int     load_mem()
#else	       /* __STDC__ */
int     load_mem(void)
#endif	       /* __STDC__ */
{
    FILE    *fp;
    char    buffer[256];

#define LOAD_FILE(name,reloc,var,siz,ret) \
    sprintf(buffer, "%s/%s", project_dir, name); \
    if (NULL != ( fp = fopen( buffer, "r" ))) { \
	if (reloc != 0) { \
	     if(((unsigned int)fgetc(fp) + 256 * (unsigned int)fgetc(fp)) != reloc) { \
	       rewind(fp); /*(void)fseek(fp,(off_t)0, SEEK_SET);*/ }\
	else printf("Skipping load address %04X on %s\n", reloc, name); } \
	if (1 != fread((char *)var, siz, 1, fp )) { \
	return ( ret ); }} \
    else { \
	perror(name); \
	return ( ret ); } \
    (void) fclose( fp );

    LOAD_FILE(kernel_name, 0xe000, kernal_rom, ROMSIZE, -3);
    LOAD_FILE(basic_name, 0xa000, basic_rom, ROMSIZE, -2);
    LOAD_FILE(ram_name, 0, ram, RAMSIZE, 1);

    if (module) {
	external_rom = (BYTE *)malloc(EXROMSIZE<<1);
	assert(external_rom);
	LOAD_FILE(exrom_name, 0,
		  ((module == 2) ? external_rom + EXROMSIZE : external_rom),
		  ((module == 3) ? EXROMSIZE<<1 : EXROMSIZE), -1);
    }
    return (0);
}

#ifdef NEW_TIMER_CODE
/*
 * This is part kludge to get both timers usable
 */
static int new_alarm_number;

#ifndef __STDC__
void    new_alarm_trap(i)
    int     i;
#else
void    new_alarm_trap(int i)
#endif
{
    while (clk > next_int_clk)
	PC = intr(next_int, PC);
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
}

#ifndef __STDC__
void update_intr_alarm()
#else
void update_intr_alarm(void)
#endif
{
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
    find_next_alarm();
}

#endif /* NEW_TIMER_CODE */

#ifndef __STDC__
int     main(argc, argv)
    int     argc;
    char  **argv;
#else	       /* __STDC__ */
int     main(int argc,
            char **argv)
#endif	       /* __STDC__ */
{
    int     assflg = 0;
    int     errflg = 0;
    char   *c;
    char   *prg = NULL;
    char   *disk = NULL;
    int     i;

    /* Initialisation. */
    progname = argv[0];

    x_start(argc, argv);


  /***************  parsing arguments  ****************/

  while (!errflg && --argc && '-' == (*++argv)[0]) {
    c = argv[0]; if (*c == '-') c++;
    for (; *c != '\0'; c++)
      switch (*c) {

	  case 'A':
	    ++autodump;
	    break;
	  case 'a':
	    ++assflg;
	    break;
	  case 'g':
	    ++debugflg;
	    break;
	  case 'h':
	    hexflg = !hexflg;
	    break;
	  case 't':
	    ++tstflg;
	    break;
	  case 'v':
	    ++verflg;
	    break;

	  case 'f':
	    if (argc > 1) {
	      disk = argv[1];
	    ++argv; --argc;
	    }
	    else
		++errflg;
	    break;

	  case 'e':
	    if (argc > 1) {
	      exrom_name = argv[1];
	    ++argv; --argc;
	    }
	    else
		++errflg;
	    break;
	  case 'm':
	    if (argc > 1) {
	      if ((module = atoi(argv[1])) < 1 || module > 3) errflg++;
	    ++argv; --argc;
	    }
	    break;

	  case 'l':
	    if (argc > 1) {
	      prg = argv[1];
	    ++argv; --argc;
	    }
	    else
		++errflg;
	    break;

	  case 'p':
	    if (argc > 1) {
	      project_dir = argv[1];
	    ++argv; --argc;
	    }
	    else
		++errflg;
	    break;
#ifdef REU
	  case 'r':
	    ++reuflg;
	    if (argc > 1 && *argv[1] != '-') {
	      reufile_name = argv[1]; /* by deafault: reu */
	    ++argv; --argc;
	    }
	    break;
#endif

	  default:
	    ++errflg;
	    break;
	}
  }

    if (errflg) {
	(void) fprintf(stderr, "\n Usage:  %s [-A] [-a] [-g] [-h] [-t] [-v]\n\
\t     [-f disk] [-e exrom] [-m #] [-l program] [-p path] %s[address]\n\n",
	  progname,
#ifdef REU
	  "[-r reufile] ");
#else
	  "");
#endif
	(void) fprintf(stderr, "\
\t-A\tdump ram on exit\n\
\t-a\tassembler/monitor\n\
\t-e\texromname for -m option\n\
\t-f\tattach floppy image\n\
\t-g\tdebug info\n\
\t-h\ttoggle decimal/hexadecimal mode (default hex)\n\
\t-l\tload program from filesystem\n\
\t-m\tload module into slot 1..3\n\
\t-p\tproject directory\n");
#ifdef REU
	(void) fprintf(stderr, "\t-r\tactivate ram expansion using reufile\n");
#endif
	(void) fprintf(stderr, "\
\t-t\ttest (monitor)\n\
\t-v\tverbose\n\n");
	exit(1);
    }

    printf ("\nWelcome to x64-%3.1f.%d, the Commodore 64 Emulator for XWindows\n", 
	    (float)VERSION, (int)PATCHLEVEL);

    switch (load_mem()) {
      case 1:
        fprintf(stderr,"Warning, RAM image not loaded\n");
	if (module) {
	    fprintf(stderr, "Cannot use module without ram image. Please dump ram.\n");
	    module = 0;
	}
	break;
      case -1:
	fprintf(stderr,"Couldn't load module. Ignoring faulty virtual cartridge.\n");
	module = 0;
	break;
      case -2:
	fprintf(stderr,"Couldn't load basic\n");
	       fprintf(stderr,"Warning, RAM image not loaded\n");if (!assflg)
	    exit(1);
	break;
      case -3:
	fprintf(stderr,"Couldn't load kernel.\n");
	if (!assflg)
	    exit(1);
    }

    if (prg) {
	FILE   *fp;
	ADDRESS start, end;

	printf("Loading program %s\n", prg);
	if (NULL == (fp = fopen(prg, "r"))) {
	    printf("Program '%s' loading failed.\n", prg);
	    perror(prg);
	} else {
	    start = fgetc(fp);
	    start += ((fgetc(fp)) << 8);
	    printf("Loading '%s' from %.4x, ", prg, start);
	    end = fread((char *) ram + start, 1, RAMSIZE, fp);
	    printf("%x bytes\n", end);
	    end += start;
	    ram[45] = (BYTE) end & 0xff;
	    ram[46] = (BYTE) (end >> 8) & 0xff;
	    (void) fclose(fp);
	}
    }

    printf("\nTesting configuration dependent code\n");
#ifdef REU
    if (reuflg)
      reset_reuc(reufile_name);
#endif

    for (i = 0; i < NUMOFINT; i++)
	int_clk[i] = 0;

    int_clk[I_SPECX] = 1;

    /* define start address */
    PC = intr(I_RESET, 0);

    if (argc)
	PC = (ADDRESS) strtol(argv[0], NULL, hexflg ? 16 : 10);

    initialize_traps();
    initialize_serial();
    initialize_1541();
    if (disk)
      attach_floppy_image(disk);

    if (assflg)
	mon(PC);

#ifdef NEW_TIMER_CODE
    new_alarm_number = new_alarm();
    update_intr_alarm();
#endif /* NEW_TIMER_CODE */

    printf("\nStarting Commodore 64 emulator with PC = %04X\n", PC);
    mainloop();
    exit(0);   /* never reached */
}

#ifndef HACK2

#ifndef __STDC__
void    mainloop()
#else	       /* __STDC__ */
void    mainloop(void)
#endif	       /* __STDC__ */
{

#ifndef HACK
    register unsigned int src = 0, result;
    register BYTE p1 = 0;
    register ADDRESS p2 = 0;
#endif
    register struct lookup_tag *op;

    SET_SR(0x20);
    SP = 0;

    while (1) {
	/* serve interrupts */
#ifdef NEW_TIMER_CODE
        if (clk >= next_alarm_clk)
            alarm_handler();
#else
	while (clk > next_int_clk)
	    PC = intr(next_int, PC);
#endif
	
#ifdef DEBUG
	instructions[LOAD(PC)]++;
	if (debugflg) {
	    show();
	}
#endif

	op = &lookup[LOAD(PC)];

#ifdef DEBUG
	if (debugflg)
	  printf(".%.4X\t%s\n", PC, sprint_opcode(PC, 1));
#endif
   
#ifdef HACK
	op -> funktio();

#else
	clk += op -> cycles;

	switch (clength[op -> addr_mode]) {
	  case 1:
	    PC++;
	    break;
	  case 2:
	    p1 = LOAD(PC + 1);
	    PC += 2;
	    break;
	  case 3:
	    p2 = LOAD(PC + 1) | (LOAD(PC + 2) << 8);
	    PC += 3;
	    break;
	  default:
	    printf("AAARGH\n");
	    end64();
	}

	switch (op -> source) {
	  case M_NONE:
	    break;
	  case M_AC:
	    src = AC;
	    break;
	  case M_XR:
	    src = XR;
	    break;
	  case M_YR:
	    src = YR;
	    break;
	  case M_SP:
	    src = SP;
	    break;
	  case M_SR:
	    src = GET_SR();
	    break;
	  case M_IMM:
	    src = p1;
	    break;
	  case M_ZERO:
	    src = LOAD_ZERO(p1);
	    break;
	  case M_ZERX:
	    src = LOAD_ZERO(p1 + XR);
	    break;
	  case M_ZERY:
	    src = LOAD_ZERO(p1 + YR);
	    break;
	  case M_ABS:
	    src = LOAD(p2);
	    break;
	  case M_ABSX:
	    src = LOAD(p2 + XR);
	    break;
	  case M_ABSY:
	    src = LOAD(p2 + YR);
	    break;
	  case M_ADDR:
	    src = p2;
	    break;
	  case M_IND:
	    src = LOAD_ADDR(p2);
	    break;
	  case M_INDX:
	    src = LOAD(LOAD_ZERO_ADDR(p1 + XR));
	    break;
	  case M_INDY:
	    src = LOAD(LOAD_ZERO_ADDR(p1) + YR);
	    break;
	  case M_REL:
	    src = p1;
	    break;
	  case M_FC:
	    src = CARRY;
	    break;
	  case M_FD:
	    src = DECIMAL;
	    break;
	  case M_FI:
	    src = INTERRUPT;
	    break;
	  case M_FV:
	    src = OVERFLOW;
	    break;

#ifndef NO_UNDOC_CMDS
	    /* These are used by undocumented commands only */
	  case M_ACIM:
	    src = (AC & p1);
	    break;
	  case M_ANXR:
	    src = (AC & XR);
	    break;
	  case M_AXIM:
	    src = ((AC | 0xee) & XR & p1);
	    break;
	  case M_SABY:
	    src = (SP & LOAD(p2 + YR));
	    break;
#endif

	  default:
	    printf("CPU: Internal bus error\n");
	    end64();
	}

#ifdef DEBUG
	if (debugflg) {
	    printf("src:%d(%d), dst:%d length:%d p1:%d, p2:%d ",
		src, op -> source, op -> destination, clength[op -> addr_mode],
		p1, p2);
	    fflush(stdout);
	    result = (op -> funktio) (src);
	    printf("result %x\n", result);
	} else
#endif

	    result = op -> funktio(src, AC);

	switch (op -> destination) {
	  case M_NONE:
	    break;
	  case M_AC:
	    AC = result;
	    break;
	  case M_XR:
	    XR = result;
	    break;
	  case M_YR:
	    YR = result;
	    break;
	  case M_SP:
	    SP = result;
	    break;
	  case M_SR:
	    SET_SR(result);
	    break;
	  case M_PC:
	    PC = result;
	    break;
	  case M_ZERO:
	    STORE_ZERO(p1, result);
	    break;
	  case M_ZERX:
	    STORE_ZERO(p1 + XR, result);
	    break;
	  case M_ZERY:
	    STORE_ZERO(p1 + YR, result);
	    break;
	  case M_ABS:
	    STORE(p2, result);
	    break;
	  case M_ABSX:
	    STORE(p2 + XR, result);
	    break;
	  case M_ABSY:
	    STORE(p2 + YR, result);
	    break;
	  case M_INDY:
	    STORE(LOAD_ZERO_ADDR(p1) + YR, result);
	    break;
	  case M_INDX:
	    STORE(LOAD_ZERO_ADDR(p1 + XR), result);
	    break;

	  case M_FC:
	    SET_CARRY(result);
	    break;
	  case M_FD:
	    SET_DECIMAL(result);
	    break;
	  case M_FI:
	    SET_INTERRUPT(result);
	    break;
	  case M_FV:
	    SET_OVERFLOW(result);
	    break;

#ifndef NO_UNDOC_CMDS
	    /* These are used by undocumented commands only */
	  case M_ACNC:
	    AC = result;
	    SET_CARRY(IF_SIGN());
	    break;
	  case M_ACXR:
	    AC = XR = result;
	    break;
	  case M_ACXS:
	    AC = XR = SP = result;
	    break;
	  case M_STH0:
	    u_stoshr(src, LOAD_ZERO_ADDR(p1), YR);
	    break;
	  case M_STH2:
	    u_stoshr(src, p2, XR);
	    break;
	  case M_STH3:
	    SP = src;	/* SHS */
	  case M_STH1:
	    u_stoshr(src, p2, YR);
	    break;
#endif

	  default:
	    end64();
	}
#endif
    }
}

#endif   /* HACK2 */

#ifndef NO_UNDOC_CMDS

#ifndef __STDC__
void    u_stoshr(val, address, index)
    unsigned int val;
    ADDRESS address;
    BYTE    index;
#else	       /* __STDC__ */
void    u_stoshr(unsigned int val, ADDRESS address, BYTE index)
#endif	       /* __STDC__ */
{
    val &= (((address >> 8) + 1) & 0xff);
    if (((address & 0xff) + index) > 0xff)
	address = val;
    STORE((address + index), val);
}
#endif

#ifndef __STDC__
void    end64()
#else	       /* __STDC__ */
void    end64(void)
#endif	       /* __STDC__ */
{
    x_exit();

#ifdef REU
    if (reuflg)
      close_reuc(reufile_name);
#endif

    if (autodump)
      dump ("ram");

#ifdef DEBUG
    /* if (debugflg) */
    {
	int     i;
	long    total = 0;
	char    buffer[256];
	FILE   *fp;

	sprintf(buffer, "%s/inst", project_dir);
	if (NULL == (fp = fopen(buffer, "w")))
	    return;

	for (i = 0; i < 256; i++) {
	    total += instructions[i];
	    fprintf(fp, "%02X %-4s %8ld\n", i, lookup[i].mnemonic,
		instructions[i]);
	}
	fprintf(fp, "----------------\n\tTotal %8ld\n", total);
	fclose(fp);

	/* Store C64 log file */
	/* if ( NULL == ( fp = fopen( "mem", "w" ))) return 1; for ( i=0;
	 * i<65536; i++ ) fprintf( fp, "%04X %010d %010d\n", i, rd[i], wr[i]
	 * ); fclose( fp );
	 */
    }
#endif

    exit(0);
}
