/*
 * $Id: memory.c,v 1.3 93/06/21 13:38:39 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * this file contains cpu <> memory <> io interface
 * memory is cutted into banks which are of $1000 (4k) lenght
 * memwr and memrd are tables for memory access functions
 * screen drawing is done by putting hook into a bank
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	memory.c,v $
 * Revision 1.3  93/06/21  13:38:39  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:23  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include "macro.h"
#include "extern.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* global */

BYTE    basic_rom[ROMSIZE];
BYTE    kernal_rom[ROMSIZE];
BYTE   *external_rom = 0;
BYTE    ram[RAMSIZE];
BYTE    color_ram[1024];
BYTE    sid[32];

/* function tables to access memory */

#ifdef __STDC__
void    (*(memwr[17])) (ADDRESS, BYTE);
BYTE(*(memrd[17])) (ADDRESS);

/* local functions */
static void ramwr(ADDRESS, BYTE);
#ifdef REU
static void ramwr_hi(ADDRESS, BYTE);
#endif
static void zerowr(ADDRESS, BYTE);
static void iowr(ADDRESS, BYTE);

static BYTE ramrd(ADDRESS);
static BYTE basicrd(ADDRESS);
static BYTE kernelrd(ADDRESS);
static BYTE exromrd(ADDRESS);
static BYTE charrd(ADDRESS);
static BYTE iord(ADDRESS);
#else

void    (*(memwr[17])) ();
BYTE(*(memrd[17])) ();

/* local functions */

static void ramwr();
#ifdef REU
static void ramwr_hi();
#endif
static void zerowr();
static void iowr();

static BYTE ramrd();
static BYTE basicrd();
static BYTE kernelrd();
static BYTE exromrd();
static BYTE charrd();
static BYTE iord();
#endif

/* ------------------------------------------------------------------------- */


#ifndef __STDC__
void    init_memmaps()
#else	       /* __STDC__ */
void    init_memmaps(void)
#endif	       /* __STDC__ */
{
    int     i;

    for (i = 0; i < 16; i++) {
	memwr[i] = ramwr;
	memrd[i] = ramrd;
    }
#ifdef REU
    if (reuflg)
	memwr[0x0f] = ramwr_hi;
#endif
    memwr[0x00] = zerowr;

    set_video_hooks();

    /*
     * Note: ULTIMAX mode (configurations 1, 5, 9 and D) is not supported
     */
    switch (ram[1] & 0x07) {
      case 7:  /* C E F */
	switch (module) {
	  case 1:
	    memrd[0x8] = memrd[0x9] = exromrd;

	  case 0:
	    memrd[0xa] = memrd[0xb] = basicrd;
	    break;

	  case 3:
	    memrd[0x8] = memrd[0x9] = exromrd;
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 6:  /* 4 6 7 */
	memrd[0xe] = memrd[0xf] = kernelrd;

	if (module == 2) {
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 5:  /* 8 A B */
	memrd[0x0d] = iord;
	memwr[0x0d] = iowr;

      case 4:  /* 0 2 3 */
	break;

      case 3:
	switch (module) {
	  case 1:
	    memrd[0x8] = memrd[0x9] = exromrd;

	  case 0:
	memrd[0xa] = memrd[0xb] = basicrd;
	    break;

	  case 3:
	    memrd[0x8] = memrd[0x9] = exromrd;
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 2:
	memrd[0xe] = memrd[0xf] = kernelrd;

	if (module == 2) {
	    memrd[0xa] = memrd[0xb] = exromrd;
	}
	memrd[0x0d] = charrd;
	break;

      case 1: /* the chargen is not accessible */
      case 0:
	break;
    }

    memrd[0x10] = memrd[0];
    memwr[0x10] = memwr[0];
}

/* memory access functions */

#ifndef __STDC__
static void ramwr(address, byte)
    ADDRESS address;
    BYTE    byte;
#else	       /* __STDC__ */
static void ramwr(ADDRESS address, BYTE byte)
#endif	       /* __STDC__ */
{
    ram[address] = byte;
}

#ifdef REU
#ifndef __STDC__
static void ramwr_hi(address, byte)
    ADDRESS address;
    BYTE    byte;
#else	       /* __STDC__ */
static void ramwr_hi(ADDRESS address, BYTE byte)
#endif	       /* __STDC__ */
{
    ram[address] = byte;
    if (address == 0xff00)  /* do not test every write */
      reu_dma(-1);
}
#endif

#ifndef __STDC__
static BYTE ramrd(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE ramrd(ADDRESS address)
#endif	       /* __STDC__ */
{
    return (ram[address]);
}

#ifndef __STDC__
static BYTE basicrd(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE basicrd(ADDRESS address)
#endif	       /* __STDC__ */
{
    return (basic_rom[address & 0x1fff]);
}

#ifndef __STDC__
static BYTE kernelrd(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE kernelrd(ADDRESS address)
#endif	       /* __STDC__ */
{
    return (kernal_rom[address & 0x1fff]);
}

#ifndef __STDC__
static BYTE exromrd(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE exromrd(ADDRESS address)
#endif	       /* __STDC__ */
{
    return (external_rom[address & 0x3fff]);
}

#ifndef __STDC__
static BYTE charrd(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE charrd(ADDRESS address)
#endif	       /* __STDC__ */
{
    return (char_rom[address & 0xfff]);
}

#ifndef __STDC__
static void iowr(address, byte)
    ADDRESS address;
    BYTE    byte;
#else	       /* __STDC__ */
static void iowr(ADDRESS address, BYTE byte)
#endif	       /* __STDC__ */
{
    switch ((address & 0xf00) >> 8) {
	case 0:case 1:case 2:case 3:	/* VIC */
	store_vic(address & 0x3f, byte);
	return;

      case 4:
      case 5:
      case 6:
      case 7: /* SID */
	sid[address & 0x1f] = byte;
	/* store sid */
	return;

      case 8:
      case 9:
      case 10:
      case 11:/* COLOR RAM */
	/* do not draw char unless its colour is changed */
	address &= 0x3ff;
	if ((byte & 0xf) != (color_ram[address] & 0xf)) {
	    color_ram[address] = byte;
	    if (address < 1000 &&
		(vmode == V_TEXT || vmode == V_ECTEXT || vmode == V_MTEXT))
		x_draw_screen(address);
	} else
	    color_ram[address] = byte;
	return;

      case 12:/* CIA 1 */
	store_cia1(address & 0xf, byte);
	return;

      case 13:/* CIA 2 */
	store_cia2(address & 0xf, byte);
	return;
#ifdef REU
      case 15: /* IO2 */
	if (reuflg)
	  store_reuc(address & 0xf, byte);
	return;
#endif
      default:
	return;
    }
}


#ifndef __STDC__
static BYTE iord(address)
    ADDRESS address;
#else	       /* __STDC__ */
static BYTE iord(ADDRESS address)
#endif	       /* __STDC__ */
{
    switch ((address & 0xf00) >> 8) {
	case 0:case 1:case 2:case 3:	/* VIC */
	return (read_vic(address & 0x3f));

      case 4:
      case 5:
      case 6:
      case 7: /* SID */
	if ((address & 0x1f) == 0x1b)
	    return (rand());
	return (sid[address & 0x1f]);

      case 8:
      case 9:
      case 10:
      case 11:/* COLOR RAM */
	return (color_ram[address & 0x3ff]);

      case 12:/* CIA 1  dc00 */
	return (read_cia1(address & 0xf));

      case 13:/* CIA 2  dd00 */
	return (read_cia2(address & 0xf));
#ifdef REU
      case 15: /* IO2   df00 */
	if (reuflg)
	  return (read_reuc(address & 0xf));  /* else drop trough */
#endif
      default:
	return (0xff);
    }
}

#ifndef __STDC__
static void zerowr(address, byte)
    ADDRESS address;
    BYTE    byte;
#else	       /* __STDC__ */
static void zerowr(ADDRESS address, BYTE byte)
#endif	       /* __STDC__ */
{
    static BYTE ram1;
    if (address == 1) {
	if ((ram[1] = (0xff & ~ram[0]) | (byte & ram[0])) != ram1) {
	    init_memmaps();
	    ram1 = ram[1];
	}
    } else
	ram[address] = byte;
}

/* ------------------------------------------------------------------------- */

/* interrupts */

#ifndef __STDC__
ADDRESS intr(sig, pc)
    int     sig;
    ADDRESS pc;
#else	       /* __STDC__ */
ADDRESS intr(int sig, ADDRESS pc)
#endif	       /* __STDC__ */
{
    BYTE    t;

    switch (sig) {
      case I_SPECX:
	/* special "interrupt" to handle x events */
	break;

      case I_RESET:
	ram[0] = 0xff;
	ram[1] = 0xff;
	vic[0x18] = 0x15;
	cia2[2] = 0;
	cia2[0] = 255;
	clk = 0;
	init_memmaps();
	if (debugflg)
	    printf("RESET\n");
	pc = LOAD_ADDR(0xFFFC);	/* Jump using RESET vector. */
	break;

      case I_BRK:
	SET_BREAK(1);	/* Set break status. */
	pc--;
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);	/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR(65534);	/* Jump using BRK vector (0xfffe). */
	return (pc);

      case I_LIGHT_PEN:
      case I_SS_COL:
      case I_SB_COL:
      case I_RASTER:
      case I_CIA1FL:
      case I_CIA1SP:
      case I_CIA1TOD:
      case I_CIA1TA:
      case I_CIA1TB:	/* check if IRQ iterrupts are not masked */
	if (IF_INTERRUPT())
	    break;

	SET_BREAK(0);
	pc--;
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);	/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR(65534);	/* Jump using IRQ vector (0xfffe). */
	break;

      case I_CIA2FL:
      case I_CIA2SP:
      case I_CIA2TOD:
      case I_CIA2TA:
      case I_CIA2TB:
      case I_RESTORE:
	SET_BREAK(0);
	pc--;
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);	/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR(0xFFFA);	/* Jump using NMI vector. */
	break;
    }

    /*
     * we must handle interrupts allthough they are not served eg. interrupts
     * are disabled, so that timers are updated
     */

    int_clk[sig] = 0;

    switch (sig) {
      case I_SPECX:
	int_clk[sig] = clk + CYCLES_PER_RFSH;
	x_loop();
	pc = PC;	/* PC may be changed in x_loop() or mon() */
	break;

      case I_RASTER:
	int_raster();
	break;

      case I_CIA1TB:
	printf("cia1tbint\n");
	int_cia1tb();
	break;

      case I_CIA1TA:
	int_cia1ta();
	break;

      case I_CIA2TB:
	printf("cia2tbint\n");
	int_cia2tb();
	break;

      case I_CIA2TA:
	printf("cia2taint\n");
	/* int_cia2ta(); */
	break;

      case I_LIGHT_PEN:
      case I_SS_COL:
      case I_SB_COL:

      case I_CIA1FL:
      case I_CIA1SP:

      case I_CIA2FL:
      case I_CIA2SP:

	printf("unsupported interrupt %d\n", sig);
	show();
	end64();
    }

    find_next_int();
    return (pc);
}

/* find next interrupt */

#ifndef __STDC__
void    find_next_int()
#else	       /* __STDC__ */
void    find_next_int(void)
#endif	       /* __STDC__ */
{
    int     i;

    for (i = 0, next_int_clk = (CLOCK) - 1; i < NUMOFINT; i++)
	if (int_clk[i] && int_clk[i] < next_int_clk) {
	    next_int_clk = int_clk[i];
	    next_int = i;
	}
    if (debugflg && next_int != I_SPECX)
	printf("%d next int %d %ld\n", clk, next_int, next_int_clk);
}

/*
 * This routine is used to skip idle on ROM busy loop
 */
void jump_to_next_alarm()
{
    clk = next_int_clk;
}

#ifndef __STDC__
void    restore()
#else	       /* __STDC__ */
void    restore(void)
#endif	       /* __STDC__ */
{
    int_clk[I_RESTORE] = clk;
    find_next_int();
#ifdef NEW_TIMER_CODE
    update_intr_alarm();
#endif
}


#ifndef __STDC__
BYTE load_rom(a)
ADDRESS a;
#else	       /* __STDC__ */
BYTE load_rom(ADDRESS a)
#endif	       /* __STDC__ */
{
    switch(a & 0xF000)
    {
    case 0xA000:
    case 0xB000:
	return basic_rom[a & 0x1FFF];
    case 0xE000:
    case 0xF000:
	return kernal_rom[a & 0x1FFF];
    }
    return 0;
}

#ifndef __STDC__
void save_rom(a, x)
ADDRESS a;
BYTE x;
#else	       /* __STDC__ */
void save_rom(ADDRESS a, BYTE x)
#endif	       /* __STDC__ */
{
    switch(a & 0xF000)
    {
    case 0xA000:
    case 0xB000:
	basic_rom[a & 0x1FFF] = x;
	break;
    case 0xE000:
    case 0xF000:
	kernal_rom[a & 0x1FFF] = x;
	break;
    }
}
