/*
 * $Id: misc.c,v 1.3 93/06/21 13:38:42 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains misc funtions to help debugging.
 * Included are:
 *	void sprint_opcode (ADDRESS pc, int mode)
 *	void sprint_binary (BYTE code)
 *
 * sprint_opcode returns mnemonic code of machine instruction.
 * sprint_binary returns binary form of given code (8bit)
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	misc.c,v $
 * Revision 1.3  93/06/21  13:38:42  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:26  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "macro.h"
#include "extern.h"

/*
 * shw
 *
 * show prints PSW and contents of registers.
 */

#ifndef __STDC__
void    show()
#else	       /* __STDC__ */
void    show(void)
#endif	       /* __STDC__ */
{
    if (verflg)
	printf(hexflg ?
	    "PC=%4X AC=%2X XR=%2X YR=%2X PF=%s SP=%2X %3X %3d %s\n" :
	    "PC=%04d AC=%03d XR=%03d YR=%03d PF=%s SP=%03d %3d %2X %s\n",
	    (int) PC, (int) AC, (int) XR, (int) YR,
	    sprint_status(), (int) SP,
	    LOAD(PC), LOAD(PC), sprint_opcode(PC, hexflg));
    else
	printf(hexflg ? "%lx %4X %s\n" : "%ld %4d %s\n",
	    clk, PC, sprint_opcode(PC, hexflg));
}


/* sprint_opcode parameters:
 *
 *	string		the name of the machine instruction
 *	addr_mode	# describing used addressing mode, see "vmachine.h"
 *	base		if 1==base => HEX, 0==base => DEC, see code
 *	opcode		address of memory where machine instruction
 *                      and argument are. First byte is unused, because
 *                      it is the machine code and it is already known -
 *                      string!
 */

#ifndef __STDC__
char   *sprint_opcode(counter, base)
    ADDRESS counter;
    int     base;
#else	       /* __STDC__ */
char   *sprint_opcode(ADDRESS counter, int base)
#endif	       /* __STDC__ */
{
    char   *string;
    int     addr_mode;
    char   *buffp;
    static char buff[200];
    int     ival;
    BYTE    x = LOAD(counter);
    BYTE    p1 = LOAD(counter + 1);
    BYTE    p2 = LOAD(counter + 2);

#define HEX 1
    buffp = buff;
    string = lookup[x].mnemonic;
    addr_mode = lookup[x].addr_mode;

    sprintf(buff, "%s", string);	/* Print opcode. */
    while (*++buffp);

    switch (addr_mode) {
	/*
	 * Bits 0 and 1 are usual marks for X and Y indexed addresses, i.e.
	 * if  bit #0 is set addressing mode is X indexed something and if
	 * bit #1 is set addressing mode is Y indexed something. This is not
	 * from MOS6502, but convention used in this program. See
	 * "vmachine.h" for details.
	 */

	/* Print arguments of the machine instruction. */

      case IMPLIED:
	break;

      case IMMEDIATE:
	ival = p1;
	if (base & HEX)
	    sprintf(buffp, " #$%02X", ival);
	else
	    sprintf(buffp, " %3d", ival);
	break;

      case ZERO_PAGE_X:
	ival = p1 & 0xFF;
	if (base & HEX)
	    sprintf(buffp, " $%02X,X", ival);
	else
	    sprintf(buffp, " %3d,X", ival);
	while (*++buffp);
	break;

      case ZERO_PAGE:
	ival = p1 & 0xFF;
	if (base & HEX)
	    sprintf(buffp, " $%02X", ival);
	else
	    sprintf(buffp, " %3d", ival);
	while (*++buffp);
	break;

      case ABSOLUTE_X:
	ival = p1 & 0xFF;
	ival += (p2 & 0xFF) * 256;
	if (base & HEX)
	    sprintf(buffp, " $%04X,X", ival);
	else
	    sprintf(buffp, " %5d,X", ival);
	while (*++buffp);
	break;

      case ABSOLUTE_Y:
	ival = p1 & 0xFF;
	ival += (p2 & 0xFF) * 256;
	if (base & HEX)
	    sprintf(buffp, " $%04X,Y", ival);
	else
	    sprintf(buffp, " %5d,Y", ival);
	while (*++buffp);
	break;

      case ABSOLUTE:
	ival = p1 & 0xFF;
	ival += (p2 & 0xFF) * 256;
	if (base & HEX)
	    sprintf(buffp, " $%04X", ival);
	else
	    sprintf(buffp, " %5d", ival);
	while (*++buffp);
	break;

      case INDIRECT_X:
	ival = p1 & 0xFF;
	if (base & HEX)
	    sprintf(buffp, " ($%02X,X)", ival);
	else
	    sprintf(buffp, " (%3d,X)", ival);
	while (*++buffp);
	break;

      case INDIRECT_Y:
	ival = p1 & 0xFF;
	if (base & HEX)
	    sprintf(buffp, " ($%02X),Y", ival);
	else
	    sprintf(buffp, " (%3d),Y", ival);
	while (*++buffp);
	break;

      case INDIRECT:
	ival = p1 & 0xFF;
	ival += (p2 & 0xFF) * 256;
	if (base & HEX)
	    sprintf(buffp, " ($%04X)", ival);
	else
	    sprintf(buffp, " (%5d)", ival);
	while (*++buffp);
	break;

      case RELATIVE:
	ival = p1 & 0xFF;
	if (0x80 & ival)
	    ival -= 256;
	ival += counter;
	ival += 2;
	if (base & HEX)
	    sprintf(buffp, " $%04X", ival);
	else
	    sprintf(buffp, " %5d", ival);
	while (*++buffp);
	break;

      case ACCUMULATOR:
	sprintf(buffp, " A");
	while (*++buffp);
	break;

    }

    return buff;
}

#ifndef __STDC__
void    print_stack(SP)
    BYTE    SP;
#else	       /* __STDC__ */
void    print_stack(BYTE SP)
#endif	       /* __STDC__ */
{
    int     i;

    printf("Stack: ");
    for (i = 0x100 + SP; i < 0x200; i += 2)
	printf("%04X  ", ram[i + 1] | ram[i] << 8);
    printf("\n");
}

/* sprint_binary parameter:
 *
 *	code		# which will be converted to binary
 */

#ifdef __STDC__
char   *sprint_binary(BYTE code)
#else

#ifndef __STDC__
char   *sprint_binary(code)
    BYTE    code;
#else	       /* __STDC__ */
char   *sprint_binary(BYTE code)
#endif	       /* __STDC__ */

#endif

#ifdef __STDC__
#endif	       /* __STDC__ */
{
    static char bin[9];
    int     i;

    bin[8] = 0;/* Leading zero. */

    for (i = 7; i >= 0; --i) {
	bin[i] = code % 2 ? '1' : '0';
	code >>= 1;
    }

    return bin;
}

#ifndef __STDC__
char   *xstrdup(str)
    char   *str;
#else
char   *xstrdup(char *str)
#endif
{
    char   *t = (char *) malloc(strlen(str) + 1);
    strcpy(t, str);
    return (t);
}

#ifndef __STDC__
char   *strndup(str, n)
    char   *str;
    int     n;
#else
char   *strndup(char *str, int n)
#endif
{
    char   *t = (char *) malloc(n + 1);
    strncpy(t, str, n);
    t[n] = '\0';
    return (t);
}

#ifndef __STDC__
void    memmov(target, source, length)
    char   *target;
    char   *source;
    unsigned int length;
#else
void    memmov(char *target, char *source, unsigned int length)
#endif
{
    if (target > source) {
	target += length;
	source += length;
	while (length--)
	    *--target = *--source;
    } else if (target < source) {
	while (length--)
	    *target++ = *source++;
    }
}

