/*
 * $Id: mon.c,v 1.3 93/06/21 13:38:48 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains ml monitor/debugger
 * for c64
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * Patches by
 *   Frank Prindle  (Frank.Prindle@lambada.oit.unc.edu)  /FCP/
 *   Teemu Rantanen (tvr@cs.hut.fi) /TVR/
 *
 * $Log:	mon.c,v $
 * Revision 1.3  93/06/21  13:38:48  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:30  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "macro.h"
#include "asm.h"
#include "extern.h"


static ADDRESS address;	/* Default place to put code in Commodore 64. */

/*
 *  Machine Language Monitor
 */


#ifdef __STDC__
static void pack_args(int start);
static int mon_fill(char *line);
static int mon_hunt(char *line);
static int mon_compare(char *line);
static int mon_move(char *line);
static int mon_asm(char *line);
static int mon_quit(char *line);
static int mon_exit(char *line);
static int mon_jump(char *line);
static int mon_dump(char *line);
static int mon_undump(char *line);
static int mon_floppy(char *line); /*TVR*/
static int mon_save(char *line);
static int mon_load(char *line);
static int mon_setreg(char *line);
static int mon_showreg(char *line);
static int mon_disassemble(char *line);
static int mon_memdump(char *line);
static int mon_bitdump(char *line);
static int mon_memstore(char *line);
static int mon_help(char *line);
#ifdef NEW_TIMER_CODE
static int mon_break(char *line);
static int mon_clear(char *line);
#endif

#else	       /* __STDC__ */

static void pack_args( /* int start */ );
static int mon_fill( /* char *line */ );
static int mon_hunt( /* char *line */ );
static int mon_compare( /* char *line */ );
static int mon_move( /* char *line */ );
static int mon_asm( /* char *line */ );
static int mon_quit( /* char *line */ );
static int mon_exit( /* char *line */ );
static int mon_jump( /* char *line */ );
static int mon_dump( /* char *line */ );
static int mon_undump( /* char *line */ );
static int mon_floppy( /* char *line */ ); /*TVR*/
static int mon_save( /* char *line */ );
static int mon_load( /* char *line */ );
static int mon_setreg( /* char *line */ );
static int mon_showreg( /* char *line */ );
static int mon_disassemble( /* char *line */ );
static int mon_memdump( /* char *line */ );
static int mon_bitdump( /* char *line */ );
static int mon_memstore( /* char *line */ );
static int mon_help( /* char *line */ );
#ifdef NEW_TIMER_CODE
static int mon_break( /* char *line */ );
static int mon_clear( /* char *line */ );
#endif
#endif	       /* __STDC__ */


struct mon_table {
    char   *command;
    int     min_args;
    int     max_args;

#ifdef __STDC__
    int     (*funcp) (char *);
#else
    int     (*funcp) ();
#endif
    char   *help_text;
};

struct mon_table mon_cmds[] = {
    {"fill", 3, MAXARG, mon_fill,
    "fill     start end|+len bytes|'string'\n"},
    {"hunt", 3, MAXARG, mon_hunt,
    "hunt     start end|+len bytes|'string'\n"},
    {"compare", 3, 3, mon_compare,
    "compare  start end|+len target\n"},
    {"transfer", 3, 3, mon_move,
    "transfer start end|+len dest\t(moves memory area)\n"},
    {"asm", 0, 1, mon_asm,
    "asm \t[addr]\t\t\t(enter to line interpreter)\n"},
    {"jump", 0, 1, mon_jump,
    "jump\t[addr]\n"},
    {"call", 0, 1, mon_jump,
    "call\t[addr]\t\t\t(run code at address specified)\n"},
    {"registers", 0, 0, mon_setreg,
    "registers [register list]\t\t(set PC AC AX YR PF SP)\n"},
    {";", 0, 5, mon_setreg,
    ";\t\t\t\t\t(same as registers)\n"},
    {"memory", 0, 2, mon_memdump,
    "memory [addr] [end|+len]\t\t(dumps memory in hex format)\n"},
    {"image", 0, 2, mon_bitdump,
    "image  [addr] [end|+len]\t\t(dumps memory in bit format)\n"},
    {":", 1, MAXARG, mon_memstore,
    ":\t  addr  bytes|'string'\t\t(write data to memory)\n"},
    {"disassemble", 0, 2, mon_disassemble,
    "disassemble [addr] [end|+len]\n"},
    {"load", 1, 2, mon_load,
    "load\t filename [addr]\n"},
    {"save", 3, 3, mon_save,
    "save\t filename  addr  end|+len\n"},
    {"help", 0, 1, mon_help,
    "help\t[topic]\n"},
    {"?", 0, 1, mon_help,
    "?\t[topic]\t\t\t\t(Same as help)\n"},
    {"x", 0, 0, mon_exit,
    "x\t\t\t\t\t(exits monitor. eg. returns to emulator)\n"},
    {"quit", 0, 0, mon_quit,
    "quit\t\t\t\t\t(exits x64)\n"},
    {"dump", 0, 1, mon_dump, /*FCP*/
    "dump\t [filename]\n"},
    {"undump", 0, 1, mon_undump, /*FCP*/
    "undump [filename]\t\t\t(loads ram image from file)\n"},
    {"floppy", 0, 1, mon_floppy, /*TVR*/
     "floppy [imagename]\t\t\t(attaches floppy image for use)\n"}, /*TVR*/
#ifdef NEW_TIMER_CODE
    {"break", 0, 2, mon_break,
     "break  [start] [end]\t\t\t(Set up breakpoints (region))\n"},
    {"clear", 1, 2, mon_clear,
     "clear   start [end]\t\t\t(Clear breakpoints (region))\n"},
#endif
    {NULL, 0, 0, 0, NULL}
};


#define T_QUOTED 1
#define T_NUMBER 2
#define T_OTHER  3
#define MASKVAL  0xffff

static char *args[MAXARG+1];
static int values[MAXARG+1];
static int types[MAXARG+1];
static int nargs;

static int mode;

/* function packs arguments after start to single piece of data */

#ifndef __STDC__
static void pack_args(start)
    int     start;
#else
static void pack_args(int start)
#endif
{
    int     i;
    static char buf[256];
    char   *p = buf;

    for (i = start; i < nargs; i++) {
	if (types[i] == T_QUOTED) {
	    memcpy(p, args[i], values[i]);
	    p += values[i];
	} else
	    *p++ = values[i];
    }
    *p = '\0';
    types[start] = T_QUOTED;
    args[start] = buf;
    values[start] = (int) p - (int) buf;
}

#ifndef __STDC__
static int mon_fill(line)
    char   *line;
#else
static int mon_fill(char *line)
#endif
{
    ADDRESS adr, end = values[2]-1;
    pack_args(3);

    if (values[3] > 0)
	for (adr = values[1]; adr <= end; adr += values[3])
	    memcpy(ram + adr, args[3], values[3]);
    return (0);
}


#ifndef __STDC__
static int mon_memstore(line)
    char   *line;
#else
static int mon_memstore(char *line)
#endif
{
    address = values[1] & 0xffff;
    pack_args(2);

    if (values[2] > 0) {
      memcpy(ram + address, args[2], values[2]);
    }
    address = (address + values[2]) & 0xffff;
    return (0);
}


#ifndef __STDC__
static int mon_hunt(line)
    char   *line;
{
#else
static int mon_hunt(char *line)
{
#endif
    ADDRESS adr = values[1], end = values[2];

    pack_args(3);
    if (values[3] > 0)
	for (; adr <= end;) {
	    if (!memcmp(ram + adr, args[3], values[3])) {
		printf("%.4X\t", adr);
		adr += values[3];
	    } else
		adr++;
	}
    printf("\n");
    return (0);
}

#ifndef __STDC__
static int mon_compare(line)
    char   *line;
#else
static int mon_compare(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2], trg = values[3];

    for (; adr <= end; adr++, trg++)
	if (LOAD(adr&65535) != LOAD(trg&65535))
	    printf("%.4X\t", adr);
    printf("\n");
    return (0);
}

#ifndef __STDC__
static int mon_move(line)
    char   *line;
#else
static int mon_move(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2], trg = values[3];

    if (adr > trg)
	for (; adr <= end; STORE(trg, LOAD(adr&65535)), trg++, adr++);
    else if (adr < trg)
	for (trg += (end - adr);
	     adr <= end; STORE(trg, LOAD(end&65535)), trg--, end-- );
    return (0);
}



#ifndef __STDC__
static int mon_asm(line)
    char   *line;
#else
static int mon_asm(char *line)
#endif
{
    if (nargs > 1)
	ass(values[1], mode);
    else
	ass(address, mode);
    return (0);
}

#ifndef __STDC__
static int mon_quit(line)
    char   *line;
#else
static int mon_quit(char *line)
#endif
{
    end64();
    return (0);
}

#ifndef __STDC__
static int mon_exit(line)
    char   *line;
#else
static int mon_exit(char *line)
#endif
{
    runflg = 0;
    return (1);
}

#ifndef __STDC__
static int mon_jump(line)
    char   *line;
#else
static int mon_jump(char *line)
#endif
{
    if (nargs > 1)
	PC = values[1];
    runflg = 1;
    return (1);
}

#ifndef __STDC__
static int mon_dump(line)
    char   *line;
#else
static int mon_dump(char *line)
#endif
{
    char    buffer[256];
    if (nargs <= 1) { /*FCP*/

	sprintf(buffer, "%s/%s", project_dir, ram_name);
	dump(buffer);
    } else
	dump(args[1]);

    printf("core dumped\n"); /*FCP*/
    return (0);
}

#ifndef __STDC__
static int mon_undump(line)
    char   *line;
#else
static int mon_undump(char *line)
#endif
{
    char    buffer[256];
    if (nargs <= 1) { /*FCP*/

	sprintf(buffer, "%s/%s", project_dir, ram_name);
	undump(buffer);
    } else
	undump(args[1]);

    printf("core loaded\n");
    return (0);
}

#ifndef __STDC__
static int mon_floppy(line) /*TVR*/
    char   *line;
#else
static int mon_floppy(char *line)
#endif
{
    /*char    buffer[256];*/
    if (nargs > 1)
	attach_floppy_image(args[1]);
    else
	attach_floppy_image(NULL);

    return (0);
}

#ifndef __STDC__
static int mon_save(line)
    char   *line;
#else
static int mon_save(char *line)
#endif
{
    FILE   *fp;
    ADDRESS adr = values[2];
    ADDRESS end = values[3];

    if (NULL == (fp = fopen(args[1], "w"))) {
	printf("Saving failed.\n");
    } else {
	printf("Saving file %s\n", args[1]);
	fputc((BYTE) adr & 0xff, fp);
	fputc((BYTE) (adr >> 8) & 0xff, fp);
	fwrite((char *) (ram + adr), 1, end - adr, fp);
	fclose(fp);
    }
    return (0);
}

#ifndef __STDC__
static int mon_load(line)
    char   *line;
#else
static int mon_load(char *line)
#endif
{
    FILE   *fp;
    ADDRESS adr = values[2];

    printf((*line == 'l') ? "Loading %s " : "Verifying %s ", args[1]);
    fflush(stdout);
    if (NULL == (fp = fopen(args[1], "r"))) {
	printf("Loading failed.\n");
	perror(args[1]);
    } else {
	int     ch;

	if (*line == 'l' && nargs < 3) {
	    adr = fgetc(fp);
	    adr += ((fgetc(fp)) << 8);
	} else {
	    (void) fgetc(fp);
	    (void) fgetc(fp);
	}

	printf("%04X - ", adr);
	if (*line == 'l')
	    while ((ch = fgetc(fp)) != EOF)
		ram[adr++] = ch;
	else {
	    while ((ch = fgetc(fp)) != EOF)
		if (ram[adr++] != ch)
		    printf("%.4X\t", adr - 1);
	    printf("\n");
	}

	/* set end of load addresses like kernel load */ /*FCP*/
	if (*line == 'l') {
	    ram[0x2d]=ram[0x2f]=ram[0x31]=adr&0xff;
	    ram[0x2e]=ram[0x30]=ram[0x32]=(adr>>8)&0xff;
	  }

	fclose(fp);
	printf("%04X\n", adr);
	}
    return (0);
}

#ifndef __STDC__
static int mon_setreg(line)
    char   *line;
#else
static int mon_setreg(char *line)
#endif
{
    /* read registers if more than one parameter */

    switch (nargs) {
	case 7:
	SP = values[4];
      case 6:
	SET_SR(values[3]);
      case 5:
	YR = values[2];
      case 4:
	XR = values[1];
      case 3:
	AC = values[0];
      case 2:
	PC = values[1];
    }

    mon_showreg(NULL);
    return (0);
}

#ifndef __STDC__
static int mon_showreg(line)
    char   *line;
#else
static int mon_showreg(char *line)
#endif
{
    printf(" PC   AC  XR  YR  nv1bdizc  SP    OPCODE\n");
    printf((mode & MODE_HEX) ?
	"%.4X  %.2X  %.2X  %.2X  %s  %.2X   %3X %3d %s\n" :
	"%04d %03d %03d %03d %s %03d  %3d %2X %s\n",
	(int) PC, (int) AC, (int) XR, (int) YR,
	sprint_status(), (int) SP,
	LOAD(PC), LOAD(PC), sprint_opcode(PC, mode & MODE_HEX));
    return (0);
}

#ifndef __STDC__
static int mon_disassemble(line)
    char   *line;
#else
static int mon_disassemble(char *line)
#endif
{
    ADDRESS adr = values[1], end =  values[2];
    int     cnt;

    switch (nargs) {
      case 1:
	adr = address;

      case 2:
	cnt = 20;
	while (cnt-- > 0) {
	    printf(".%.4X\t%s\n", adr&65535, sprint_opcode(adr&65535, 1));
	    adr += clength[lookup[LOAD(adr&65535)].addr_mode];
	}
	break;

      case 3:
	while (adr <= end) {
	    printf(".%.4X\t%s\n", adr&65535, sprint_opcode(adr&65535, 1));
	    adr += clength[lookup[LOAD(adr&65535)].addr_mode];
	}
    }
    address = adr & 0xffff;
    return (0);
}

#ifndef __STDC__
static int mon_memdump(line)
    char   *line;
#else
static int mon_memdump(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2];
    char    arg[17];
    int     cnt;

    switch (nargs) {
      case 1:
	adr = address;

      case 2:
	end = adr + 254;
	break;
    }

    arg[16] = 0;

    while (adr < end) {
	printf(" %.4X:", adr&65535);
	for (cnt = 0; cnt < 16; cnt++, adr++) {
	    printf(" %.2X", LOAD(adr&65535)); /* timing isn't so critical */
	    arg[cnt] = isprint(LOAD(adr&65535)) ? LOAD(adr&65535) : '.';
	}
	printf(" :%s\n", arg);
    }

    address = adr & 0xffff;
    return (0);
}


#ifndef __STDC__
static int mon_bitdump(line)
    char   *line;
#else
static int mon_bitdump(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2];
    char    arg[17];
    int     cnt;

    switch (nargs) {
      case 1:
	adr = address;

      case 2:
	end = adr + 60;
	break;
    }

    arg[3] = 0;

    while (adr <= end) {
	unsigned int bit, val, mask;
	printf(" %.4X: ", adr&0xffff);
	for (cnt = 0; cnt < 3; cnt++, adr++) {
	    val = LOAD(adr&0xffff) & 0xff;
	    for (bit = 7, mask = 0x80; bit--; mask >>= 1)
		printf((val & mask) ? "*" : "-");
	}
	printf("\n");
	if ((adr & 0x3f) == 0x3f)
	    adr++;	/* Skip the unused byte */
    }

    address = adr & 0xffff;
    return (0);
}


#ifdef NEW_TIMER_CODE
static BYTE break_bits[65536/8];
static int  break_alarm = 0;

#ifndef __STDC__
static void break_alarm_cb(i)
int i;
#else
static void break_alarm_cb(int i)
#endif
{
    /*
     * This will be called twice / cycle. This way we can be sure
     * breakpoint is tested if interrupt happens on same cycle on other
     * alarm.
     */
    static int old_pc = -1;
    int    kludge;

    kludge = (PC == old_pc) ? 1 : 0;

    old_pc = PC;
    
    set_alarm(break_alarm, clk + kludge, break_alarm_cb, 0);

    if (kludge)
	return;
    
    if (break_bits[PC/8] & (1 << (PC%8)))
    {
	printf("Monitor break\n");
	mon(PC);
    }
}
    

#ifndef __STDC__
static void set_break_alarm()
#else
static void set_break_alarm(void)
#endif
{
    int i;
    if (!break_alarm)
	break_alarm = new_alarm();

    for (i=0; i<65536/8; i++)
	if (break_bits[i])
	{
	    break_alarm_cb(0);
	    return;
	}
    clear_alarm(break_alarm);
}


#ifndef __STDC__
static int mon_break(line)
    char   *line;
#else
static int mon_break(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2];
    int found = 0;
    int i;
    int started;
    
    /*
     * Show current breakpoints
     */
    if (nargs == 1)
    {
	for (i=0; i<65536; i++)
	{
	    if (break_bits[i/8] & (1 << (i%8)))
	    {
		if (!found)
		{
		    printf("Breakpoints: ");
		    found = 1;
		}
		started = i;
		i++;
		while (i < 65536 && (break_bits[i/8] & (1 << (i%8))))
		    i++;

		i--;
		if (found == 1)
		    found = 2;
		else
		    printf(",");

		if (i > started + 1)
		    printf("%04x-%04x", started, i);
		else if (i == started + 1)
		    printf("%04x,%04x", started, i);
		else
		    printf("%04x", started);
	    }
	}
	if (!found)
	    printf("No breakpoints set.\n");
	else
	    printf("\n");

	return 0;
    }

    
    if (nargs == 2)
	end = adr;

    while (adr <= end)
    {
	break_bits[adr/8] |= 1 << (adr%8);
	adr++;
    }

    set_break_alarm();
    return (0);
}


#ifndef __STDC__
static int mon_clear(line)
    char   *line;
#else
static int mon_clear(char *line)
#endif
{
    ADDRESS adr = values[1], end = values[2];
    
    if (nargs == 2)
	end = adr;

    while (adr <= end)
    {
	break_bits[adr/8] &= ~(1 << (adr%8));
	adr++;
    }

    set_break_alarm();
    return (0);
}
#endif /* NEW_TIMER_CODE */


#ifndef __STDC__
static int mon_help(line)
    char   *line;
#else
static int mon_help(char *line)
#endif
{
    int     i;

    printf("\nFollowing commands are available:\n\
  #%%&$\t\t\t\t\t(base conversions)\n");
    for (i = 0; mon_cmds[i].command; i++) {
	printf("  %s", mon_cmds[i].help_text);
    }
    printf("\n\tAll commands may be abbreviated\n");
    return (0);
}


#ifndef __STDC__
void    mon(adr)
    ADDRESS adr;
#else	       /* __STDC__ */
void    mon(ADDRESS adr)
#endif	       /* __STDC__ */
{
    int     len;
    char   *line;
    int     temparg;
    int     i, j;

    int     quote = 0, b = 0, bi = 0;
    char   *p, *q, *r;

#ifdef EDIT
    e_setterm();
#endif

    runflg = 0;/* terminate RUN */
    address = adr;

    printf("\nMonitor.\n");
    mode = MODE_SPACE | (MODE_HEX & hexflg);

    mon_showreg(NULL);

    /* monitor main loop */

    do {
	len = 0;
	if ((line = read_line(address, MODE_MON)) == NULL)
	    end64();

	if (*line) {

	    /* split command line to table and parse possible numeric values */
	    quote = nargs = 0;
	    for (i = 0; i < MAXARG; i++) {
		values[i] = 0;
		types[i] = 0;
		args[i] = NULL;
	    }

	    p = q = r = line;
	    do {
		if (*p == '\'' && quote != 1) {
		    quote ^= 2;
		    types[nargs] = T_QUOTED;
		} else if (*p == '"' && quote != 2) {
		    quote ^= 1;
		    types[nargs] = T_QUOTED;

		} else if (!*p || (!quote && strchr(" \t\n,", *p))) {
		    *p = '\0';
		    /* r points to start of current word */
		    if (*r) {
			args[nargs] = r;
			if (types[nargs] == T_QUOTED) {
			    values[nargs] = strlen(r);
			} else {
			    int cb = 0;
			    if (*r == '+') {
				cb++; r++;
			    }
			    if (nargs && sconv(r, 0, mode | MODE_QUERY)) {
				values[nargs] = sconv(r, 0, mode);
				if (cb) values[nargs] += values[nargs-1];
				if (values[nargs] > MASKVAL+1) {
				    values[nargs] &= MASKVAL;
				    printf("Value too large\n");
				}
				types[nargs] = T_NUMBER;
				/*  formats: "m 400 4ff", "m 400,4ff"
				 * ("m 400-4ff"), ("m 400+ff"), "m 400 +ff"
				 */
			    } else {
				values[nargs] = strlen(r);
				types[nargs] = T_OTHER;
			    }
			} /* (!*r) */
			nargs++;
			r = ++q;
		    }
		} else if (*p >= 0x20)
		    *q++ = *p;

		/* else discard character */
	    } while (*++p && nargs < MAXARG);

	    /*
	     * for (i = 0; i < MAXARG; i++) {
	     * printf ("arg %2d = '%s' type %d  value %d\n",
	     * i, args[i], types[i], values[i]); }
	     */

	    if (!nargs || !args[0] || !args[0][0])
		continue;

	    /* now search through command table and find nearest command */
	    b = 0;
	    for (i = 0; mon_cmds[i].command; i++) {
		for (p = mon_cmds[i].command, q = args[0], j = 0;
		     *p && *q && *p == *q; p++, q++, j++);
		/*
		 * printf ("compare %s %s - %d %d %d\n", mon_cmds[i].command,
		 * args[0], j, b, bi);
		 */
		if (j > b && !*q) {
		    b = j;
		    bi = i;

		    /* found an exact command */
		    if (*p == *q)
			break;
		}
	    }

	    /* command should now have been identified */
	    if (b) {
		if (nargs - 1 > mon_cmds[bi].max_args ||
		    nargs - 1 < mon_cmds[bi].min_args) {
		    printf("Wrong number of arguments.\n%s\n",
			   mon_cmds[bi].help_text);
		    continue;
		}
		if (mon_cmds[bi].funcp(line))
		    break;	/* used for exit and jump to prg */
		continue;
	    }
	    switch (*line) {
		/*case '+':*/
	      case '-':
	      case '%':
	      case '&':
	      case '#':
	      case '$':
		/* Base conversion. */
		temparg = sconv(line, 0, mode);

		printf("\n %17x x\n %17d d\n %17o o\n %s %s b\n\n",
		       temparg, temparg, temparg, sprint_binary(UPPER(temparg)),
		       sprint_binary(LOWER(temparg)));
		break;

	      case '*':
		address = sconv(line + 1, 0, mode) & 0xffff;
		break;

	      default:
		printf(" Unknown command. try 'help'\n");
	    }
	} /* line */
    } while (!runflg);
}


/*****************  evaluate numeric  ****************/
/************   --- with error checking   ************/

#ifndef __STDC__
int     sconv(s, level, mode)
    char   *s; /* pointer to input string */
    int     level;	/* recursion level     */
    int     mode;	/* flag: dec/hex mode  */
#else	       /* __STDC__ */
int     sconv(char *s, int level, int mode)
#endif	       /* __STDC__ */
{
    static char hexas[] = "0123456789abcdefg";
    char   *p = s;
    int     base = 0;
    int     result = 0, sign = 1;
    int     i;

    if (!p)
	return (0);

    switch (*p) {
      case '%':
	p++;
	base = 2;
	break;

      case 'o':
      case '&':
	p++;
	base = 8;
	break;

      case 'x':
      case '$':
	p++;
	base = 16;
	break;

      case 'u':
      case 'i':
      case '#':
	p++;
	base = 10;
	break;

      case '0':	/* 0x 0b 0d */
	p++;
	if (!isdigit(*p))
	    return (sconv(p, level+1, mode & ~MODE_HEX));

      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	base = (mode & MODE_HEX) ? 16 : 10;
	break;

      case 'a':
      case 'c':
      case 'e':
      case 'f':
	if (mode & MODE_HEX)
	    base = 16;
	break;

      case 'b':	/* hex or binary */
	if (mode & MODE_HEX)
	    base = 16;
	else {
	    base = 2;
	    p++;
	}
	break;

      case 'd':	/* hex or decimal */
	if (mode & MODE_HEX)
	    base = 16;
	else {
	    base = 10;
	    p++;
	}
	break;

      default:
	break;
    }

    /*
     * now p points to start of string to convert and base hold its base
     * number 2, 8, 10 or 16
     */

    if (!base)
	return (0);

    if (*p == '-') {
	sign = -1;
	p++;
    }
    while (*p) {
	for (i = 0; i < base; i++)
	    if (*p == hexas[i]) {
		result = result * base + i;
		break;
	    }
	if (i >= base) {
	    /* unknown char has occurred, return value or error */
	  if (!level && !(mode & MODE_QUERY) &&
	      !strchr(",-+()", *p) && !isspace(*p))
	    printf ("Bad character near '%s'\n", p);
	  else
	      i = 0;
	  break;
	}
	p++;
    }
    /* printf ("base %d value %d\n", base, result); */
    /* return final value */
    return ((mode & MODE_QUERY) ? (i < base) : result * sign);
}


#ifndef __STDC__
char   *read_line(adr, mode)
    ADDRESS adr;
    int     mode;
#else	       /* __STDC__ */
char   *read_line(ADDRESS adr, int mode)
#endif	       /* __STDC__ */
{
    static char line[256];
    char   *linep, *r_linep;

    sprintf(line, (mode & MODE_MON) ? "%04x >" : ".%04x ", adr);

#ifdef EDIT
    r_linep = linep = e_gets(line);
    if (linep && *linep)
	e_push_history(linep);
#else
    fputs(line, stdout);
    fflush(stdout);
    r_linep = linep = fgets(line, 255, stdin);
#endif

    while (r_linep && *r_linep && isspace(*r_linep))
	r_linep++;

    return (r_linep);
}


#ifndef __STDC__
int     dump(name)
    char   *name;
#else	       /* __STDC__ */
int     dump(char *name)
#endif	       /* __STDC__ */
{
    FILE   *fp;
    if (NULL == (fp = fopen(name, "w")) ||
	1 != fwrite((char *) ram, RAMSIZE, 1, fp))
	return (0);
    (void) fclose(fp);

    return (0);
}

#ifndef __STDC__
int     undump(name)
    char   *name;
#else	       /* __STDC__ */
int     undump(char *name)
#endif	       /* __STDC__ */
{
    FILE   *fp;

    if (NULL == (fp = fopen(name, "r")) ||
	1 != fread((char *) ram, RAMSIZE, 1, fp))
	return (0);
    (void) fclose(fp);

    return (0);
}
