/*
 * $Id: op.c,v 1.3 93/06/21 13:38:52 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains 6502 machine language instructions.
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	op.c,v $
 * Revision 1.3  93/06/21  13:38:52  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:08:35  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include "macro.h"
#include "extern.h"
#include "op.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#ifndef HACK

#ifndef __STDC__
unsigned int load(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int load(unsigned int src)
#endif	       /* __STDC__ */
{
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int store(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int store(unsigned int src)
#endif	       /* __STDC__ */
{
    return (src);
}

#ifndef __STDC__
unsigned int nop(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int nop(unsigned int src)
#endif	       /* __STDC__ */
{
    return (src);
}

#ifndef __STDC__
unsigned int inc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int inc(unsigned int src)
#endif	       /* __STDC__ */
{
    src = (src + 1) & 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int dec(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int dec(unsigned int src)
#endif	       /* __STDC__ */
{
    src = (src - 1) & 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int setf(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int setf(unsigned int src)
#endif	       /* __STDC__ */
{
    return (1);
}

#ifndef __STDC__
unsigned int clrf(src)
    unsigned int src;

#else	       /* __STDC__ */
unsigned int clrf(unsigned int src)
#endif	       /* __STDC__ */
{
    return (0);
}

#ifndef __STDC__
unsigned int ora(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int ora(unsigned int src)
#endif	       /* __STDC__ */
{
    src |= AC;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int and(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int and(unsigned int src)
#endif	       /* __STDC__ */
{
    src &= AC;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int eor(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int eor(unsigned int src)
#endif	       /* __STDC__ */
{
    src ^= AC;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src & 0xff);
}

#ifndef __STDC__
unsigned int adc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int adc(unsigned int src)
#endif	       /* __STDC__ */
{
    register unsigned int temp = src + AC + (IF_CARRY() ? 1 : 0);
    SET_SIGN(temp);
    SET_ZERO(temp & 0xff);	/* These are not valid in decimal mode */

    SET_OVERFLOW(!((AC ^ src) & 0x80) && ((AC ^ temp) & 0x80));

    /*
     * In decimal mode this is "correct" but not exact emulation. However,
     * the propability of different result when using undefined BCD codes is
     * less than 6%
     */

    if (IF_DECIMAL()) {
	if ((AC & 0xf) + (src & 0xf) > 9)
	    temp += 6;
	if (temp > 159)
	    temp += 96;
	SET_CARRY(temp > 0x99);
    } else
	SET_CARRY(temp > 0xff);

    return (temp & 0xff);
}

#ifndef __STDC__
unsigned int sbc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int sbc(unsigned int src)
#endif	       /* __STDC__ */
{
   register unsigned int temp = AC - src - (IF_CARRY() ? 0 : 1);

    SET_SIGN(temp);
    SET_ZERO(temp & 0xff);	/* These are not valid in decimal mode */

    SET_OVERFLOW(((AC ^ temp) & 0x80) && ((AC ^ src) & 0x80));

    if (IF_DECIMAL()) {
	if ((AC & 0xf) < (src & 0xf))
	    temp -= 6;
	if (temp > 0x99)
	    temp -= 0x60;
    }
    SET_CARRY(temp < 0x100);
    return (temp & 0xff);
}

#ifndef __STDC__
unsigned int cmp(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int cmp(unsigned int src)
#endif	       /* __STDC__ */
{
    src = AC - src;
    SET_CARRY(src < 0x100);
    SET_SIGN(src);
    SET_ZERO(src &= 0xff);
    return (src);
}

#ifndef __STDC__
unsigned int cpx(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int cpx(unsigned int src)
#endif	       /* __STDC__ */
{
    src = XR - src;
    SET_CARRY(src < 0x100);
    SET_SIGN(src);
    SET_ZERO(src &= 0xff);
    return (src);
}

#ifndef __STDC__
unsigned int cpy(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int cpy(unsigned int src)
#endif	       /* __STDC__ */
{
    src = YR - src;
    SET_CARRY(src < 0x100);
    SET_SIGN(src);
    SET_ZERO(src &= 0xff);
    return (src);
}

#ifndef __STDC__
unsigned int bit(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bit(unsigned int src)
#endif	       /* __STDC__ */
{
    SET_SIGN(src);
    SET_OVERFLOW(0x40 & src);	/* Copy bit 6 to OVERFLOW flag. */
    SET_ZERO(src & AC);
    return (src);
}

#ifndef __STDC__
unsigned int asl(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int asl(unsigned int src)
#endif	       /* __STDC__ */
{
    SET_CARRY(src & 0x80);
    src <<= 1;
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int asr(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int asr(unsigned int src)
#endif	       /* __STDC__ */
{
    SET_CARRY(src & 0x01);
    src >>= 1;
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int rol(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int rol(unsigned int src)
#endif	       /* __STDC__ */
{
    src <<= 1;
    if (IF_CARRY())
	src |= 0x1;
    SET_CARRY(src > 0xff);
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int ror(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int ror(unsigned int src)
#endif	       /* __STDC__ */
{
    if (IF_CARRY())
	src |= 0x100;
    SET_CARRY(src & 0x01);
    src >>= 1;
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}


/* BRANCH instructions. */

#ifndef __STDC__
unsigned int bcc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bcc(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if carry flag is clear. */
    if (!IF_CARRY())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bcs(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bcs(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if carry flag is set. */
    if (IF_CARRY())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bne(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bne(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if  zero flag is clear. */
    if (!IF_ZERO())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int beq(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int beq(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if  zero flag is set. */
    if (IF_ZERO())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bpl(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bpl(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if sign flag is clear. */
    if (!IF_SIGN())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bmi(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bmi(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if sign flag is set. */
    if (IF_SIGN())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bvc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bvc(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if overflow flag is clear. */
    if (!IF_OVERFLOW())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int bvs(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int bvs(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Branch if overflow flag is set. */
    if (IF_OVERFLOW())
	PC = REL_ADDR(PC, src);
    return (src);
}

#ifndef __STDC__
unsigned int push(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int push(unsigned int src)
#endif	       /* __STDC__ */
{
    PUSH(src);
    /* First push item onto stack and then decrement SP. */
    return (src);
}

#ifndef __STDC__
unsigned int pla(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int pla(unsigned int src)
#endif	       /* __STDC__ */
{
    /* First increment stack pointer and then pull item from stack. */
    src = PULL();
    SET_SIGN(src);	/* Change sign and zero flag accordingly. */
    SET_ZERO(src);
    return (src);
}

#ifndef __STDC__
unsigned int plp(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int plp(unsigned int src)
#endif	       /* __STDC__ */
{
    /* First increment stack pointer and then pull item from stack. */
    src = PULL();
    return (src);
}

#ifndef __STDC__
unsigned int jsr(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int jsr(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Jump to subroutine. */
    PC--;
    PUSH((PC >> 8) & 0xff);	/* Push return address onto the stack. */
    PUSH(PC & 0xff);

#ifdef DEBUG
    if (debugflg)
	print_stack(SP);
#endif

    return (src);
}

#ifndef __STDC__
unsigned int rts(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int rts(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Return from subroutine. */
    src = PULL();
    src += ((PULL()) << 8) + 1;	/* Load return address from stack. */

#ifdef DEBUG
    if (debugflg)
	print_stack(SP);
#endif

    return (src);
}

#ifndef __STDC__
unsigned int rti(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int rti(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Return from interrupt. */
    /* Load program status and program counter from stack. */
    src = PULL();
    SET_SR(src);
    src = PULL();
    src += (PULL() << 8) + 1;	/* Load return address from stack. */

#ifdef DEBUG
    if (debugflg)
	print_stack(SP);
#endif

    return (src);
}

#ifndef __STDC__
unsigned int brk(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int brk(unsigned int src)
#endif	       /* __STDC__ */
{	       /* Force break. */

#ifdef DEBUG
    if (verflg) {
	printf("BRK %d\n", clk);
	show();
    }
#endif

    if (runflg) {
	--PC;
	printf("BRK\n");
	mon(PC);
	return (PC);
    }
    if (trap_handler())
	return (intr(I_BRK, PC));
    return (PC);
}


#ifndef __STDC__
unsigned int noop(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int noop(unsigned int src)
#endif	       /* __STDC__ */
{	       /* No such opcode. */
    (void) printf("Illegal instruction.\n");
    verflg = 1;
    --PC;
    show();
    mon(PC);
    return (PC);
}


#ifndef NO_UNDOC_CMDS
/*
 * MOS 6510 undocumented commands
 * If you are emulating the 8502, be careful with these since
 * it may get the xxxxxx10's the other way around.
 *
 * These functions emulate a 6510 (manuf. week 5183) and 8502 R0 (week 4685).
 * However, your CPU may have them differently.
 *
 * BTW: I used 68000 based 80286 emulator to run this editor.
 */

#ifndef __STDC__
unsigned int u_rlan(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_rlan(unsigned int src)
#endif	       /* __STDC__ */
{
    /* src = rol(src); AC = and(src); */

    src <<= 1;
    if (IF_CARRY())	/* ROL+AND */
	src |= 0x1;
    SET_CARRY(src > 0xff);
    src &= 0xff;

    AC &= src;
    SET_SIGN(AC);
    SET_ZERO(AC);
    return (src);	/* rotated, not disturbed */
}

#ifndef __STDC__
unsigned int u_rrad(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_rrad(unsigned int src)
#endif	       /* __STDC__ */
{
  register unsigned int temp;

    /* src = ror(src); AC = adc(src); */


    if (IF_CARRY())	/* ROR+ADC */
	src |= 0x100;
    SET_CARRY(src & 0x01);
    src >>= 1;
    src &= 0xff;

    temp = src + AC + (IF_CARRY() ? 1 : 0);
    SET_SIGN(temp);
    SET_ZERO(temp & 0xff);	/* These are not valid in decimal mode */

    SET_OVERFLOW(!((AC ^ src) & 0x80) && ((AC ^ temp) & 0x80));

    if (IF_DECIMAL()) {
	if ((AC & 0xf) + (src & 0xf) > 9)
	    temp += 6;
	if (temp > 159)
	    temp += 96;
	SET_CARRY(temp > 0x99);
    } else
	SET_CARRY(temp > 0xff);

    AC = temp;
    return (src);
}

#ifndef __STDC__
unsigned int u_slor(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_slor(unsigned int src)
#endif	       /* __STDC__ */
{
    /* src = asl(src); AC = ora(src); */

    SET_CARRY(src & 0x80);	/* ASL+ORA */
    src <<= 1;
    src |= AC;
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}


#ifndef __STDC__
unsigned int u_sreo(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_sreo(unsigned int src)
#endif	       /* __STDC__ */
{
    /* src = asr(src); AC = eor(src); */

    SET_CARRY(src & 0x01);	/* LSR+EOR */
    src >>= 1;
    src ^= AC;
    src &= 0xff;
    SET_SIGN(src);
    SET_ZERO(src);

    return (src);
}

#ifndef __STDC__
unsigned int u_decp(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_decp(unsigned int src)
#endif	       /* __STDC__ */
{
    /* cmp(--src & 0xff)); return (src); */

    src = (src - 1) & 0xff;	/* DEC+CMP */
    SET_CARRY(AC >= src);
    SET_SIGN(AC - src);
    SET_ZERO(AC != src);
    return (src);
}

#ifndef __STDC__
unsigned int u_isbc(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_isbc(unsigned int src)
#endif	       /* __STDC__ */
{
  register unsigned int temp;

    /* src = ++src & 0xff; AC = sbc(src); */

    src = ((src + 1) & 0xff);	/* INC+SBC */
    temp = AC - src - (IF_CARRY() ? 0 : 1);
    SET_SIGN(temp);
    SET_ZERO(temp);

    SET_OVERFLOW(((AC ^ temp) & 0x80) && ((AC ^ src) & 0x80));

    if (IF_DECIMAL()) {
	if ((AC & 0xf) < (src & 0xf))
	    temp -= 6;
	if (temp > 0x99)
	    temp -= 0x60;
    }
    SET_CARRY(temp < 0x100);

    AC = temp;
    return (src);	/* src saved */
}

#ifndef __STDC__
unsigned int u_sbx(src)
    unsigned int src;
#else	       /* __STDC__ */
unsigned int u_sbx(unsigned int src)
#endif	       /* __STDC__ */
{
    src = (AC & XR) - src;	/* Carry is ignored (CMP) */
    /* Overflow flag may be affected */
    SET_CARRY(src < 0x100);

    src &= 0xff;	/* No decimal mode */
    SET_SIGN(src);
    SET_ZERO(src);
    return (src);
}

#endif  /* undoc */

#endif  /* hack */
