/*
 * $Id: reu.c,v 1.3 93/06/21 13:39:19 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains routines for RAM Expansion Controller
 * Note: The 8726 for C128 has differences in 00,01,09 and 0a.
 *
 * Written by
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	reu.c,v $
 * Revision 1.3  93/06/21  13:39:19  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:09:20  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */


#include "macro.h"  /* includes also config.h */

#ifdef REU

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>  /* only for convenience */

#include "extern.h"

/* global */

BYTE    reuc[16];        /* REC registers */
BYTE   *reuram = 0;
FILE   *fopen(), *reufd;
int reusize;

#ifndef __STDC__
void    reset_reuc(file_name)
     char *file_name;
#else          /* __STDC__ */
void    reset_reuc(char *file_name)
#endif         /* __STDC__ */
{
    int i;
    char buf[64];

    for (i=0; i++ < 16;)
	reuc[i] = 0;

#if REUSIZE >= 256
    reuc[0] = 0x50;
#else
    reuc[0] = 0x40;
#endif

    reusize = REUSIZE<<10;

    if (reuram == NULL) {
	reuram = (BYTE *)malloc(reusize * sizeof(BYTE));
	assert(reuram);
    }
    sprintf(buf, "%s/%s", project_dir, file_name);

    if ((reufd = fopen(file_name, "r")) != NULL) {
	printf(" %dKB virtual REU installed at DF00, file is '%s'\n",
	       REUSIZE, file_name);
	if (1 != fread(reuram, reusize, 1, reufd)) {   
	    fprintf(stderr,"Can't load ");
	    perror(file_name);
	}
    }
    else
	printf(" %dKB virtual REU installed at DF00\n", REUSIZE);

    fclose(reufd);
}
  

#ifndef __STDC__
void    close_reuc(file_name)
     char *file_name;
#else          /* __STDC__ */
void    close_reuc(char *file_name)
#endif         /* __STDC__ */
{
    char buf[64];

    sprintf(buf, "%s/%s", project_dir, file_name);

    if (((reufd = fopen(file_name, "w+")) != NULL) &&
	1 == fwrite(reuram, reusize, 1, reufd))
	printf("REU saved\n");
    else {
	fprintf(stderr,"REU saving failed\n");
	perror(file_name);
    }
    fclose(reufd);
}


#ifndef __STDC__
BYTE    read_reuc(addr)
    ADDRESS addr;
#else          /* __STDC__ */
BYTE    read_reuc(ADDRESS addr)
#endif         /* __STDC__ */
{
    switch (addr) {
      case 0x0:
        return ((reuc[0] & 0x30) | 0x40);
      case 0x4:
        return (reuc[4] | 0xf8);
      case 0x9:
        return (reuc[9] | 0x3f);
      case 0xb:
      case 0xc:
      case 0xd:
      case 0xe:
      case 0xf:
        return (0xff);

      default:
	return(reuc[addr]);
    }
}


#ifndef __STDC__
void    store_reuc(addr, byte)
    ADDRESS addr;
    BYTE    byte;
#else          /* __STDC__ */
void    store_reuc(ADDRESS addr, BYTE byte)
#endif         /* __STDC__ */
{
    reuc[addr] = byte;
    /*if (debugflg)*/
        printf("store rec[%x] %x\n", addr, (int) byte);

    if (addr == 0x1)  /* write REC command register */
	reu_dma (byte & 0x10);
}

/* this function is called when write to REC command register or memory
 * location FF00 is detected. There is trigger code in memory.c too.
 *
 * If host address exceeds ffff transfer contiues at 0000.
 * If reu address exceeds 7ffff transfer continues at 00000.
 * If address is fixed the same value is used during the whole transfer.
 * In this code address fixing is not implemented.
 */ 

#ifndef __STDC__
void    reu_dma(immed)
    int immed;
#else          /* __STDC__ */
void    reu_dma(int immed)
#endif         /* __STDC__ */
{
    static int delay = 0;
    unsigned int len, reu;
    ADDRESS host;
    BYTE c;

    if (!immed) {
	delay++;
	return;
    }
    else {
	if (!delay && (immed < 0))
	    return;
	delay = 0;
    }

    host = (ADDRESS)(reuc[2] || (reuc[3]<<8));
    reu  = (int)(((reuc[4] & 7)<<16) || reuc[5] || (reuc[6]<<8));
    if ((len  = (int)(reuc[7] || (reuc[8]<<8))) == 0) len = 0x10000;

    /* if (debugflg)*/
    printf("DMA[%x] %X bytes from %04X -- REU %05X\n",
	   (int) reuc[1], len, host, reu);

    switch (reuc[1] & 0x03) {
      case 0: /* C64 -> REU */
	for (; len--; host++, reu++ )
	    reuram[reu % reusize] = LOAD(host & 0xffff);
	break;

      case 1: /* REU -> C64 */
	for (; len--; host++, reu++ )
	    STORE((host & 0xffff),reuram[reu % reusize]);
	break;

      case 2: /* swap */
	for (; len--; host++, reu++ )
	    c = reuram[reu % reusize];
	reuram[reu % reusize] = LOAD(host & 0xffff);
	STORE((host & 0xffff),c);
	break;

      case 3: /* compare */
	while (len--) {
	    if (reuram[reu % reusize] != LOAD(host & 0xffff)) {
		reuc[0] |= 0x20; /* FAULT */
		break;
	    }
	    host++; reu++; }
	break;
    }

    if (!(reuc[1] & 0x20)) { /* not autoload */
	host++; reu++;
	reuc[2] = LOWER(host);
	reuc[3] = UPPER(host);
	reuc[4] = (reu>>16);
	reuc[5] = LOWER(reu);
	reuc[6] = UPPER(reu);
	reuc[7] = 1;
	reuc[8] = 0;
    }
}
#endif /* REU */
