/*
 * $Id: serial.c,v 1.3 93/06/21 13:39:21 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *
 * $Log:	serial.c,v $
 * Revision 1.3  93/06/21  13:39:21  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:09:23  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * serial.c -- Serial device interface.
 */

#include <stdio.h>

#include "macro.h"
#include "extern.h"
#include "serial.h"

/*
 * Store name here for serial-open
 */
#define SERIAL_NAMELENGTH 256

static BYTE SerialBuffer[SERIAL_NAMELENGTH];
static int SerialPtr;


/*
 * On which channel did listen happen to ?
 */
static int SerialLatestListen;
static int SerialTrapListen;

/*
 * And what about salisten
 */
static int SerialLatestSaListen;


/*
 * Initialized serialdevices
 */
static serial_t serialdevices[16];


static int fn();

static void serialsendbyte()
{
    int     data;
    serial_t *p;

    /*
     * Get data to send from address 0x95
     */
    data = LOAD(0x95);

#ifdef DEBUG
    printf("SerialSendByte(0x%x)\n", data);
#endif

    p = &serialdevices[SerialLatestListen & 0x0f];

    if (!p -> isopen) {
	/*
	 * Store name here
	 */
	if (SerialPtr < SERIAL_NAMELENGTH)
	    SerialBuffer[SerialPtr++] = data;
    } else {
	/*
	 * Send to device
	 */
	p -> putf(data, SerialLatestSaListen & 0x0f);
    }

    /*
     * Set PC to RTS
     */
    PC = 0xEDAC;
}

static void serialreceivebyte()
{
    int     i;
    BYTE    pp;
    serial_t *p;

#ifdef DEBUG
    printf("SerialReceiveByte()\n");
#endif

    p = &serialdevices[SerialLatestListen & 0x0f];
    i = p -> getf(&pp, SerialLatestSaListen & 0x0f);

    /*
     * Set up serial success / data
     */
    STORE(0x90, i);
    STORE(0xA4, pp);

    /*
     * Set PC to RTS
     */
    PC = 0xEDAC;

    /*
     * Set AC and CARRY like assembler routine does
     */
    AC = pp;
    SET_CARRY(0);
}


static void seriallisten()
{

#ifdef DEBUG
    printf("SerialListen(0x%x)\n", AC);
#endif

    SerialTrapListen = AC;

    /*
     * Set PC to 'RTS'
     */
    PC = 0xEDAC;
}

static void serialsalisten()
{
    serial_t *p;
    int     b;
    int     i;

    /*
     * Which SaListen ?
     */
    b = LOAD(0x95);

#ifdef DEBUG
    printf("SerialSaListen(0x%x)\n", b);
#endif

    /*
     * which device ?
     */
    p = &serialdevices[SerialLatestListen & 0x0f];

    if (b == 0x3f) {
	switch (SerialLatestSaListen & 0xf0) {
	    /*
	     * Open
	     */
	  case 0xF0:
	    if (!p -> isopen) {
		p -> isopen = 1;
		i = p -> openf((char *)SerialBuffer, SerialPtr,
		    SerialLatestSaListen & 0x0f);
		SerialPtr = 0;
		STORE(0x90, i);
		if (i) {
		    p -> isopen = 0;
		    p -> closef(SerialLatestSaListen & 0x0f);
		}
	    }
	    if (p -> flushf)
		p -> flushf(SerialLatestSaListen & 0x0f);
	    break;

	    /*
	     * A bit mystery but seems to work this way :-)
	     */
	  case 0x60:
	    if (!p -> isopen) {
		p -> isopen = 1;
		p -> openf(NULL, 0, SerialLatestSaListen & 0x0f);
		for (i = 0; i < SerialPtr; i++)
		    p -> putf(SerialBuffer[i], SerialLatestSaListen & 0x0f);
		SerialPtr = 0;
	    }
	    if (p -> flushf)
		p -> flushf(SerialLatestSaListen & 0x0f);
	    break;

	    /*
	     * Close
	     */
	  case 0xE0:
	    p -> isopen = 0;
	    p -> closef(SerialLatestSaListen & 0x0f);
	    break;

	  default:
	    printf("Unknown salisten\n");
	}
    }
    SerialLatestListen = SerialTrapListen;
    SerialLatestSaListen = b;

    PC = 0xEDAC;
}


/*
 * Initialize serial module. Set up traps.
 */
static trap_t init_traps[] =
{
    {
	"SerialSendByte",
	0xED40,
	{0x78, 0x20, 0x97},
	serialsendbyte,
    },
    {
	"SerialReceiveByte",
	0xEE13,
	{0x78, 0xA9, 0x00},
	serialreceivebyte,
    },
    {
	"SerialListen",
	0xED0E,
	{0x20, 0xA4, 0xF0},
	seriallisten,
    },
    {
	"SerialSaListen",
	0xED36,
	{0x78, 0x20, 0x8E},
	serialsalisten,
    },
};



void    initialize_serial()
{
    int     i;

    /*
     * Which traps to initialize
     */

    set_trap(&init_traps[0]);
    set_trap(&init_traps[1]);
    set_trap(&init_traps[2]);
    set_trap(&init_traps[3]);

    /*
     * Clear serial device functions
     */
    for (i = 0; i < 16; i++) {
	serial_t *p;

	p = &serialdevices[i];

#ifdef __STDC__
	p -> getf = (int (*)(BYTE *, int ))fn;
	p -> putf = (int (*)(BYTE, int ))fn;
	p -> openf = (int (*)(char *, int , int ))fn;
	p -> closef = (int (*)(int ))fn;
	p -> flushf = (void (*)(int ))NULL;
#else
	p -> getf = fn;
	p -> putf = fn;
	p -> openf = fn;
	p -> closef = fn;
	p -> flushf = NULL;
#endif
    }
}




#ifndef __STDC__
int     attach_serial_device(device, name, getf, putf, openf, closef, flushf)
    int     device;
    char   *name;
    int     (*getf) ();
    int     (*putf) ();
    int     (*openf) ();
    int     (*closef) ();
    void    (*flushf) ();
#else	       /* __STDC__ */
int     attach_serial_device(int device, char   *name, 
			     int (*getf)(BYTE * , int), 
			     int (*putf)(BYTE , int), 
			     int (*openf)(char *, int , int), 
			     int (*closef)(int), 
			     void (*flushf)(int))
#endif	       /* __STDC__ */
{
    serial_t *p;

    if (device < 0 || device > 15)
	return 1;

    p = &serialdevices[device];

    if (p -> inuse) {
	printf("warning. serial device %d (%s) in use\n", device, name);
	return 1;
    }
    p -> getf = getf;
    p -> putf = putf;
    p -> openf = openf;
    p -> closef = closef;
    p -> flushf = flushf;
    p -> inuse = 1;

    p -> name = xstrdup(name);
    return 0;
}


/*
 * Call this if device is not attached
 */
static int fn()
{
    return 2;
}
