/*
 * $Id: table.c,v 1.3 93/06/21 13:39:31 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains lookup-table which is used to translate
 * MOS6502 machine instructions. Machine code is used as index
 * to array called lookup. Pointer to function is then fetched
 * from array and function is called.
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log:	table.c,v $
 * Revision 1.3  93/06/21  13:39:31  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:09:35  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include "vmachine.h"
#ifndef HACK
#include "mnemonics.h"
/*
 * File contains #defines for BRK, ORA, NOOP... which
 * are character strings, i.e. #define BRK    "BRK"
 * #define ORA    "ORA" . . . Used mainly to reduce
 * typing...
*/

/*
 * There is #defines for addressing modes i.e. IMPLIED, INDIRECT_X,
 * ZERO_PAGE... These can be used to make a diassembler.
 */

#include "op.h"

int     clength[] = {1, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 1, 0};

struct lookup_tag lookup[] = {

    /****  Positive  ****/

     /* 00 */ {brk, IMPLIED, M_NONE, M_PC, BRK, 7},	/* Pseudo Absolute */
     /* 01 */ {ora, INDIRECT_X, M_INDX, M_AC, ORA, 6},	/* (Indirect,X) */
     /* 02 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 03 */ {u_slor, INDIRECT_X, M_INDX, M_INDX, SLO, 8},

     /* 04 */ {u_nop, ZERO_PAGE, M_NONE, M_NONE, NOOP, 3},
     /* 05 */ {ora, ZERO_PAGE, M_ZERO, M_AC, ORA, 3},	/* Zeropage */
     /* 06 */ {asl, ZERO_PAGE, M_ZERO, M_ZERO, ASL, 5},	/* Zeropage */
     /* 07 */ {u_slor, ZERO_PAGE, M_ZERO, M_ZERO, SLO, 5},

     /* 08 */ {push, IMPLIED, M_SR, M_NONE, PHP, 3},
     /* 09 */ {ora, IMMEDIATE, M_IMM, M_AC, ORA, 2},	/* Immediate */
     /* 0a */ {asl, ACCUMULATOR, M_AC, M_AC, ASL, 2},	/* Accumulator */
     /* 0b */ {u_and, IMMEDIATE, M_ACIM, M_ACNC, ANC, 2},

     /* 0c */ {u_nop, ABSOLUTE, M_NONE, M_NONE, NOOP, 4},
     /* 0d */ {ora, ABSOLUTE, M_ABS, M_AC, ORA, 4},	/* Absolute */
     /* 0e */ {asl, ABSOLUTE, M_ABS, M_ABS, ASL, 6},	/* Absolute */
     /* 0f */ {u_slor, ABSOLUTE, M_ABS, M_ABS, SLO, 6},

     /* 10 */ {bpl, RELATIVE, M_REL, M_NONE, BPL, 2},	/* 10 */
     /* 11 */ {ora, INDIRECT_Y, M_INDY, M_AC, ORA, 5},	/* (Indirect),Y */
     /* 12 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 13 */ {u_slor, INDIRECT_Y, M_INDY, M_INDY, SLO, 8},

     /* 14 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* 15 */ {ora, ZERO_PAGE_X, M_ZERX, M_AC, ORA, 4},	/* Zeropage,X */
     /* 16 */ {asl, ZERO_PAGE_X, M_ZERX, M_ZERX, ASL, 6},	/* Zeropage,X */
     /* 17 */ {u_slor, ZERO_PAGE_X, M_ZERX, M_ZERX, SLO, 6},

     /* 18 */ {clrf, IMPLIED, M_NONE, M_FC, CLC, 2},
     /* 19 */ {ora, ABSOLUTE_Y, M_ABSY, M_AC, ORA, 4},	/* Absolute,Y */
     /* 1a */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* 1b */ {u_slor, ABSOLUTE_Y, M_ABSY, M_ABSY, SLO, 7},

     /* 1c */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* 1d */ {ora, ABSOLUTE_X, M_ABSX, M_AC, ORA, 4},	/* Absolute,X */
     /* 1e */ {asl, ABSOLUTE_X, M_ABSX, M_ABSX, ASL, 7},	/* Absolute,X */
     /* 1f */ {u_slor, ABSOLUTE_X, M_ABSX, M_ABSX, SLO, 7}, /****************/

     /* 20 */ {jsr, ABSOLUTE, M_ADDR, M_PC, JSR, 6},	/* 20 */
     /* 21 */ {and, INDIRECT_X, M_INDX, M_AC, AND, 6},	/* (Indirect ,X) */
     /* 22 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 23 */ {u_rlan, INDIRECT_X, M_INDX, M_INDX, RLA, 8},

     /* 24 */ {bit, ZERO_PAGE, M_ZERO, M_NONE, BIT, 3},	/* Zeropage */
     /* 25 */ {and, ZERO_PAGE, M_ZERO, M_AC, AND, 3},	/* Zeropage */
     /* 26 */ {rol, ZERO_PAGE, M_ZERO, M_ZERO, ROL, 5},	/* Zeropage */
     /* 27 */ {u_rlan, ZERO_PAGE, M_ZERO, M_ZERO, RLA, 5},

     /* 28 */ {plp, IMPLIED, M_NONE, M_SR, PLP, 4},
     /* 29 */ {and, IMMEDIATE, M_IMM, M_AC, AND, 2},	/* Immediate */
     /* 2a */ {rol, ACCUMULATOR, M_AC, M_AC, ROL, 2},	/* Accumulator */
     /* 2b */ {u_and, IMMEDIATE, M_ACIM, M_ACNC, ANC, 2},

     /* 2c */ {bit, ABSOLUTE, M_ABS, M_NONE, BIT, 4},	/* Absolute */
     /* 2d */ {and, ABSOLUTE, M_ABS, M_AC, AND, 4},	/* Absolute */
     /* 2e */ {rol, ABSOLUTE, M_ABS, M_ABS, ROL, 6},	/* Absolute */
     /* 2f */ {u_rlan, ABSOLUTE, M_ABS, M_ABS, RLA, 6},

     /* 30 */ {bmi, RELATIVE, M_REL, M_NONE, BMI, 2},	/* 30 */
     /* 31 */ {and, INDIRECT_Y, M_INDY, M_AC, AND, 5},	/* (Indirect),Y */
     /* 32 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 33 */ {u_rlan, INDIRECT_Y, M_INDY, M_INDY, RLA, 8},

     /* 34 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* 35 */ {and, ZERO_PAGE_X, M_ZERX, M_AC, AND, 4},	/* Zeropage,X */
     /* 36 */ {rol, ZERO_PAGE_X, M_ZERX, M_ZERX, ROL, 6},	/* Zeropage,X */
     /* 37 */ {u_rlan, ZERO_PAGE_X, M_ZERX, M_ZERX, RLA, 6},

     /* 38 */ {setf, IMPLIED, M_NONE, M_FC, SEC, 2},
     /* 39 */ {and, ABSOLUTE_Y, M_ABSY, M_AC, AND, 4},	/* Absolute,Y */
     /* 3a */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* 3b */ {u_rlan, ABSOLUTE_Y, M_ABSY, M_ABSY, RLA, 7},

     /* 3c */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* 3d */ {and, ABSOLUTE_X, M_ABSX, M_AC, AND, 4},	/* Absolute,X */
     /* 3e */ {rol, ABSOLUTE_X, M_ABSX, M_ABSX, ROL, 7},	/* Absolute,X */
     /* 3f */ {u_rlan, ABSOLUTE_X, M_ABSX, M_ABSX, RLA, 7},	/****************/

     /* 40 */ {rti, IMPLIED, M_NONE, M_PC, RTI, 6},	/* 40 */
     /* 41 */ {eor, INDIRECT_X, M_INDX, M_AC, EOR, 6},	/* (Indirect,X) */
     /* 42 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 43 */ {u_sreo, INDIRECT_X, M_INDX, M_INDX, SRE, 8},

     /* 44 */ {u_nop, ZERO_PAGE, M_NONE, M_NONE, NOOP, 3},
     /* 45 */ {eor, ZERO_PAGE, M_ZERO, M_AC, EOR, 3},	/* Zeropage */
     /* 46 */ {asr, ZERO_PAGE, M_ZERO, M_ZERO, LSR, 5},	/* Zeropage */
     /* 47 */ {u_sreo, ZERO_PAGE, M_ZERO, M_ZERO, SRE, 5},

     /* 48 */ {push, IMPLIED, M_AC, M_NONE, PHA, 3},
     /* 49 */ {eor, IMMEDIATE, M_IMM, M_AC, EOR, 2},	/* Immediate */
     /* 4a */ {asr, ACCUMULATOR, M_AC, M_AC, LSR, 2},	/* Accumulator */
     /* 4b */ {u_asr, IMMEDIATE, M_ACIM, M_AC, ASR, 2},	/* (AC & IMM) >>1 */

     /* 4c */ {store, ABSOLUTE, M_ADDR, M_PC, JMP, 3},	/* Absolute */
     /* 4d */ {eor, ABSOLUTE, M_ABS, M_AC, EOR, 4},	/* Absolute */
     /* 4e */ {asr, ABSOLUTE, M_ABS, M_ABS, LSR, 6},	/* Absolute */
     /* 4f */ {u_sreo, ABSOLUTE, M_ABS, M_ABS, SRE, 6},

     /* 50 */ {bvc, RELATIVE, M_REL, M_NONE, BVC, 2},	/* 50 */
     /* 51 */ {eor, INDIRECT_Y, M_INDY, M_AC, EOR, 5},	/* (Indirect),Y */
     /* 52 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 53 */ {u_sreo, INDIRECT_Y, M_INDY, M_INDY, SRE, 8},

     /* 54 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* 55 */ {eor, ZERO_PAGE_X, M_ZERX, M_AC, EOR, 4},	/* Zeropage,X */
     /* 56 */ {asr, ZERO_PAGE_X, M_ZERX, M_ZERX, LSR, 6},	/* Zeropage,X */
     /* 57 */ {u_sreo, ZERO_PAGE_X, M_ZERX, M_ZERX, SRE, 6},

     /* 58 */ {clrf, IMPLIED, M_NONE, M_FI, CLI, 2},
     /* 59 */ {eor, ABSOLUTE_Y, M_ABSY, M_AC, EOR, 4},	/* Absolute,Y */
     /* 5a */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* 5b */ {u_sreo, ABSOLUTE_Y, M_ABSY, M_ABSY, SRE, 7},

     /* 5c */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* 5d */ {eor, ABSOLUTE_X, M_ABSX, M_AC, EOR, 4},	/* Absolute,X */
     /* 5e */ {asr, ABSOLUTE_X, M_ABSX, M_ABSX, LSR, 7},	/* Absolute,X */
     /* 5f */ {u_sreo, ABSOLUTE_X, M_ABSX, M_ABSX, SRE, 7},	/****************/

     /* 60 */ {rts, IMPLIED, M_NONE, M_PC, RTS, 6},	/* 60 */
     /* 61 */ {adc, INDIRECT_X, M_INDX, M_AC, ADC, 6},	/* (Indirect,X) */
     /* 62 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* 63 */ {u_rrad, INDIRECT_X, M_INDX, M_INDX, RRA, 8},

     /* 64 */ {u_nop, ZERO_PAGE, M_NONE, M_NONE, NOOP, 3},
     /* 65 */ {adc, ZERO_PAGE, M_ZERO, M_AC, ADC, 3},	/* Zeropage */
     /* 66 */ {ror, ZERO_PAGE, M_ZERO, M_ZERO, ROR, 5},	/* Zeropage */
     /* 67 */ {u_rrad, ZERO_PAGE, M_ZERO, M_ZERO, RRA, 5},

     /* 68 */ {pla, IMPLIED, M_NONE, M_AC, PLA, 4},
     /* 69 */ {adc, IMMEDIATE, M_IMM, M_AC, ADC, 2},	/* Immediate */
     /* 6a */ {ror, ACCUMULATOR, M_AC, M_AC, ROR, 2},	/* Accumulator */
     /* 6b */ {u_ror, IMMEDIATE, M_ACIM, M_AC, ARR, 2},	/* ARR isn't typo */

     /* 6c */ {store, INDIRECT, M_IND, M_PC, JMP, 5},	/* Indirect */
     /* 6d */ {adc, ABSOLUTE, M_ABS, M_AC, ADC, 4},	/* Absolute */
     /* 6e */ {ror, ABSOLUTE, M_ABS, M_ABS, ROR, 6},	/* Absolute */
     /* 6f */ {u_rrad, ABSOLUTE, M_ABS, M_ABS, RRA, 6},

     /* 70 */ {bvs, RELATIVE, M_REL, M_NONE, BVS, 2},	/* 70 */
     /* 71 */ {adc, INDIRECT_Y, M_INDY, M_AC, ADC, 5},	/* (Indirect),Y */
     /* 72 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT relative? */
     /* 73 */ {u_rrad, INDIRECT_Y, M_INDY, M_INDY, RRA, 8},

     /* 74 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* 75 */ {adc, ZERO_PAGE_X, M_ZERX, M_AC, ADC, 4},	/* Zeropage,X */
     /* 76 */ {ror, ZERO_PAGE_X, M_ZERX, M_ZERX, ROR, 6},	/* Zeropage,X */
     /* 77 */ {u_rrad, ZERO_PAGE_X, M_ZERX, M_ZERX, RRA, 6},

     /* 78 */ {setf, IMPLIED, M_NONE, M_FI, SEI, 2},
     /* 79 */ {adc, ABSOLUTE_Y, M_ABSY, M_AC, ADC, 4},	/* Absolute,Y */
     /* 7a */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* 7b */ {u_rrad, ABSOLUTE_Y, M_ABSY, M_ABSY, RRA, 7},

     /* 7c */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* 7d */ {adc, ABSOLUTE_X, M_ABSX, M_AC, ADC, 4},	/* Absolute,X */
     /* 7e */ {ror, ABSOLUTE_X, M_ABSX, M_ABSX, ROR, 7},	/* Absolute,X */
     /* 7f */ {u_rrad, ABSOLUTE_X, M_ABSX, M_ABSX, RRA, 7},	/****************/

    /****  Negative  ****/

     /* 80 */ {u_nop, IMMEDIATE, M_NONE, M_NONE, NOOP, 2},	/* 80 */
     /* 81 */ {store, INDIRECT_X, M_AC, M_INDX, STA, 6},	/* (Indirect,X) */
     /* 82 */ {u_nop, IMMEDIATE, M_NONE, M_NONE, NOOP, 2},
     /* 83 */ {u_store, INDIRECT_X, M_ANXR, M_INDX, SAX, 6},

     /* 84 */ {store, ZERO_PAGE, M_YR, M_ZERO, STY, 3},	/* Zeropage */
     /* 85 */ {store, ZERO_PAGE, M_AC, M_ZERO, STA, 3},	/* Zeropage */
     /* 86 */ {store, ZERO_PAGE, M_XR, M_ZERO, STX, 3},	/* Zeropage */
     /* 87 */ {u_store, ZERO_PAGE, M_ANXR, M_ZERO, SAX, 3},

     /* 88 */ {dec, IMPLIED, M_YR, M_YR, DEY, 2},
     /* 89 */ {u_nop, IMMEDIATE, M_NONE, M_NONE, NOOP, 2},
     /* 8a */ {load, IMPLIED, M_XR, M_AC, TXA, 2},
    /****  very abnormal: usually AC = AC | #$EE & XR & #$oper  ****/
     /* 8b */ {u_load, IMMEDIATE, M_AXIM, M_AC, ANE, 2},

     /* 8c */ {store, ABSOLUTE, M_YR, M_ABS, STY, 4},	/* Absolute */
     /* 8d */ {store, ABSOLUTE, M_AC, M_ABS, STA, 4},	/* Absolute */
     /* 8e */ {store, ABSOLUTE, M_XR, M_ABS, STX, 4},	/* Absolute */
     /* 8f */ {u_store, ABSOLUTE, M_ANXR, M_ABS, SAX, 4},

     /* 90 */ {bcc, RELATIVE, M_REL, M_NONE, BCC, 2},	/* 90 */
     /* 91 */ {store, INDIRECT_Y, M_AC, M_INDY, STA, 6},	/* (Indirect),Y */
     /* 92 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT relative? */
     /* 93 */ {u_store, INDIRECT_Y, M_ANXR, M_STH0, SHA, 6},

     /* 94 */ {store, ZERO_PAGE_X, M_YR, M_ZERX, STY, 4},	/* Zeropage,X */
     /* 95 */ {store, ZERO_PAGE_X, M_AC, M_ZERX, STA, 4},	/* Zeropage,X */
     /* 96 */ {store, ZERO_PAGE_Y, M_XR, M_ZERY, STX, 4},	/* Zeropage,Y */
     /* 97 */ {u_store, ZERO_PAGE_Y, M_ANXR, M_ZERY, SAX, 4},

     /* 98 */ {load, IMPLIED, M_YR, M_AC, TYA, 2},
     /* 99 */ {store, ABSOLUTE_Y, M_AC, M_ABSY, STA, 5},	/* Absolute,Y */
     /* 9a */ {store, IMPLIED, M_XR, M_SP, TXS, 2},
/*** This is very mysterious command ... */
     /* 9b */ {u_store, ABSOLUTE_Y, M_ANXR, M_STH3, SHS, 5},

     /* 9c */ {u_store, ABSOLUTE_X, M_YR, M_STH2, SHY, 5},
     /* 9d */ {store, ABSOLUTE_X, M_AC, M_ABSX, STA, 5},	/* Absolute,X */
     /* 9e */ {u_store, ABSOLUTE_Y, M_XR, M_STH1, SHX, 5},
     /* 9f */ {u_store, ABSOLUTE_Y, M_ANXR, M_STH1, SHA, 5},	/****************/

     /* a0 */ {load, IMMEDIATE, M_IMM, M_YR, LDY, 2},	/* a0 *//* Immediate */
     /* a1 */ {load, INDIRECT_X, M_INDX, M_AC, LDA, 6},	/* (indirect,X) */
     /* a2 */ {load, IMMEDIATE, M_IMM, M_XR, LDX, 2},	/* Immediate */
     /* a3 */ {u_load, INDIRECT_X, M_INDX, M_ACXR, LAX, 6},

     /* a4 */ {load, ZERO_PAGE, M_ZERO, M_YR, LDY, 3},	/* Zeropage */
     /* a5 */ {load, ZERO_PAGE, M_ZERO, M_AC, LDA, 3},	/* Zeropage */
     /* a6 */ {load, ZERO_PAGE, M_ZERO, M_XR, LDX, 3},	/* Zeropage */
     /* a7 */ {u_load, ZERO_PAGE, M_ZERO, M_ACXR, LAX, 3},

     /* a8 */ {load, IMPLIED, M_AC, M_YR, TAY, 2},
     /* a9 */ {load, IMMEDIATE, M_IMM, M_AC, LDA, 2},	/* Immediate */
     /* aa */ {load, IMPLIED, M_AC, M_XR, TAX, 2},
     /* ab */ {u_load, IMMEDIATE, M_ACIM, M_ACXR, LXA, 2},	/* LXA isn't a typo */

     /* ac */ {load, ABSOLUTE, M_ABS, M_YR, LDY, 4},	/* Absolute */
     /* ad */ {load, ABSOLUTE, M_ABS, M_AC, LDA, 4},	/* Absolute */
     /* ae */ {load, ABSOLUTE, M_ABS, M_XR, LDX, 4},	/* Absolute */
     /* af */ {u_load, ABSOLUTE, M_ABS, M_ACXR, LAX, 4},

     /* b0 */ {bcs, RELATIVE, M_REL, M_NONE, BCS, 2},	/* b0 */
     /* b1 */ {load, INDIRECT_Y, M_INDY, M_AC, LDA, 5},	/* (indirect),Y */
     /* b2 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* b3 */ {u_load, INDIRECT_Y, M_INDY, M_ACXR, LAX, 5},

     /* b4 */ {load, ZERO_PAGE_X, M_ZERX, M_YR, LDY, 4}, /* Zeropage,X */
     /* b5 */ {load, ZERO_PAGE_X, M_ZERX, M_AC, LDA, 4}, /* Zeropage,X */
     /* b6 */ {load, ZERO_PAGE_Y, M_ZERX, M_XR, LDX, 4}, /* Zeropage,Y */
     /* b7 */ {u_load, ZERO_PAGE_Y, M_ZERY, M_ACXR, LAX, 4},

     /* b8 */ {clrf, IMPLIED, M_NONE, M_FV, CLV, 2},
     /* b9 */ {load, ABSOLUTE_Y, M_ABSY, M_AC, LDA, 4},	/* Absolute,Y */
     /* ba */ {load, IMPLIED, M_SP, M_XR, TSX, 2},
     /* bb */ {u_load, ABSOLUTE_Y, M_SABY, M_ACXS, LAS, 4},

     /* bc */ {load, ABSOLUTE_X, M_ABSX, M_YR, LDY, 4},	/* Absolute,X */
     /* bd */ {load, ABSOLUTE_X, M_ABSX, M_AC, LDA, 4},	/* Absolute,X */
     /* be */ {load, ABSOLUTE_Y, M_ABSY, M_XR, LDX, 4},	/* Absolute,Y */
     /* bf */ {u_load, ABSOLUTE_Y, M_ABSY, M_ACXR, LAX, 4},	/****************/

     /* c0 */ {cpy, IMMEDIATE, M_IMM, M_NONE, CPY, 2},	/* c0 *//* Immediate */
     /* c1 */ {cmp, INDIRECT_X, M_INDX, M_NONE, CMP, 6},	/* (Indirect,X) */
     /* c2 */ {u_nop, IMMEDIATE, M_NONE, M_NONE, NOOP, 2},	/* occasional TILT */
     /* c3 */ {u_decp, INDIRECT_X, M_INDX, M_INDX, DCP, 8},

     /* c4 */ {cpy, ZERO_PAGE, M_ZERO, M_NONE, CPY, 3},	/* Zeropage */
     /* c5 */ {cmp, ZERO_PAGE, M_ZERO, M_NONE, CMP, 3},	/* Zeropage */
     /* c6 */ {dec, ZERO_PAGE, M_ZERO, M_ZERO, DEC, 5},	/* Zeropage */
     /* c7 */ {u_decp, ZERO_PAGE, M_ZERO, M_ZERO, DCP, 5},

     /* c8 */ {inc, IMPLIED, M_YR, M_YR, INY, 2},
     /* c9 */ {cmp, IMMEDIATE, M_IMM, M_NONE, CMP, 2},	/* Immediate */
     /* ca */ {dec, IMPLIED, M_XR, M_XR, DEX, 2},
     /* cb */ {u_sbx, IMMEDIATE, M_IMM, M_XR, SBX, 2},

     /* cc */ {cpy, ABSOLUTE, M_ABS, M_NONE, CPY, 4},	/* Absolute */
     /* cd */ {cmp, ABSOLUTE, M_ABS, M_NONE, CMP, 4},	/* Absolute */
     /* ce */ {dec, ABSOLUTE, M_ABS, M_ABS, DEC, 6},	/* Absolute */
     /* cf */ {u_decp, ABSOLUTE, M_ABS, M_ABS, DCP, 6},

     /* d0 */ {bne, RELATIVE, M_REL, M_NONE, BNE, 2},	/* d0 */
     /* d1 */ {cmp, INDIRECT_Y, M_INDY, M_NONE, CMP, 5},	/* (Indirect),Y */
     /* d2 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* d3 */ {u_decp, INDIRECT_Y, M_INDY, M_INDY, DCP, 8},

     /* d4 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* d5 */ {cmp, ZERO_PAGE_X, M_ZERX, M_NONE, CMP, 4},	/* Zeropage,X */
     /* d6 */ {dec, ZERO_PAGE_X, M_ZERX, M_ZERX, DEC, 6},	/* Zeropage,X */
     /* d7 */ {u_decp, ZERO_PAGE_X, M_ZERX, M_ZERX, DCP, 6},

     /* d8 */ {clrf, IMPLIED, M_NONE, M_FD, CLD, 2},
     /* d9 */ {cmp, ABSOLUTE_Y, M_ABSY, M_NONE, CMP, 4},	/* Absolute,Y */
     /* da */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* db */ {u_decp, ABSOLUTE_Y, M_ABSY, M_ABSY, DCP, 7},

     /* dc */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* dd */ {cmp, ABSOLUTE_X, M_ABSX, M_NONE, CMP, 4},	/* Absolute,X */
     /* de */ {dec, ABSOLUTE_X, M_ABSX, M_ABSX, DEC, 7},	/* Absolute,X */
     /* df */ {u_decp, ABSOLUTE_X, M_ABSX, M_ABSX, DCP, 7},	/****************/

     /* e0 */ {cpx, IMMEDIATE, M_IMM, M_NONE, CPX, 2},	/* e0 *//* Immediate */
     /* e1 */ {sbc, INDIRECT_X, M_INDX, M_AC, SBC, 6},	/* (Indirect,X) */
     /* e2 */ {u_nop, IMMEDIATE, M_NONE, M_NONE, NOOP, 2},
     /* e3 */ {u_isbc, INDIRECT_X, M_INDX, M_INDX, ISB, 8},

     /* e4 */ {cpx, ZERO_PAGE, M_ZERO, M_NONE, CPX, 3},	/* Zeropage */
     /* e5 */ {sbc, ZERO_PAGE, M_ZERO, M_AC, SBC, 3},	/* Zeropage */
     /* e6 */ {inc, ZERO_PAGE, M_ZERO, M_ZERO, INC, 5},	/* Zeropage */
     /* e7 */ {u_isbc, ZERO_PAGE, M_ZERO, M_ZERO, ISB, 5},

     /* e8 */ {inc, IMPLIED, M_XR, M_XR, INX, 2},
     /* e9 */ {sbc, IMMEDIATE, M_IMM, M_AC, SBC, 2},	/* Immediate */
     /* ea */ {nop, IMPLIED, M_NONE, M_NONE, NOP, 2},
     /* eb */ {u_sbc, IMMEDIATE, M_IMM, M_AC, USBC, 2},	/* same as e9 */

     /* ec */ {cpx, ABSOLUTE, M_ABS, M_NONE, CPX, 4},	/* Absolute */
     /* ed */ {sbc, ABSOLUTE, M_ABS, M_AC, SBC, 4},	/* Absolute */
     /* ee */ {inc, ABSOLUTE, M_ABS, M_ABS, INC, 6},	/* Absolute */
     /* ef */ {u_isbc, ABSOLUTE, M_ABS, M_ABS, ISB, 6},

     /* f0 */ {beq, RELATIVE, M_REL, M_NONE, BEQ, 2},	/* f0 */
     /* f1 */ {sbc, INDIRECT_Y, M_INDY, M_AC, SBC, 5},	/* (Indirect),Y */
     /* f2 */ {noop, IMPLIED, M_NONE, M_NONE, JAM, 0},	/* TILT */
     /* f3 */ {u_isbc, INDIRECT_Y, M_INDY, M_INDY, ISB, 8},

     /* f4 */ {u_nop, ZERO_PAGE_X, M_NONE, M_NONE, NOOP, 4},
     /* f5 */ {sbc, ZERO_PAGE_X, M_ZERX, M_AC, SBC, 4},	/* Zeropage,X */
     /* f6 */ {inc, ZERO_PAGE_X, M_ZERX, M_ZERX, INC, 6},	/* Zeropage,X */
     /* f7 */ {u_isbc, ZERO_PAGE_X, M_ZERX, M_ZERX, ISB, 6},

     /* f8 */ {setf, IMPLIED, M_NONE, M_FD, SED, 2},
     /* f9 */ {sbc, ABSOLUTE_Y, M_ABSY, M_AC, SBC, 4},	/* Absolute,Y */
     /* fa */ {u_nop, IMPLIED, M_NONE, M_NONE, NOOP, 2},
     /* fb */ {u_isbc, ABSOLUTE_Y, M_ABSY, M_ABSY, ISB, 7},

     /* fc */ {u_nop, ABSOLUTE_X, M_NONE, M_NONE, NOOP, 4},
     /* fd */ {sbc, ABSOLUTE_X, M_ABSX, M_AC, SBC, 4},	/* Absolute,X */
     /* fe */ {inc, ABSOLUTE_X, M_ABSX, M_ABSX, INC, 7},	/* Absolute,X */
     /* ff */ {u_isbc, ABSOLUTE_X, M_ABSX, M_ABSX, ISB, 7}
};
#endif
