/*
 * $Id: traps.c,v 1.3 93/06/21 13:39:40 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 * $Log:	traps.c,v $
 * Revision 1.3  93/06/21  13:39:40  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:09:49  jopi
 * *** empty log message ***
 * 
 * Revision 1.3  1993/06/13  08:39:52  sonninen
 * *** empty log message ***
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * traps.c -- This file implements ROM traps that makes possible to replace
 *            some ROM code with C-language functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "macro.h"
#include "extern.h"

typedef struct _traplist_t {
    struct _traplist_t *next;
    trap_t *trap;
} traplist_t;

static traplist_t *traplist;

#ifdef IDLE_TRAP
static void trap_idle_E5D4();

static trap_t idle_trap =
{
    "Idle trap",
    0xE5D4,
    {0xF0, 0xF7, 0x78},
    trap_idle_E5D4,
};
#endif

void    initialize_traps()
{
    traplist = NULL;

#ifdef IDLE_TRAP
    set_trap(&idle_trap);
#endif
}



#ifndef __STDC__
int     set_trap(t)
    trap_t *t;
#else	       /* __STDC__ */
int     set_trap(trap_t *t)
#endif	       /* __STDC__ */
{
    int     i;
    traplist_t *p;

#ifdef DEBUG
    printf("setting up trap %s at 0x%x\n", t -> name, t -> address);
#endif

    for (i = 0; i < 3; i++) {
	if (load_rom(t -> address + i) != t -> check[i]) {
	    printf("incorrect checkbyte for trap %s. not installed\n",
		t -> name);
	    return 1;
	}
    }

    /*
     * BRK (0x00) is trap-opcode
     */
    save_rom(t -> address, 0x00);
    
    p = (traplist_t *) malloc (sizeof (traplist_t));
    p->next = traplist;
    p->trap = t;
    traplist = p;

    return 0;
}

#ifndef __STDC__
int     remove_trap(t)
    trap_t *t;
#else	       /* __STDC__ */
int     remove_trap(trap_t *t)
#endif	       /* __STDC__ */
{
    traplist_t *p = traplist, *prev = NULL;


#ifdef DEBUG
    printf("removing trap %s at 0x%x\n", t -> name, t -> address);
#endif

    if (load_rom(t -> address) != 0x00)
	printf("incorrect checkbyte for trap %s.\n", t -> name);

    save_rom(t -> address, t->check[0]);
    
    
    while (p) {
	if (p -> trap -> address == PC - 1)
	    break;
	prev = p;
	p = p -> next;
    }
    
    if (!p) {
	printf("cannot remove trap %s.\n", t -> name);
	return -1;
    }

    if (prev)
	prev -> next = p -> next;
    else
	traplist = p -> next;

    free ((char *)p);
    return 0;
}

int     trap_handler()
{
    traplist_t *p = traplist;
    
    while (p) {
	if (p -> trap -> address == PC - 1) {
	    (*p -> trap -> func) ();
	    return 0;
	}
	p = p -> next;
    }

    return 1;
}



#ifdef IDLE_TRAP
static void trap_idle_E5D4()
{
    if (!IF_ZERO())
	PC = 0xe5d6;
    else {
	jump_to_next_alarm();
	PC = 0xe5cd;
    }
}
#endif
