/*
 * $Id: x64.c,v 1.3 93/06/21 13:39:52 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * Patch by
 *   Richard Hable (k3027e7@edvz.uni-linz.ac.at) /RH/
 *
 * $Log:	x64.c,v $
 * Revision 1.3  93/06/21  13:39:52  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  93/06/21  13:09:59  jopi
 * *** empty log message ***
 * 
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */


#define XPIX 320
#define YPIX 200

#define BORDER_WIDTH 4
#define MAX_COLORS 16

#define BITMAPDEPTH 1

/* Bitmap data for icon */
#include "c64map.h"
#define icon_bitmap_bits dummy_bits
#define icon_bitmap_width dummy_width
#define icon_bitmap_height dummy_height


#include "macro.h"
#include "extern.h"

/* Xlib include files */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>

#include "x64.h"

#ifdef __STDC__

#include <stdarg.h>

static void open_win ( void );
static void open_cntl_win ( void );
static void handle_debugwin_event ( XEvent * );
static void handle_mainwin_event ( XEvent * );
static void get_colors ( void );
static int draw_string ( Window w, int x, int y, char *s, ... );
static void redraw_cntl_win ( void );
static void update_cntl_win ( void );
static void update_asm ( void );
static void toggle_cntl_win ( void );
extern void open_screen_win ( Window borderwin );

#else /* __STDC__ */

#include <varargs.h>

static void open_win (/* void */);
static void open_cntl_win (/* void */);
static void handle_debugwin_event (/* XEvent * */);
static void handle_mainwin_event (/* XEvent * */);
static void get_colors (/* void */);
static int draw_string (/* ... */);
static void redraw_cntl_win (/* void */);
static void update_cntl_win (/* void */);
static void update_asm (/* void */);
static void toggle_cntl_win (/* void */);
extern void open_screen_win (/* Window borderwin */);

#endif /* __STDC__ */


/*
 * Display and screen are used as arguments to nearly every Xlib routine
 * so it simplifies routine calls to declare them global; if there were
 * additional source files, these variables would be declared ``extern''
 * in them
 */

Display *display;
int     screen;
static Window  rootw, borderwin;

static GC      debuggc;
static Window  debugwin, debug_stop, debug_step, debug_exit, debug_asm;
static int     debugstate;
static int     halt, trace;

int     redraw;


/* keyboard */

#define MAXDOWN 16

int     keypressed = -1;
int     keys_down = 0;
/*static int     keyarr_modified = 0;*/

BYTE    joy1, joy2;

#ifdef KBDBUG
static int     dbg_dn = 0, dbg_up = 0;
#endif

keyconv keyconvmap[] = {
    {XK_Delete, 0, 1, 0},	/* del */
    {XK_Return, 0, 2, 0},
    {XK_Right, 0, 4, 0},
    {XK_F7, 0, 8, 0},
    {XK_F1, 0, 16, 0},
    {XK_F3, 0, 32, 0},
    {XK_F5, 0, 64, 0},
    {XK_Down, 0, 128, 0},

    {XK_3, 1, 1, 0},
    {XK_w, 1, 2, 0},
    {XK_a, 1, 4, 0},
    {XK_4, 1, 8, 0},
    {XK_z, 1, 16, 0},
    {XK_s, 1, 32, 0},
    {XK_e, 1, 64, 0},
    {XK_Shift_L, 1, 128, 0},

    {XK_5, 2, 1, 0},
    {XK_r, 2, 2, 0},
    {XK_d, 2, 4, 0},
    {XK_6, 2, 8, 0},
    {XK_c, 2, 16, 0},
    {XK_f, 2, 32, 0},
    {XK_t, 2, 64, 0},
    {XK_x, 2, 128, 0},

    {XK_7, 3, 1, 0},
    {XK_y, 3, 2, 0},
    {XK_g, 3, 4, 0},
    {XK_8, 3, 8, 0},
    {XK_b, 3, 16, 0},
    {XK_h, 3, 32, 0},
    {XK_u, 3, 64, 0},
    {XK_v, 3, 128, 0},

    {XK_9, 4, 1, 0},
    {XK_i, 4, 2, 0},
    {XK_j, 4, 4, 0},
    {XK_0, 4, 8, 0},
    {XK_m, 4, 16, 0},
    {XK_k, 4, 32, 0},
    {XK_o, 4, 64, 0},
    {XK_n, 4, 128, 0},

    {XK_plus, 5, 1, 0},
    {XK_p, 5, 2, 0},
    {XK_l, 5, 4, 0},
    {XK_minus, 5, 8, 0},
    {XK_period, 5, 16, 0},
    {XK_colon, 5, 32, 528},
    {XK_at, 5, 64, 528},	/* shifted @ gives 'grave' */
    {XK_comma, 5, 128, 0},

    {XK_sterling, 6, 1, 0},	/* pound sign */
    {XK_bar, 6, 1, 0},		/* pound sign */
    {XK_asterisk, 6, 2, 528},
    {XK_semicolon, 6, 4, 528},
    {XK_Home, 6, 8, 0},

#ifdef NOKEYRELEASE
    {XK_Shift_R, 6, 16, 16},
#else
    {XK_Shift_R, 0, 0, 16},	/* invisible shift ;) */
#endif

    {XK_equal, 6, 32, 528},
    /* { XK_asciicircum, 6,64,0 }, *//* up arrow (US kbd) */
    {XK_asciitilde, 6, 64, 0},	/* up arrow (scandinavian kbd) */
    {XK_slash, 6, 128, 528},

    {XK_1, 7, 1, 0},
    {XK_underscore, 7, 2, 0},	/* right arrow */
    {XK_Control_L, 7, 4, 0},	/* CNTRL */
    {XK_2, 7, 8, 0},
    {XK_space, 7, 16, 0},
    {XK_Alt_L, 7, 32, 0},	/* CBM */
    {XK_q, 7, 64, 0},
    {XK_Escape, 7, 128, 0},	/* STOP */

    /* some keys for DECstations */
    {XK_KP_F1, 0, 16, 0},	/* F1 */
    {XK_KP_F2, 0, 32, 0},	/* F3 */
    {XK_KP_F3, 0, 64, 0},	/* F5 */
    {XK_KP_F4, 0, 4, 0},	/* F7 */
    {XK_Find, 6, 8, 0},		/* Home */
    {XK_Insert, 0, 1, 16},	/* Inst */

#ifdef DECMAP
    {DXK_Remove, 0, 1, 0},	/* del (DECkeysym.h) */
    {XK_Help, 7, 128, 0},	/* because no real ESC-key on the keyboard /RH/ */
#endif

    /* shifted keys */
    {XK_BackSpace, 0, 1, 16},	/* Inst */
    {XK_Linefeed, 0, 2, 16},
    {XK_Left, 0, 4, 16},
    {XK_F8, 0, 8, 16},
    {XK_F2, 0, 16, 16},
    {XK_F4, 0, 32, 16},
    {XK_F6, 0, 64, 16},
    {XK_Up, 0, 128, 16},

    {XK_numbersign, 1, 1, 16},
    {XK_W, 1, 2, 16},
    {XK_A, 1, 4, 16},
    {XK_dollar, 1, 8, 16},
    {XK_Z, 1, 16, 16},
    {XK_S, 1, 32, 16},
    {XK_E, 1, 64, 16},
    /* { XK_Shift_L,	1,128,0 }, */

    {XK_percent, 2, 1, 16},
    {XK_R, 2, 2, 16},
    {XK_D, 2, 4, 16},
    {XK_ampersand, 2, 8, 16},
    {XK_C, 2, 16, 16},
    {XK_F, 2, 32, 16},
    {XK_T, 2, 64, 16},
    {XK_X, 2, 128, 16},

    {XK_quoteright, 3, 1, 16},	/* apostrophe */
    {XK_Y, 3, 2, 16},
    {XK_G, 3, 4, 16},
    {XK_parenleft, 3, 8, 16},
    {XK_B, 3, 16, 16},
    {XK_H, 3, 32, 16},
    {XK_U, 3, 64, 16},
    {XK_V, 3, 128, 16},

    {XK_parenright, 4, 1, 16},
    {XK_I, 4, 2, 16},
    {XK_J, 4, 4, 16},
    /* { XK_0, 		4,8,16 }, */
    {XK_M, 4, 16, 16},
    {XK_K, 4, 32, 16},
    {XK_O, 4, 64, 16},
    {XK_N, 4, 128, 16},

    /* { XK_plus, 5,1,16 }, */
    {XK_P, 5, 2, 16},
    {XK_L, 5, 4, 16},
    /* { XK_minus, 5,8,16 }, */
    {XK_greater, 5, 16, 16},
    {XK_bracketleft, 5, 32, 16},
    {XK_grave, 5, 64, 16},	/* shifted @ */
    {XK_less, 5, 128, 16},

    /* { XK_sterling,	6,1,16 }, */
    {XK_backslash, 6, 1, 16},
    /* { XK_asterisk, 	6,2,16 }, */
    {XK_bracketright, 6, 4, 16},
    /* {XK_Home, 6,8,16 },
       {XK_Shift_R, 6,16,16 },
       {XK_equal, 6,32,16 }, */
    {XK_asciicircum, 6, 64, 16},	/* shift up arrow */
    {XK_question, 6, 128, 16},

    {XK_exclam, 7, 1, 16},
    /* { XK_underscore,	7,2,16 }, *//* right arrow */
    /* { XK_Control_L,	7,4,16 }, *//* CNTRL */
    {XK_quotedbl, 7, 8, 16},
    /* { XK_space, 7,16,16 },
       { XK_Alt_L, 7,32,16 }, *//* CBM */
    {XK_Q, 7, 64, 16},
    /* { XK_Escape, 7,128,16 }, *//* STOP */
    {0, 0, 0, 0}
};

#define CONV_KEYS (sizeof(keyconvmap)/sizeof(keyconv))

#ifdef NOKEYRELEASE
CLOCK   keydelay[CONV_KEYS];	/* keypress time for each key */
#endif

/* colors */
unsigned long bgcolors[20];
/* colors for border, background & sprites */

XColor  exact_defs[MAX_COLORS];
unsigned long colors[MAX_COLORS];
GC      cgcs[MAX_COLORS];

static char   *cnames[] = {
    "Black",
    "White",
    "Red",
    "Cyan",
    "Purple",
    "Green3",
    "Blue",
    "Yellow",
    "Orange",
    "Brown",
    "Pink",
    "Dim Gray",
    "Gray",
    "Pale Green",
    "Light Blue",
    "Light Gray",
};


/* fonts */

static XFontStruct *font_info;	/* Structure containing font info */
static int     font_height;
static int     font_width;

static char   *font_list[] = {
    "-*-helvetica-bold-r-*-*-12-*",
    "-*-courier-bold-r-*-*-12-*",
    "-*-times-bold-r-*-*-12-*",
    "fixed",
    NULL,
};

unsigned int width, height;

/* ------------------------------------------------------------------------- */

#ifndef __STDC__
static void open_win()
#else	       /* __STDC__ */
static void open_win(void)
#endif	       /* __STDC__ */
{
    char   *window_name = "C64 Emulator";
    char   *icon_name = "C64";
    unsigned int display_width, display_height;

    XTextProperty windowName, iconName;
    XWMHints wm_hints;
    XClassHint class_hints;
    XSizeHints size_hints;	/* Preferred sizes for window man */
    XSetWindowAttributes wattr;

    /* Display size is a member of display structure */
    display_width = DisplayWidth(display, screen);
    display_height = DisplayHeight(display, screen);

    size_hints.width = width = XPIX + 60;
    size_hints.height = height = YPIX + 60;
    size_hints.min_width = XPIX;
    size_hints.min_height = YPIX;
    size_hints.x = (display_width - width) / 2;
    size_hints.y = (display_height - height) / 2;
    size_hints.flags = PSize | PMinSize | PPosition;

    wattr.backing_store = Always;
    wattr.background_pixel = bgcolors[C_BORDER];
    wattr.cursor = XCreateFontCursor(display, XC_gumby);
    wattr.event_mask = ExposureMask | KeyReleaseMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask |
	FocusChangeMask |
	EnterWindowMask | LeaveWindowMask;

    borderwin =
	XCreateWindow(display, rootw, size_hints.x, size_hints.y,
	width, height, 0,
	CopyFromParent, CopyFromParent, CopyFromParent,
	CWBackPixel | CWEventMask | CWCursor,
	&wattr);

    /* Display window */
    XMapWindow(display, borderwin);
    if (debugflg)
	printf("Window open\n");

    XSetStandardProperties(display, borderwin, "hmm", "64",
	None, NULL, 0, &size_hints);
    if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for windowName failed.\n",
	    progname);
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for iconName failed.\n",
	    progname);
	exit(-1);
    }
    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "C64";

    XSetWMProperties(display, borderwin, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);

}

/*---------------------------------------------------------------------------*/

#ifndef __STDC__
static void open_cntl_win()
#else	       /* __STDC__ */
static void open_cntl_win(void)
#endif	       /* __STDC__ */
{
    XSizeHints size_hints;
    XTextProperty windowName, iconName;
    XWMHints wm_hints;
    XClassHint class_hints;

    char   *window_name = "C64 Control";
    char   *icon_name = "C64 CTRL";

    size_hints.min_width = size_hints.width = font_width * 27;
    size_hints.min_height = size_hints.height = font_height * 16;
    size_hints.flags = PSize | PMinSize;

    debugwin = XCreateSimpleWindow(display, rootw, 0, 0,
	size_hints.width, size_hints.height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    if (debugflg)
	printf("Control Window open\n");

    XSetStandardProperties(display, debugwin, "hmm", "64",
	None, NULL, 0, &size_hints);

    /* Select event types wanted */
    XSelectInput(display, debugwin,
	ExposureMask | KeyReleaseMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask |
	FocusChangeMask | EnterWindowMask | LeaveWindowMask);

    if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for windowName failed.\n",
	    progname);
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf(stderr,
	    "%s: structure allocation for iconName failed.\n",
	    progname);
	exit(-1);
    }
    wm_hints.initial_state = NormalState;

    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "C64";

    XSetWMProperties(display, debugwin, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);


    debug_stop = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 10,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_stop);

    debug_step = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 12,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_step);

    debug_exit = XCreateSimpleWindow(display, debugwin,
	font_width, font_height * 14,
	font_width * 10, font_height,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_exit);

    debug_asm = XCreateSimpleWindow(display, debugwin,
	font_width * 13, font_height * 1,
	font_width * 13, font_height * 14,
	BORDER_WIDTH,
	BlackPixel(display, screen),
	WhitePixel(display, screen));
    XMapWindow(display, debug_asm);

    /* Display window XMapWindow (display, debugwin); */
}


/*---------------------------------------------------------------------------*/

#ifndef __STDC__
void    x_start(argc, argv)
    int     argc;
    char  **argv;
#else	       /* __STDC__ */
void    x_start(int argc, char **argv)
#endif	       /* __STDC__ */
{
    char   *display_name = NULL;	/* Server to connect to */
    char  **p = font_list;
    Pixmap icon_pixmap;

    if (debugflg)
	printf("Xstart\n");

    /* Connect to X server */
    if ((display = XOpenDisplay(display_name)) == NULL) {
	(void) fprintf(stderr, "%s: cannot connect to X server %s\n",
	    progname, XDisplayName(display_name));
	exit(-1);
    }
    /* Get screen size from display structure macro */
    screen = DefaultScreen(display);
    rootw = RootWindow(display, screen);

    get_colors();

    while (p && *p && !(font_info = XLoadQueryFont(display, *p))) {
	(void) fprintf(stderr, "%s: allocation of font %s failed.\n",
	    progname, *p);
	exit(-1);
    }
    font_height = font_info -> ascent + font_info -> descent;
    font_width = font_info -> max_bounds.width;

    XSetFont(display, debuggc, font_info -> fid);

    open_win();
    open_cntl_win();

    open_screen_win (borderwin);
}	       /* End x_start */


/* ------------------------------------------------------------------------- */

#ifndef __STDC__
void    x_exit()
#else	       /* __STDC__ */
void    x_exit(void)
#endif	       /* __STDC__ */
{
    int     i;

    XAutoRepeatOn(display);
    for (i = 0; i < MAX_COLORS; i++)
	XFreeGC(display, cgcs[i]);
    if (font_info)
	XFreeFont(display, font_info);
    XCloseDisplay(display);
}

/* ------------------------------------------------------------------------- */

#ifndef __STDC__
static void handle_debugwin_event(report)
    XEvent *report;
#else	       /* __STDC__ */
static void handle_debugwin_event(XEvent *report)
#endif	       /* __STDC__ */
{
    /* printf("debugwin event\n"); */
    switch (report -> type) {
      case Expose:
	/*
	 * Unless this is the last contiquous expose, to place_text and
	 * place_notify in next Expose
	 */
	if (report -> xexpose.count != 0)
	    break;

	redraw_cntl_win();
	break;

      case ButtonPress:
	if (report -> xbutton.subwindow == debug_stop) {
	    halt = !halt;
	    redraw_cntl_win();

	} else if (report -> xbutton.subwindow == debug_step) {
	    trace = !trace;

	} else if (report -> xbutton.subwindow == debug_exit) {
	    end64();

	} else
	    switch (report -> xbutton.button) {
	      case Button1:
		debugflg = !debugflg;
		break;
	      case Button2:
		/* mon (PC); */
	      case Button3:
		/* open debugging window */
		toggle_cntl_win();
	    }
	break;
      case KeyPress:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count;
	    int     show_keyarr();

	    count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	    switch (key) {
	      case XK_Up:
		joy1 |= 1;
		break;
	      case XK_Down:
		joy1 |= 2;
		break;
	      case XK_Left:
		joy1 |= 4;
		break;
	      case XK_Right:
		joy1 |= 8;
		break;
	      case XK_Return:
		joy1 |= 16;
		break;
	    }
	    if (debugflg)
		printf("press joy1 %x\n", joy1);
	    break;
	}

      case KeyRelease:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count;
	    int     show_keyarr();

	    count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	    switch (key) {
	      case XK_Up:
		joy1 &= ~1;
		break;
	      case XK_Down:
		joy1 &= ~2;
		break;
	      case XK_Left:
		joy1 &= ~4;
		break;
	      case XK_Right:
		joy1 &= ~8;
		break;
	      case XK_Return:
		joy1 &= ~16;
		break;
	    }
	    if (debugflg)
		printf("release joy1 %x\n", joy1);
	    break;
	}

      case FocusIn:
      case EnterNotify:
	XAutoRepeatOff(display);
	break;

      case FocusOut:
      case LeaveNotify:
	XAutoRepeatOn(display);
	break;
    }
}

#ifndef __STDC__
static void handle_mainwin_event(report)
    XEvent *report;
#else	       /* __STDC__ */
static void handle_mainwin_event(XEvent *report)
#endif	       /* __STDC__ */
{
    switch (report -> type) {

      case Expose:
	if (debugflg)
	    printf("Expose %d\n", report -> xexpose.count);
	/* Unless this is the last contiquous expose, don't draw the window */
	if (report -> xexpose.count != 0)
	    break;

	/* Place text in window */
	XClearWindow(display, borderwin);
	x_redraw();
	break;

      case ConfigureNotify:
	if (debugflg)
	    printf("Config notify\n");
	/* window has been resized; change width and height */
	height = report -> xconfigure.height;
	width = report -> xconfigure.width;
	break;

      case ButtonPress:
	switch (report -> xbutton.button) {
	  case Button1:
	    /* debugflg = !debugflg; */
	    break;
	  case Button2:
	    /* -- drop to ass; continue with x or C-d */
	    mon(PC);
	    break;
	  case Button3:
	    /* open debugging window */
	    toggle_cntl_win();
	}
	break;

      case KeyPress:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count, i;
	    int     show_keyarr();

#ifdef KBDBUG
	    char   *p;
	    dbg_dn++;	/* event counter */
#endif

	    count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	    if (key == XK_Tab)	/* Restore key */
		restore();

	    for (i = 0; i < CONV_KEYS; i++) {
		if (key == keyconvmap[i].sym) {

		    keyarr[keyconvmap[i].row] |= keyconvmap[i].keybit;
		    if (keyconvmap[i].shift < 0x100)
			keyarr[6] |= (BYTE) keyconvmap[i].shift;
		    else
			keyarr[6] &= ~0x10;
		    keypressed = i;

#ifdef NOKEYRELEASE

#ifdef KBDBUG
		    printf("%d\n", clk - keydelay[i]);
#endif

		    if (!keydelay[i])	/* set pressed keys */
			keys_down++;
		    keydelay[i] = clk + KEYDELAY;	/* c64 system clock */
#else
		    keys_down++;
#endif

		    if (debugflg)
			printf("KEY FOUND %d\n", report -> xkey.time);
		    break;
		}
	    }

	    if (debugflg) {
		show_keyarr();
		printf("count %d Key %d %x\n", count, (int) *buffer, key);
	    }
	    /* this hack puts keystrings directly into char buffer */
#ifdef KBD_HACK
#ifdef KBD_HACKALL
	      if( ( key == XK_Escape ) /*RH*/
		|| ( key == XK_Help )
		|| ( key == XK_Home )
		|| ( key == XK_Control_L )
		|| ( key == XK_Alt_L )
		|| ( key == XK_Delete )
		|| ( key == XK_Right )
	 	|| ( key == XK_Down )
		|| ( key == XK_Find )
		|| ( key == XK_Insert )
		|| ( ( key >= XK_F1 ) && ( key <= XK_F8 ) )  /* F1..F8 */
		|| ( key == XK_Left )
		|| ( key == XK_Up ) ) {;} else {
#endif
	     if (count>0) { if( ((*buffer & 0xc0) == 0x40)&&((*buffer
	       &0x1f) <0x1e) ) *buffer ^= 0x20; ram[0x0277 + ram[0xc6]++] =
	       *buffer; }

#ifdef KBD_HACKALL
	} /*RH*/
#endif
#endif

#ifdef KBDBUG
	    p = XKeysymToString(key);
	    printf("%8s (%d)\n", p, keypressed);

	    printf("%2d keys: %3d %2d %2d %2d  %2d %2d %2d %2d\n",
		keys_down, keyarr[0], keyarr[1], keyarr[2], keyarr[3],
		keyarr[4], keyarr[5], keyarr[6], keyarr[7]);
#endif

	    break;
	}

#ifndef NOKEYRELEASE
      case KeyRelease:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count, j;

#ifdef KBDBUG
	    dbg_up++;
	    printf("left: %d\n", dbg_dn - dbg_up);
#endif

	    count = XLookupString((XKeyEvent *) report,
		buffer, 20, &key, &compose);
	    for (j = 0; j < CONV_KEYS; j++)
		if (key == keyconvmap[j].sym) {
		    keypressed = j;
		    break;
		}
	    if (keypressed >= 0) {
		if (!(--keys_down)) {

#ifdef KBDBUG
		    printf("release: no keys left\n");
#endif

		    for (j = 0; j < 8; j++)
			keyarr[j] = 0;
		    keypressed = -1;
		} else
		    /* do not use "XOR" here... */
		    keyarr[keyconvmap[keypressed].row] &=
			~keyconvmap[keypressed].keybit;
		/* keyarr[6] &= ~keyconvmap[keypressed].shift; */

		if (debugflg)
		    printf("KEY RELEASED %d\n", keypressed);
	    }
	    if (debugflg)
		printf("count %d Key %d %x\n", count, (int) *buffer, key);
	    break;
	}

      case FocusIn:
      case EnterNotify:

	XAutoRepeatOff(display);
	break;
#endif

      case FocusOut:
      case LeaveNotify:
	{
	    register int i, j;

#ifndef NOKEYRELEASE
	    XAutoRepeatOn(display);
#endif

#ifdef KBDBUG
	    dbg_dn = 0;
	    dbg_up = 0;
	    printf("no focus\n");
#endif

	    /* this is just to recover from server's panic */

	    for (i = 0, j = 0; j < 8; j++)
		keyarr[j] = 0;

	    keys_down = 0;

#ifdef NOKEYRELEASE
	    for (i = 0; i < CONV_KEYS; i++)
		keydelay[i] = 0;	/* clear kbd */
#endif
	}
	break;

      case ClientMessage:
	flush_queue();
	break;

      default:
	/*
	 * All events selected by StructureNotifyMask except ConfigureNotify
	 * are thrown away here, since nothing is done with them
	 */
	break;
    }	       /* End switch */
}

#ifndef __STDC__
void    x_loop()
#else	       /* __STDC__ */
void    x_loop(void)
#endif	       /* __STDC__ */
{
    XEvent  report;	/* Structure for event information */

    if (debugstate)
	update_cntl_win();

    if (redraw) {
	x_redraw();
	redraw = 0;
    }

#ifndef SETOWN
    if (queuesize > 0)
	flush_queue();
#endif

    while (XEventsQueued(display, QueuedAfterFlush) > 0 || (halt && !trace)) {
	XNextEvent(display, &report);

	/* Control Window Event */
	if (report.xany.window == debugwin) {
	    handle_debugwin_event(&report);
	    continue;
	}
	/* Main Window Event */
	if (report.xany.window == borderwin) {
	    handle_mainwin_event(&report);
	    continue;
	}
    }	       /* End while */

    if (trace) {
	int_clk[I_SPECX] = clk;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	if (halt)
	    trace = 0;
    }
}

/* ------------------------------------------------------------------------- */

#ifndef __STDC__
static void get_colors()
#else	       /* __STDC__ */
static void get_colors(void)
#endif	       /* __STDC__ */
{
    int     def_depth = DefaultDepth(display, screen);
    XColor  exact_def, rgb_def;
    Colormap def_cmap = DefaultColormap(display, screen);
    int     i;
    XGCValues values;

    if (def_depth < 4) {
	printf("Cannot allocate colours\n");
	exit(0);
    }
    /* allocate read/write colors for backgrounds & sprites */
    if (!XAllocColorCells(display, def_cmap, False, NULL, 0, bgcolors, 20)) {
	printf("Cannot allocate r/w colors\n");
	exit(0);
    }
    /* allocate read only colors for characters in screen */
    for (i = 0; i < MAX_COLORS; i++) {
	if (!XAllocNamedColor
	    (display, def_cmap, cnames[i], &exact_def, &rgb_def)) {
	    printf("Cannot allocate color %s\n", cnames[i]);
	    exit(0);
	}
	values.foreground = exact_def.pixel;

	cgcs[i] = XCreateGC(display, rootw, GCForeground, &values);

	colors[i] = exact_def.pixel;
	exact_defs[i] = exact_def;
	exact_defs[i].flags = DoRed | DoGreen | DoBlue;
    }

    values.foreground = BlackPixel(display, screen);
    values.background = WhitePixel(display, screen);
    debuggc = XCreateGC(display, rootw, GCForeground | GCBackground, &values);
}


#ifndef __STDC__
static int draw_string(va_alist) va_dcl
{
    va_list args;
    char    buf[1000];

    Window  w;
    int     x, y;
    char   *s;
    int     vsprintf();

    va_start(args);
    w = va_arg(args, Window);
    x = va_arg(args, int);
    y = va_arg(args, int);
    s = va_arg(args, char *);

    vsprintf(buf, s, args);
    va_end(args);

#else	       /* __STDC__ */

static int draw_string(Window w, int x, int y, char *s,...)
{
    va_list args;
    char    buf[1000];
    va_start(args, s);
    vsprintf(buf, s, args);
    va_end(args);
#endif	       /* __STDC__ */

    XDrawImageString(display, w, debuggc,
	x * font_width, font_info -> ascent + y * font_height,
	buf, strlen(buf));

    return (0);
}

#ifndef __STDC__
static void    redraw_cntl_win()
#else	       /* __STDC__ */
static void    redraw_cntl_win(void)
#endif	       /* __STDC__ */
{
    XClearWindow(display, debugwin);
    draw_string(debugwin, 1, 1, "AC:");
    draw_string(debugwin, 1, 2, "XR:");
    draw_string(debugwin, 1, 3, "YR:");
    draw_string(debugwin, 1, 4, "SP:");
    draw_string(debugwin, 1, 5, "PC:");
    draw_string(debugwin, 1, 6, "SR:");
    /* draw_string(debugwin, 1, 7, "MEM:");*/

    draw_string(debug_stop, 3, 0, halt ? "CONT " : "STOP ");
    draw_string(debug_step, 3, 0, halt ? "STEP  " : "TRACE ");
    draw_string(debug_exit, 3, 0, "EXIT");
    update_cntl_win();
    update_asm();
}

#ifndef __STDC__
static void    update_cntl_win()
#else	       /* __STDC__ */
static void    update_cntl_win(void)
#endif	       /* __STDC__ */
{
    draw_string(debugwin, 4, 1, " %02X ", AC);
    draw_string(debugwin, 4, 2, " %02X ", XR);
    draw_string(debugwin, 4, 3, " %02X ", YR);
    draw_string(debugwin, 4, 4, " %02X ", SP);
    draw_string(debugwin, 4, 5, " %04X ", PC);

    draw_string(debugwin, 4, 6, " %s  ", sprint_status());
   /* draw_string(debugwin, 4, 7, " %02X ", src); */
    draw_string(debugwin, 1, 8, "%-15s", sprint_opcode(PC, 1));

    if (halt || trace)
	update_asm();
}

#ifndef __STDC__
static void    update_asm()
#else	       /* __STDC__ */
static void    update_asm(void)
#endif	       /* __STDC__ */
{
    ADDRESS p = PC;
    int     i;

    for (i = 0; i < 14; i++) {
	draw_string(debug_asm, 1, i, " %-15s", sprint_opcode(p, 1));
	p += clength[lookup[LOAD(p)].addr_mode];
    }
    draw_string(debug_asm, 0, 0, ">");
    draw_string(debugwin, 14, 0, " %04X  ", PC);
}

#ifndef __STDC__
static void    toggle_cntl_win()
#else	       /* __STDC__ */
static void    toggle_cntl_win(void)
#endif	       /* __STDC__ */
{
    if ((debugstate = !debugstate)) {
	XMapWindow(display, debugwin);
    } else {
	XUnmapWindow(display, debugwin);
    }
}


